/* 
 * Copyright (c) 2001 by Matt Welsh and The Regents of the University of 
 * California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Matt Welsh <mdw@cs.berkeley.edu>
 * 
 */

package mdw.sandStorm.internal;

import mdw.sandStorm.api.*;
import mdw.sandStorm.api.internal.*;
import mdw.sandStorm.core.*;
import java.util.*;

/**
 * A StageWrapper is a basic implementation of StageWrapperIF for 
 * application-level stages.
 * 
 * @author   Matt Welsh
 */

class StageWrapper implements StageWrapperIF {

  private String name;
  private StageIF stage;
  private EventHandlerIF handler;
  private ConfigDataIF config;
  private FiniteQueue eventQ;
  private ThreadManagerIF threadmgr;
  private StageStats stats;

  /**
   * Create a StageWrapper with the given name, handler, config data, and 
   * thread manager.
   */
  StageWrapper(String name, EventHandlerIF handler, ConfigDataIF config, ThreadManagerIF threadmgr) {
    this.name = name;
    this.handler = handler;
    this.config = config;
    this.threadmgr = threadmgr;
    eventQ = new FiniteQueue(name);

    this.stage = new Stage(name, this, (SinkIF)eventQ, config);
    config.setStage(this.stage);
    stats = new StageStats(this);
  }

  /**
   * Create a StageWrapper with the given name, handler, config data, thread 
   * manager, and queue threshold.
   */
  StageWrapper(String name, EventHandlerIF handler, ConfigDataIF config, ThreadManagerIF threadmgr, int queueThreshold) {
    this.name = name;
    this.handler = handler;
    this.config = config;
    this.threadmgr = threadmgr;

    eventQ = new FiniteQueue(name);
    QueueThresholdPredicate pred = new QueueThresholdPredicate(eventQ, queueThreshold);
    eventQ.setEnqueuePredicate(pred);

    this.stage = new Stage(name, this, (SinkIF)eventQ, config);
    config.setStage(this.stage);
    stats = new StageStats(this);
  }

  /**
   * Initialize this stage.
   */
  public void init() throws Exception {
    handler.init(config);
    threadmgr.register(this);
  }

  /**
   * Destroy this stage.
   */
  public void destroy() throws Exception {
    threadmgr.deregister(this);
    handler.destroy();
  }

  /**
   * Return the event handler associated with this stage.
   */
  public EventHandlerIF getEventHandler() {
    return handler;
  }

  /**
   * Return the stage handle for this stage.
   */
  public StageIF getStage() {
    return stage;
  }

  /**
   * Return the set of sources from which events should be pulled to 
   * pass to this EventHandlerIF.
   */
  public SourceIF getSource() {
    return (SourceIF)eventQ;
  }

  /**
   * Return the thread manager which will run this stage.
   */
  public ThreadManagerIF getThreadManager() {
    return threadmgr;
  }

  /**
   * Return execution statistics for this stage.
   */
  StageStats getStats() {
    return stats;
  }

  public String toString() {
    return "SW["+stage.getName()+"]";
  }

}

