/* 
 * Copyright (c) 2000 by Matt Welsh and The Regents of the University of 
 * California. All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: Matt Welsh <mdw@cs.berkeley.edu>
 * 
 */

package mdw.sandStorm.lib.aSocket;

import mdw.nbio.*;
import mdw.sandStorm.api.*;
import mdw.sandStorm.core.*;
import java.io.*;
import java.net.*;

/**
 * This is a simple extension of the AMcastSocket class, which creates
 * uses a NonblockingMulticastSocket as the underlying socket, rather
 * than a NonblockingDatagramSocket.
 *
 * NOTE: packets cannot be received on a socket once connect() is
 * called, due to the semantics of connect() for multicast sockets.
 * Instead, applications should generally use joinGroup(), and then
 * explicitly specify the group address as the destination in all the
 * outbound packets.
 * 
 * FIXME: possibly change the behavior of this class, to mimic more
 * normal connected behavior?  This would simply involve _not_ doing
 * the connect when requested, but instead keeping track of the
 * outbound address to use, and filling in all blank outbound packets
 * with it.  Would this break expected semantics somewhere?
 *
 * @author Rob von Behren <jrvb@cs.berkeley.edu> 
 * @see NonblockingMulticastSocket
 * @see AMcastSocket
 **/
public class AMcastSocket extends AUdpSocket {

  /**
   * Create a socket bound to the given local port. 
   */
  public AMcastSocket(int localport, SinkIF compQ) throws IOException {
    this(null, localport, compQ, DEFAULT_MAX_PACKETSIZE, -1);
  }

  /**
   * Create a socket bound to the given local address and local port.
   *
   * @param maxPacketSize The maximum size, in bytes, of packets that
   * this socket will attempt to receive. The default is 
   * DEFAULT_MAX_PACKETSIZE, which is 16 KBytes.
   *
   * @param writeClogThreshold The maximum number of outstanding writes
   * on this socket before a SinkCloggedEvent is pushed to the
   * connection's completion queue. This is effectively the maximum depth
   * threshold for this connection's SinkIF. The default value is -1, which
   * indicates that no SinkCloggedEvents will be generated.
   *
   */
  public AMcastSocket(InetAddress localaddr, int localport, SinkIF compQ, int maxPacketSize, int writeClogThreshold) throws IOException {

    //super(null, port, compQ, maxPacketSize, writeClogThreshold);

    // NOTE: the superclass constror is not used, since it creates
    // the wrong kind of underlying socket.  With the exception of
    // the socket type, however, the rest is copied verbatim from the
    // superclass.
    this.remaddress = null;
    this.remport = -1;
    this.maxPacketSize = maxPacketSize;
    this.writeClogThreshold = writeClogThreshold;
    this.compQ = compQ;

    this.dgsock = new NonblockingMulticastSocket(localport, localaddr);
    sockState = new DatagramSockState(this);
  }



  // FIXME: should these be made asynchronous?  all they do is a JNI
  // call w/ a setsockopt()

  /**
   * join a multicast group
   */
  public void joinGroup(InetAddress addr) throws IOException {
    ((NonblockingMulticastSocket)dgsock).joinGroup(addr);
  }

  /**
   * leave a multicast group
   */
  public void leaveGroup(InetAddress addr) throws IOException {
    ((NonblockingMulticastSocket)dgsock).leaveGroup(addr);
  }

  /**
   * set the multicast ttl
   */
  public void setTimeToLive(int ttl) throws IOException {
    ((NonblockingMulticastSocket)dgsock).setTimeToLive(ttl);
  }

  /**
   * get the multicast ttl
   */
  public int getTimeToLive() throws IOException {
    return ((NonblockingMulticastSocket)dgsock).getTimeToLive();
  }

  /**
   * set the multicast interface
   */
  public void setInterface(InetAddress addr) throws IOException {
    ((NonblockingMulticastSocket)dgsock).setInterface(addr);
  }

  /**
   * set the multicast interface
   */
  public InetAddress getInterface() throws IOException {
    return ((NonblockingMulticastSocket)dgsock).getInterface();
  }

  /**
   * Set the looping state of the Multicast socket.  If state==true,
   * the socket will receive packets it sends out.  If false, it will
   * not.
   *
   * <b>NOTE: This doesn't work like you would probably expect it to!!
   * See the NonblockingMulticastSocket docs for more info.</b>
   *
   * @see NonblockingMulticastSocket
   **/
  public void seeLocalMessages(boolean state) throws IOException {
    ((NonblockingMulticastSocket)dgsock).seeLocalMessages(state);
  }


}


