/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;
import ostore.util.Assert;
import ostore.util.ByteUtils;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.QuickSerializable;
import ostore.util.Types;
import ostore.network.NetworkMessage;
import java.util.Vector;

/**
 * Update message.
 *
 * @author     Sean C. Rhea
 * @version    $Id: UpdateMessage.java,v 1.1 2001/12/01 01:42:59 srhea Exp $
 */
public class UpdateMessage extends NetworkMessage {

    private static final boolean DEBUG = DebugFlags.BLOOM;

    public UpdateId id;
    public long local_seq;
    public int filter_width, filter_depth;

    public UpdateMessage (NodeId peer, UpdateId id, long local_seq,
	    int filter_width, int filter_depth) {
	super (peer, false /* outbound */);
	this.id = id;
	this.local_seq = local_seq;
	this.filter_width = filter_width;
	this.filter_depth = filter_depth;
	this._bits = new Vector (10, 10);
    }

    public UpdateMessage (byte [] data, int [] offset) {
	super (data, offset);
	id = new UpdateId (data, offset);
	local_seq = ByteUtils.bytesToLong (data, offset);
	filter_width = ByteUtils.bytesToInt (data, offset);
	filter_depth = ByteUtils.bytesToInt (data, offset);

	// TODO: faking this for now
	Assert.assert (false, "BUG");
    }

    public void to_bytes (byte [] data, int [] offset) {
	super.to_bytes (data, offset);
	id.to_bytes (data, offset);
	ByteUtils.longToBytes (local_seq, data, offset);
	ByteUtils.intToBytes (filter_width, data, offset);
	ByteUtils.intToBytes (filter_depth, data, offset);

	// TODO: faking this for now
	Assert.assert (data == null, "BUG");
	int largest_bit_value = filter_width - 1;
	int bits_per_bit = 0;
	while (largest_bit_value-- > 0)
	    ++bits_per_bit;
	bits_per_bit += 1; // for sign
	int bits_for_list = bits_per_bit * _bits.size ();
	bits_for_list = (bits_for_list - 1) / 8 + 1; // round to nearest byte
	offset [0] += bits_for_list / 8;
    }

    public int type_code () {
	return Types.BLOOM_UPDATE_MESSAGE;
    }

    public void add_bit (int row, int col, boolean clear) {
	if (DEBUG) Assert.assert (row >= 0, "BUG");
	if (DEBUG) Assert.assert (col >= 0, "BUG");
	if (DEBUG) Assert.assert (row < filter_depth, "BUG");
	if (DEBUG) Assert.assert (col < filter_width, "BUG");
	if (DEBUG) Assert.assert (! has_bit (row, col, true), "BUG");
	if (DEBUG) Assert.assert (! has_bit (row, col, false), "BUG");
	int bit = row * filter_width + col;
	if (! clear)
	    bit |= 0x80000000;
	_bits.add (_bits.size (), new Integer (bit));

    }

    public void rem_bit (int row, int col, boolean clear) {
	if (DEBUG) Assert.assert (row < filter_depth, "BUG");
	if (DEBUG) Assert.assert (col < filter_width, "BUG");
	int bit = row * filter_width + col;
	if (clear)
	    bit |= 0x80000000;

	int i = 0; 
	while (((Integer) _bits.get (i)).intValue () != bit)
	    if (++i >= _bits.size ())
		if (DEBUG) Assert.assert (false, "BUG");
	if (DEBUG) Assert.assert (
		((Integer) _bits.get (i)).intValue () == bit, "BUG");

	while (i < _bits.size ()) {
	    _bits.set (i, _bits.get (i+1));
	    ++i;
	}

	_bits.remove (_bits.size () - 1);
    }

    public boolean has_bit (int row, int col, boolean clear) {
	if (DEBUG) Assert.assert (row < filter_depth, "BUG");
	if (DEBUG) Assert.assert (col < filter_width, "BUG");
	int bit = row * filter_width + col;
	if (clear)
	    bit |= 0x80000000;
	for (int i = 0; i < _bits.size (); ++i) {
	    if (((Integer) _bits.get (i)).intValue () == bit)
		return true;
	}
	return false;
    }

    public int num_bits () {
	return _bits.size ();
    }

    public int row (int b) {
	if (DEBUG) Assert.assert (b < num_bits (), "BUG");
	return nosign (b) / filter_width;
    }

    public int col (int b) {
	if (DEBUG) Assert.assert (b < num_bits (), "BUG");
	return nosign (b) % filter_width;
    }

    public boolean clear (int b) {
	if (DEBUG) Assert.assert (b < num_bits (), "BUG");
	return ((((Integer) _bits.get (b)).intValue ()) & 0x80000000) == 0;
    }

    public Object clone () throws CloneNotSupportedException {
	UpdateMessage result = (UpdateMessage) super.clone ();
	result.id = id;
	result.local_seq = local_seq;
	return result;
    }

    public String toString () {
	String result = "(UpdateMessage " + id + " " + 
	    (inbound ? "from " : "to ") + peer.port () + " " + local_seq;
	for (int i = 0; i < num_bits (); ++i) 
	    result += " " + (clear (i) ? "-" : "+") + "(" + row (i) + "," +
		col (i) + ")";
	result += ")";
	return result;
    }

    private int nosign (int b) {
	return (((Integer) _bits.get (b)).intValue ()) & 0x7fffffff;
    }

    private Vector _bits;
}

