/**
 * MACedQS.
 *
 * @author     Gifford Cheung 
 * @version    $Id: MACedQS.java,v 1.4 2001/10/26 23:51:29 srhea Exp $
 */

package ostore.mac;
import ostore.util.*;
import java.security.*;

/**
 * MACedQS<br></br>
 * <i>Pronounced: "mack'd kue ess"</i><br></br>
 * This class contains two major items: <br>(1) A QuickSerializable
 * object and <br>(2) the MAC of (1)
 * <p></p>
 * <b>How to use:</b><br></br> 
 * (1) Construct a new <tt>MACedQS</tt>: This will automatically generate the
 * byte encoding of the <tt>user_data</tt>.<br></br>
 * (2) Pass this <tt>MACedQS</tt> object to a {@link MacHandler} using
 * {@link MacHandler#doMAC} to generate the MAC. (<i>Internal Note: 
 * What the <tt>MacHandler</tt> typically does is grab the {@link #getBytes} 
 * compute the MAC and, using {@link #setMAC}, sets the MAC.)</i>
 * <br></br>
 * (3) To verify: Pass this <tt>MACedQS</tt> object to a <tt>MacHandler</tt>
 * through {@link MacHandler#authenticate} <br></br>
 * <b>What you cannot do:</b> <br></br>
 * + You cannot alter the user_data and user_bytes contents once this
 * object has been constructed.<br></br>
 * 
 * @author   Gifford Cheung 
 * @version    $Id: MACedQS.java,v 1.4 2001/10/26 23:51:29 srhea Exp $
 */
public class MACedQS implements QuickSerializable {

    /**
     * @param user_data the message to MAC 
     */
     public MACedQS ( QuickSerializable user_data ) {

      // copy over values
      _user_data = user_data;

      // create _bytes from _user_data 
      int[] offset = {0};
      to_bytes_helper(null, offset);
      _bytes = new byte[offset[0]];
      offset[0] = 0;
      to_bytes_helper(_bytes, offset);
    }

    /**
     * The deserialization constructor.  See the comments at the top of
     * {@link ostore.util.QuickSerializable} for details.
     */
    public MACedQS (byte [] data, int [] offset) throws QSException {

      int starting_offset = offset [0];
      // user data
      int user_data_type = ByteUtils.bytesToInt (data, offset);
      _user_data = TypeTable.new_object (user_data_type, data, offset);

      int blength = offset[0] - starting_offset;
      // _bytes (from the same bytes as user data)
      _bytes = new byte[blength];
      ByteUtils.memcpy(_bytes, 0, data, starting_offset, blength);

      // MAC
      int MAClen = ByteUtils.bytesToInt (data, offset);
      if (MAClen != 0) {
         _MAC = new byte [MAClen];
         offset [0] += ByteUtils.memcpy (_MAC, 0, data, 
                                      offset [0], MAClen);
      } else {
         _MAC = null;
      }
    }

    /**
     * The data which this object MAC'd
     *
     * @return the data which is MAC'd 
     */
    public QuickSerializable user_data () { return _user_data; }

    /**
     * The bytes of the data MAC'd
     *
     * @return the to_bytes of user_data
     */
    public byte[] getBytes() { return _bytes; }

    /**
     * The MAC of the user_data
     * 
     * @return the bytes of the (precomputed) MAC of the data
     */
    public byte[] getMAC() { return _MAC; }

    /**
     * Sets the MAC value for user_data, meant for MacHandler to use.
     */
    protected void setMAC(byte[] MAC) {
      //TODO  perhaps you want to set this to a one way
      //      function where, once the _MAC is no longer null,
      //      you will not allow anymore changes...
       _MAC = MAC;
    }

    public void to_bytes (byte [] data, int [] offset) {
      // _user_data's _bytes
      offset [0] += ByteUtils.memcpy (data, offset [0], _bytes, 0,
                              _bytes.length);
   
      // _MAC
      // follows the same principle of SignedQS in assuming that
      // receivers may not know the MAC type, and need to have the
      // length of the rest of the message to skip ahead to the next
      // message in line.
      if (_MAC == null) {
         ByteUtils.intToBytes(0, data, offset);
      } else {
         ByteUtils.intToBytes (_MAC.length, data, offset);
         offset [0] += ByteUtils.memcpy (data, offset [0], 
                     _MAC, 0, _MAC.length);
      }
    }

    public int type_code () {
        return Types.MAC_MACEDQS;
    }

    private void to_bytes_helper (byte [] data, int [] offset) {
        ByteUtils.intToBytes (_user_data.type_code (), data, offset);
        _user_data.to_bytes (data, offset);
    }

    private QuickSerializable _user_data;
    private byte [] _bytes;
    private byte [] _MAC;

}
