/**
 * Model.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.QuickSerializable;
import mdw.sandStorm.api.QueueElementIF;

/**
 * An interface for introspective models.
 * A <code>Model</code> can observe <code>QuickSerializable</code>
 * objects, estimating their likelihood based on an internal learning model
 * and updating that model based on those observations.
 * <code>Prediction</code>s may be estimates of the current state of a
 * simple model, or a confidence-weighted guess for future observations,
 * depending on the particular <code>Model</code>. 
 *
 * @see SegmentedModel
 * @see HMM
 *
 * @author	Dennis Geels
 * @version	$Id: Model.java,v 1.7 2001/11/28 23:03:15 geels Exp $
 */

public interface Model extends QuickSerializable
{
    //////////////////// 	METHODS		////////////////////

    /**
     * Notes an observation for later processing.
     * This method should execute very quickly, so that OIL does not create
     * a bottleneck in a critical path.  Any significant processing should
     * be postponed until the <code>recalculate</code> method is called.
     *
     * @param observation	Any event, value, etc. that this
     * 				<code>Model</code> understands.
     * @exception IllegalArgumentException if the particular subclass of
     * <code>Model</code> cannot process the <code>observation</code>.
     */
    void record( QuickSerializable observation );

    /**
     * Notes a group of observations for later processing.  This method
     * allows the <code>Model</code> to process a group of observations in
     * bulk, potentially saving resources.
     *
     * @param observations	An array of events, values, etc., not
     * 				necessarily all the same type, that this
     * 				<code>Model</code> understands. 
     * @exception IllegalArgumentException if the particular subclass of
     * <code>Model</code> cannot process the <code>observation</code>.
     */
    void record( QuickSerializable[] observations );

    /**
     * Forget the least-recent <code>record</code>ed observations.
     */
    void clear( int num );
    
    /**
     * Forget all observations <code>record</code>ed since the last call to
     * <code>clear</code> or <code>recalculate</code>.
     */
    void clear();
    
    /**
     * Calculates the log-likelihood of all <code>record</code>ed
     * observations.  Only observations recorded since the last call to
     * <code>clear</code> or <code>recalculate</code> are considered.
     *
     * @return a non-positive number equal to the logarithm of the
     * probability of the observations (<code>D</code>) given the current
     * model (<code>M</code>); <code>log(p(D|M))</code>.  If no
     * observations exist, it returns <code>Double.NaN</code>.
     */
    double loglikelihood();

    /**
     * Process recently <code>record</code>ed observations.  The internal
     * model parameters are updated to produce the model which maximizes
     * the likelihood of the observations.  Observations are forgotten
     * after processing, as if <code>clear</code> were called.
     * Different <code>Model</code>s may provide different semantics for
     * how prior parameters are included in the resulting ones.  For
     * instance, sufficient statistics may be added cumulatively,
     * parameters may have finite moments, etc.
     *
     * @return A <code>Delta</code> storing the difference between the new
     * 		model parameters and the old.  */
    Delta recalculate();

    /**
     * Make a prediction based on current model parameters and
     * observations.
     *
     * @return A <code>Prediction</code>, whose semantics depends on the
     * particular subclass of <code>Model</code>.
     */
    Prediction predict( int horizon );

    /**
     * Returns the recent observations which the model assigns the lowest
     * likelihood.  This method is useful both for standard outlier
     * detections as well as for swapping in model segments using reverse
     * lookup.
     *
     * @return an array containing recent observations which the model
     * did not expect.
     */
    QuickSerializable[] outliers();
    
    /**
     * Incorporate the information from a <code>Delta</code> into this
     * <code>Model</code>.
     * The resulting <code>Model</code> should be equivalent to that which
     * produced the <code>Delta</code>, given that they began in the same
     * state.  Subclasses of <code>Model</code> employ specific subclasses
     * of <code>Delta</code>.
     *
     * @exception IllegalArgumentException if the <code>Delta</code> object
     * is not of the correct subclass. 
     */
    void add_delta( Delta d );

    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A summary of the knowledge learned by a <code>Model</code>. It
     * contains updates to the model parameters, usually in the form of
     * sufficient statistics deltas, sufficient to update or initialize a
     * different <code>Model</code>.  In general, each subclass of
     * <code>Model</code> has its own subclass of <code>Delta</code>.
     *
     * @author	Dennis Geels
     * @version	$Id: Model.java,v 1.7 2001/11/28 23:03:15 geels Exp $
     */
    interface Delta extends QueueElementIF, QuickSerializable
    {
	/**
	 * Produce the sum of this and another <code>Delta</code>.
	 *
	 * @param other	The other <code>Delta</code>.
	 * @return a new <code>Delta</code>, constructed such that applying
	 * it to a <code>Model</code> produces an equivalent
	 * <code>Model</code> to that produced by adding first
	 * <code>this</code>, then <code>other</code>.
	 * @exception IllegalArgumentException if <code>other</code> is not
	 * an instance of a compatible <code>Delta</code> subclass.
	 */
	Delta sum( Delta other );
    }
    
    /**
     * A summary of the current state predicted by the <code>Model</code>.
     * A <code>Prediction</code> may be an estimated average, the
     * probability of a future outlier, or a weighted list of possible
     * future states.
     *
     * @author	Dennis Geels
     * @version	$Id: Model.java,v 1.7 2001/11/28 23:03:15 geels Exp $
     */
    interface Prediction extends QueueElementIF, QuickSerializable {}
}
