/**
 * SegmentPager.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.QSIO;
import ostore.util.NodeId;
import ostore.util.Types;
import ostore.util.TypeTable;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InvalidConfigFileException;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.UnknownEventException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;

/**
 * A stage which fetches model segments to/from remote carriers.
 *
 * @author	Dennis Geels
 * @version	$Id: SegmentPager.java,v 1.6 2001/12/13 20:50:29 geels Exp $
 */

public class SegmentPager implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////

    private static final boolean DEBUG = DebugFlags.OIL_SEGMENTPAGER;
    private static final String class_tag = "SegmentPager";

    // A few strings used in the config files:
    private static final String IMAGE_STR 	= "snapshot";

    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The name by which to refer to this stage.  It should be
     * the name of the corresponding <code>StageIF</code>.
     */
    protected String name;

    /**
     * For now we store segments inside one huge <code>Model</code>.
     */
    protected Model model;
    
    private Classifier classifier;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	if( DEBUG ) Debug.turnOn( class_tag );
    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Processes a <code>Request</code>.
     * If this <code>SegmentPager</code> has a copy of the requested
     * <code>Segment</code>, it dispatches it.  If not, it should fetch the
     * <code>Segment</code> from a remote carrier (not yet implemented).
     */
    public void handle_request( Request req )
    {
	final String tag = class_tag+".handle_request";
	if( DEBUG ) Debug.printtagln( tag, name + " called" );

	// TODO:
	// Reimplement storage using cache, falling back to network.
	
	if( model instanceof HMM ) {
	    // Pull the segment from our local model.
	    HMM hmm = (HMM)model;

	    QuickSerializable state = req.id;
	    HMM.Segment seg = new HMM.Segment( state,
					       hmm.marginals.get(state),
					       hmm.transitions.get(state),
					       hmm.emissions.get(state) );

	    // don't worry about dropped segments.
	    classifier.dispatch_lossy( seg );
		
	} else {
	    throw new UnsupportedOperationException();
	}
	

	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /**
     * Processes a <code>ReverseRequest</code>.
     * The <code>SegmentPager</code> scans its local <code>Segment</code>s
     * for any containing the specified entry.  If none are found, it
     * should forward the request to a remote carrier (not yet implemented).
     */
    public void handle_reverse_request( ReverseRequest req )
    {
	final String tag = class_tag+".handle_reverse_request";
	if( DEBUG ) Debug.printtagln( tag, name + " called" );

	// TODO:
	// Reimplement storage using cache, falling back to network.
	
	if( model instanceof HMM ) {
	    // Pull the segment from our local model.
	    HMM hmm = (HMM)model;

	    // FIXME: proper level?
	    double threshold = .001;

	    QuickSerializable output = req.id;
	    for( int i=0; i<hmm.states.size(); i++ ) {
		QuickSerializable state = hmm.states.get(i);

		// Check for significant emission probability.
		if( hmm.emissions.get( state ).get( output )
		    > threshold ) {
		    HMM.Segment seg =
			new HMM.Segment( state,
					 hmm.marginals.get(state),
					 hmm.transitions.get(state),
					 hmm.emissions.get(state) );

		    // don't worry about dropped segments.
		    classifier.dispatch_lossy( seg );
		}
	    }
	} else {
	    throw new UnsupportedOperationException();
	}
	

	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void handleEvent(QueueElementIF elem) throws UnknownEventException
    {
	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, name + " called:\t" + elem );

	if( elem instanceof Request ) {
	    handle_request( (Request)elem );
	} else if( elem instanceof ReverseRequest ) {
	    handle_reverse_request( (ReverseRequest)elem );
  	} else {
  	    throw new UnknownEventException( "Not a known Request", elem);
  	}

	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }
	
    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws UnknownEventException
    {
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, name + " called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
	if( DEBUG ) Debug.printtagln( tag, name + " done" );
    }

    /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void init(ConfigDataIF config) throws Exception
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	name = config.getStage().getName();

	////////////////////////////////
	// Initialize Filters
	
	classifier = Classifier.getClassifier(
                new NodeId (config.getString ("node_id")));

	SinkIF my_sink = config.getStage().getSink();

	Filter filter = new Filter();
	if( !filter.requireType( "ostore.oil.SegmentPager$Request" ) ||
	    !classifier.subscribe( filter, my_sink ) )
	    Carp.die( "Could not subscribe to Requests" );

	filter = new Filter();	
	if( !filter.requireType( "ostore.oil.SegmentPager$ReverseRequest" ) ||
	    !classifier.subscribe( filter, my_sink ) )
	    Carp.die( "Could not subscribe to ReverseRequests" );
	
	////////////////////////////////
	// Initialize model

	String image_name = config.getString( IMAGE_STR );
	model = null;
	if( image_name != null ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "reading model from " +
				  "snapshot: \"" + image_name + "\"" );
	    }
	    try {
		model = (Model)QSIO.read( image_name );
	    } catch( Exception e ) {
		if( DEBUG ) {
		    Carp.carp( "invalid model image" );
		    Carp.carp( e );
		}
		model = null;
	    }
	}
	if( model == null ) {
	    throw new InvalidConfigFileException( name, IMAGE_STR );
	}

	if( DEBUG ) Debug.printtagln( tag, "done: " + name );	
    }
    
     /* Specified by mdw.sandStorm.api.EventHandlerIF */
    public void destroy()
    {
	final String tag = class_tag+".destroy";
	if( DEBUG ) Debug.printtagln( tag, "destroyed: " + name  );
    }

   ////////////////////     INNER CLASSES	////////////////////
    
    public static class Request implements QueueElementIF, QuickSerializable
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The unique identifier for the requested <code>Segment</code>
	 */
	public QuickSerializable id;
	
	////////////////////    CONSTRUCTORS	////////////////////

	public Request( QuickSerializable id )
	{
	    this.id = id;
	}
	
	public Request( byte[] data, int[] offset ) throws QSException
	{
	    int type = ByteUtils.bytesToInt( data, offset );
	    id = TypeTable.new_object( type, data, offset );
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	public synchronized void to_bytes( byte[] data, int[] offset )
	{
	    ByteUtils.intToBytes( id.type_code(), data, offset );
	    id.to_bytes( data, offset );
	}

	public int type_code()
	{
	    return Types.SEGMENT_REQUEST;
	}
	////////////////////   INNER CLASSES	////////////////////
    }
    
    public static class ReverseRequest implements QueueElementIF,
						   QuickSerializable
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The unique identifier for some <code>Segment</code> entry.
	 */
	public QuickSerializable id;
	
	////////////////////    CONSTRUCTORS	////////////////////

	public ReverseRequest( QuickSerializable id )
	{
	    this.id = id;
	}
	
	public ReverseRequest( byte[] data, int[] offset ) throws QSException
	{
	    int type = ByteUtils.bytesToInt( data, offset );
	    id = TypeTable.new_object( type, data, offset );
	}
	
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////

	public synchronized void to_bytes( byte[] data, int[] offset )
	{
	    ByteUtils.intToBytes( id.type_code(), data, offset );
	    id.to_bytes( data, offset );
	}

	public int type_code()
	{
	    return Types.SEGMENT_REV_REQUEST;
	}
	////////////////////   INNER CLASSES	////////////////////
    }
    
}
