/**
 * TraceReader.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.oil;

import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.dispatch.Classifier;
import ostore.dispatch.Signal;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.SinkFullException;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.ConfigDataIF;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Reader;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.StreamTokenizer;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.lang.Runtime;
import ostore.util.InvalidConfigFileException;
import ostore.util.QSString;

/**
 * A SandStorm stage which reads trace files and generates events for each
 * trace entry.
 *
 * @author	Dennis Geels
 * @version	$Id: TraceReader.java,v 1.21 2001/11/30 18:26:18 geels Exp $
 */

public class TraceReader implements EventHandlerIF
{
    ////////////////////       CONSTANTS	////////////////////
    
    private final static boolean DEBUG = true;
    private final static boolean TRACE_DEBUG = false;
    private final static String tag = "TraceReader";
    
    private final static Integer REC_UID = new Integer(-3);
    private final static Integer REC_RESTART = new Integer(-2);
    private final static Integer REC_TIMESTAMP = new Integer(-1);

    private final static String STR_UID = "UID";
    private final static String STR_RESTART = "RESTART";
    private final static String STR_TIMESTAMP = "TIMESTAMP";

    private final static Integer OP_OPEN = new Integer(2);
    private final static Integer OP_CREAT = new Integer(3);
    private final static Integer OP_EXECV = new Integer(4);
    private final static Integer OP_EXECVE = new Integer(5);
    private final static Integer OP_MKNOD = new Integer(6);
    private final static Integer OP_STAT = new Integer(7);
    private final static Integer OP_LSTAT = new Integer(8);
    private final static Integer OP_OLDSTAT = new Integer(9);
    private final static Integer OP_OLDLSTAT = new Integer(10);
    private final static Integer OP_UTIME = new Integer(11);
    private final static Integer OP_UTIMES = new Integer(12);
    private final static Integer OP_ACCESS = new Integer(13);
    private final static Integer OP_READLINK = new Integer(14);
    private final static Integer OP_FCNTL = new Integer(15);
    private final static Integer OP_TRUNCATE = new Integer(16);
    private final static Integer OP_CHDIR = new Integer(17);
    private final static Integer OP_RENAME = new Integer(18);
    private final static Integer OP_EXIT = new Integer(19);
    private final static Integer OP_UNLINK = new Integer(20);
    private final static Integer OP_CLOSE = new Integer(21);
    private final static Integer OP_FORK = new Integer(22);
    private final static Integer OP_DUP = new Integer(23);
    private final static Integer OP_DUP2 = new Integer(24);
    private final static Integer OP_LINK = new Integer(25);
    private final static Integer OP_RMDIR = new Integer(26);
    private final static Integer OP_FCHDIR = new Integer(27);    
    private final static Integer OP_MKDIR = new Integer(28);    
    private final static Integer OP_SYMLINK = new Integer(29);    

    private final static String STR_OPEN = "open";
    private final static String STR_CREAT = "creat";
    private final static String STR_EXECV = "execv";
    private final static String STR_EXECVE = "execve";
    private final static String STR_MKNOD = "mknod";
    private final static String STR_STAT = "stat";
    private final static String STR_LSTAT = "lstat";
    private final static String STR_OLDSTAT = "oldstat";
    private final static String STR_OLDLSTAT = "oldlstat";
    private final static String STR_UTIME = "utime";
    private final static String STR_UTIMES = "utimes";
    private final static String STR_ACCESS = "access";
    private final static String STR_READLINK = "readlink";
    private final static String STR_FCNTL = "fcntl";
    private final static String STR_TRUNCATE = "truncate";
    private final static String STR_CHDIR = "chdir";
    private final static String STR_RENAME = "rename";
    private final static String STR_EXIT = "exit";
    private final static String STR_UNLINK = "unlink";
    private final static String STR_CLOSE = "close";
    private final static String STR_FORK = "fork";
    private final static String STR_DUP = "dup";
    private final static String STR_DUP2 = "dup2";
    private final static String STR_LINK = "link";
    private final static String STR_RMDIR = "rmdir";
    private final static String STR_FCHDIR = "fchdir";
    private final static String STR_MKDIR = "mkdir";
    private final static String STR_SYMLINK = "symlink";
//      private final static String STR_ = "";
    

    /**
     * The current working directory for processes whose real cwd we do not
     * know.
     */
    private final static String DEFAULT_CWD = "/NOCLUE/";
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The stage name of the ClusterDetector we work with.
     */
    private String cd_name;

    /**
     * The name of the traces to read.
     */
    private String[] traces;

    /**
     * The name of the trace to test (if any).
     */
    private String test_trace;
    
    /**
     * Maps strings to record types.
     */
    private Hashtable record_types;

    /**
     * Maps strings to syscall opcodes.
     */
    private Hashtable opcodes;

    /**
     * Maps process ID's to the process' current working directory.
     */
    private Hashtable cwds;

    /**
     * Tracks opcodes processed.
     */
    private Hashtable opcodes_seen;

    /**
     * Tracks filenames processed.
     */
    private Hashtable files_seen;

    private NodeId node_id;
    private Classifier classifier;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	if( DEBUG ) {
	    Debug.turnOn( tag+".init" );
	    Debug.turnOn( tag+".parse_line" );
  	    Debug.turnOn( tag+".read_trace" );
//  	    Debug.turnOn( tag+".open_file" );
//  	    Debug.turnOn( tag+".read_list" );
//  	    Debug.turnOn( tag+".print_summary" );
	    Debug.turnOn( tag+".handleEvent" );

  	    Debug.turnOn( "HMMBuilder" );
  	    Debug.turnOn( "HMMPredictor" );
//  	    Debug.turnOn( "HMM" );
    	    Debug.turnOn( "HMM.EM" );
    	    Debug.turnOn( "HMM.log_likelihood" );
    	    Debug.turnOn( "HMM.time" );
    	    Debug.turnOn( "HMM.record" );
//      	    Debug.turnOn( "HMM.M_step" );
      	    Debug.turnOn( "HMM.E_step" );
    	    Debug.turnOn( "HMM.add_delta" );
    	    Debug.turnOn( "HMM.predict" );
    	    Debug.turnOn( "HMM.sync_arrays" );
	}
	if( TRACE_DEBUG ) {
	    Debug.turnOn( tag );
	}
    }

    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by EventHandlerIF
     */
    public void handleEvent(QueueElementIF elem) throws EventHandlerException
    {
	// Assume that we caught the SIGINIT.
	if( DEBUG ) Debug.printtagln( tag+".handleEvent", "Caught " + elem +
				      " Starting trace..." );

	StreamTokenizer st;
	for( int i=0; i<traces.length; i++ ) {
	    try {
		if( DEBUG ) Debug.printtagln( tag+".handleEvent",
					      "Opening trace \"" + traces[i] +
					      "\"" );
		st = open_file( traces[i] );
		read_trace ( st );
	    } catch( IOException ioe ) {
		Carp.die( ioe );
	    }
	}
	if( DEBUG ) Debug.printtagln( tag+".handleEvent",
				      "All traces complete" );
	// Send one last event to the ClusterDetector
	while( true ) {
	    try {
		Signal.getSignal(node_id).signal( Signal.RBLD, cd_name );
		break;
	    } catch( SinkFullException se ) {
		try {
		    if( DEBUG ) Debug.printtagln( tag+".handleEvent",
						  "stalled on final event." );
		    Thread.currentThread().sleep( 5000 );
		} catch ( InterruptedException ie ) {
		    // Ignore.
		}
	    }
	}
	if( DEBUG ) Debug.printtagln( tag+".handleEvent", "traces complete." );

	// Print out opcodes and files seen, then stop.
	print_summary();
    }

    /**
     * Specified by EventHandlerIF
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i] );
	}
    }

    /**
     * Specified by EventHandlerIF
     */
    public void init(ConfigDataIF config) throws Exception
    {
	node_id = new NodeId (config.getString ("node_id"));
	classifier = Classifier.getClassifier (node_id);

	cd_name = config.getString( "detector" );
	if( cd_name == null ) {
	    throw new InvalidConfigFileException( tag, "detector" );
	}

	String type = config.getString( "type" );
	if( type == null ) {
	    throw new InvalidConfigFileException( tag, "type" );
	}
	if( type.equals( "single" ) || type.equals( "test" )) {
	    traces = config.getStringList( "trace" );
	    if( traces == null ) {
		throw new InvalidConfigFileException( tag, "trace" );
	    }
	    
	} else if( type.equals( "batch" ) ){
	    traces = config.getStringList( "tracelist" );
	    if( traces == null ) {
		throw new InvalidConfigFileException( tag, "tracelist" );
	    }

	} else {
	    throw new InvalidConfigFileException( tag, "type" );
	}

	record_types = new Hashtable();
	record_types.put( STR_UID, REC_UID );
	record_types.put( STR_RESTART, REC_RESTART );
	record_types.put( STR_TIMESTAMP, REC_TIMESTAMP );

	opcodes = new Hashtable();
	opcodes.put( STR_OPEN, OP_OPEN );
	opcodes.put( STR_CREAT, OP_CREAT );
	opcodes.put( STR_EXECV, OP_EXECV );
	opcodes.put( STR_EXECVE, OP_EXECVE );
	opcodes.put( STR_MKNOD, OP_MKNOD );
	opcodes.put( STR_STAT, OP_STAT );
	opcodes.put( STR_LSTAT, OP_LSTAT );
	opcodes.put( STR_OLDSTAT, OP_OLDSTAT );
	opcodes.put( STR_OLDLSTAT, OP_OLDLSTAT );
	opcodes.put( STR_UTIME, OP_UTIME );
	opcodes.put( STR_UTIMES, OP_UTIMES );
	opcodes.put( STR_ACCESS, OP_ACCESS );
	opcodes.put( STR_READLINK, OP_READLINK );
	opcodes.put( STR_FCNTL, OP_FCNTL );
	opcodes.put( STR_TRUNCATE, OP_TRUNCATE );
	opcodes.put( STR_CHDIR, OP_CHDIR );
	opcodes.put( STR_RENAME, OP_RENAME );
	opcodes.put( STR_EXIT, OP_EXIT );
	opcodes.put( STR_UNLINK, OP_UNLINK );
	opcodes.put( STR_CLOSE, OP_CLOSE );
	opcodes.put( STR_FORK, OP_FORK );
	opcodes.put( STR_DUP, OP_DUP );
	opcodes.put( STR_DUP2, OP_DUP2 );
	opcodes.put( STR_LINK, OP_LINK );
	opcodes.put( STR_RMDIR, OP_RMDIR );
	opcodes.put( STR_FCHDIR, OP_FCHDIR );
	opcodes.put( STR_MKDIR, OP_MKDIR );
	opcodes.put( STR_SYMLINK, OP_SYMLINK );

	cwds = new Hashtable();

	opcodes_seen = new Hashtable();

	files_seen = new Hashtable();
	
	// Now wait five seconds, then start reading the trace.
	String my_name = config.getStage().getName();
	SinkIF sink = config.getStage().getSink();
	if( !Signal.getSignal(node_id).register(Signal.INIT, my_name, sink)) {
	    throw new Exception( "signals failed" );
	}
	
	if( DEBUG ) Debug.printtagln( tag+".init", "TraceReader initialized" );
    }

    /**
     * Specified by EventHandlerIF
     */
    public void destroy() throws Exception
    {
	Debug.printtagln( tag+".destroy()", "called" );
	Debug.printtagln( tag+".destroy()", "done" );
    }

    /**
     * Initialize a <code>StreamTokenizer</code> for a specified trace file.
     */
    private StreamTokenizer open_file( String filename ) throws IOException
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".open_file", "called" );
	Reader r;
	if( ! filename.endsWith( ".gz" ) ) {
	    Debug.printtagln( tag+".init", "Opening trace file \"" +
			      filename + "\" for reading..." );
	    r = new FileReader(filename);
	} else {
	    Debug.printtagln( tag+".init", "Opening pipe from dumpobs "
			      + "to read \"" + filename + "\"..." );
	    Runtime rt = java.lang.Runtime.getRuntime();
	    Process gzip = rt.exec("dumpobs " + filename);
	    r = new InputStreamReader(gzip.getInputStream());
	}
	BufferedReader br = new BufferedReader(r);
	StreamTokenizer trace = new StreamTokenizer( br );
	
	// adjust syntax table
	trace.eolIsSignificant( true );
	trace.ordinaryChars( (int)'.', (int)'.' );
	trace.ordinaryChars( (int)'/', (int)'/' );
	trace.ordinaryChars( (int)'0', (int)'9' );
	trace.wordChars( (int)'?', (int)'?' );
	trace.wordChars( (int)'0', (int)'9' );
	trace.wordChars( (int)'/', (int)'/' );

	if( TRACE_DEBUG ) Debug.printtagln( tag+".open_file", "done" );
	return trace;
    }

    /**
     * Read a list of strings (trace names) from a file, one per line.
     */
    private String[] read_list( String filename ) throws IOException
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".read_list", "called" );
	BufferedReader br = new BufferedReader( new FileReader( filename ));
	Vector names = new Vector();
	String name;
	while( br.ready() ) {
	    name = br.readLine();
	    if( DEBUG ) Debug.printtagln( tag+".read_list",
					  "Found trace name \""+ name +"\"");
	    names.add(name);
	}
	int len = names.size();
	if( DEBUG ) Debug.printtagln( tag+".read_list", "Found " + len +
				      "traces" );
	String[] final_list = new String[len];
	for( int i=0; i<len; i++ ) {
	    final_list[i] = (String)names.get(i);
	}
	if( TRACE_DEBUG ) Debug.printtagln( tag+".read_list", "done" );
	return final_list;
    }

    /**
     * Iterate through a trace file, dispatching each open call.
     */
    private void read_trace( StreamTokenizer trace )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".read_trace", "called" );
	FileOp fop = parse_line( trace );
	while( fop != null ) {
	    TestFileEvent event = new TestFileEvent();
	    event.filename = new QSString(fop.filename);
	    event.timestamp = fop.timestamp;
	    while( true ) {
		try {
		    if( DEBUG ) Debug.printtagln( tag+".read_trace",
						  "dispatched " + event );
		    classifier.dispatch( event );
		    break;
		} catch( SinkFullException se ) {
		    // TODO: We need some sort of asynchronous notification that
		    // the SinkIF has become unclogged.
		    if( DEBUG ) Debug.printtagln( tag+".read_trace",
						  "waiting..." );
		    try {
			// Just poll every 1 seconds.
			// FIXME!! set an alarm instead.
			if( TRACE_DEBUG ) Debug.printtagln(tag+".read_trace",
					     "waiting..." );
			Thread.currentThread().sleep( 1000 );
		    } catch ( InterruptedException ie ) {
			// Ignore.
		    }
		}
	    }
	    fop = parse_line(trace);
	}
	if( TRACE_DEBUG ) Debug.printtagln( tag+".read_trace", "done" );
    }

    /**
     * Print out a few useful bits of information about the trace.
     */
    private void print_summary()
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".print_summary", "called" );
	Enumeration op_strings = opcodes.keys();
	String op_string;
	while( op_strings.hasMoreElements() ) {
	    op_string = (String)op_strings.nextElement();
	    if( opcodes_seen.containsKey(opcodes.get(op_string)) ) {
		Debug.printtagln( tag+".print_summary", "Saw syscall: " +
				  op_string );
	    }
	}
	Enumeration filenames = files_seen.keys();
	while( filenames.hasMoreElements() ) {
	    Debug.printtagln( tag+".print_summary", "Saw File: " +
			      filenames.nextElement() );
	}
	if( TRACE_DEBUG ) Debug.printtagln( tag+".print_summary", "done" );
    }

    
    /**
     * Convert the next line of the trace into a <code>FileOp</code> object.
     *
     * @param st	The stream containing the token.
     * @return a <code>FileOp</code> object representing the next line of the
     * trace that contins a real file operation (skipping other lines).  If
     * the trace file contains no more operations, <code>parse_line</code>
     * returns <code>null</code>.
     */
    private FileOp parse_line( StreamTokenizer st )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_line", "called" );
	FileOp fop = null;
	int ttype = 0;
	
	long rec_num;	// ignored
	Integer record_type;	// ignored unless equal to REC_UID
	long uid;	// ignored
	String PID;	// ignored
	Integer pid;	// ignored
	String progname;	// ignored
	String flag;	// ignored
	long seconds;	
	char period;	// ignored
	long microseconds;	// ignored
	Integer opcode;	// ignored unless equal to OP_OPEN
	char paren;
	String filename = null;	

    parse_loop:
	while( ttype != StreamTokenizer.TT_EOF ) {

	parse: {
	    // Step 1: parse record number
	    rec_num = parse_long( st );
	    if( rec_num == Long.MIN_VALUE ) break parse;

	    // Step 2: parse record type
	    record_type = parse_word( st, record_types );
	    if( record_type == null ) {
		break parse;
	    } else if ( record_type == REC_RESTART ) {
		break parse;
	    } else if ( record_type == REC_TIMESTAMP ) {
		break parse;
	    } // else record_type == REC_UID.  Keep parsing.

	    // Step 3: parse uid
	    uid = parse_long( st );
	    if( uid == Long.MIN_VALUE ) break parse;

	    // Step 4: parse 'PID'
	    PID = lex_word( st );
	    if( (PID == null) || !(PID.equals("PID")) ) break parse;

	    // Step 5: parse pid
	    long long_pid = parse_long( st );
	    if( long_pid == Long.MIN_VALUE ) {
		break parse;
	    } else if( long_pid > Integer.MAX_VALUE ) {
		Debug.printtagln( tag+".parse_line", "Huge PID: " + long_pid);
	    }
	    pid = new Integer((int)long_pid);
	    
	    // Step 6: parse program name
	    progname = lex_word( st );
	    if( progname == null ) break parse;

	    // Step 7: parse flag
	    flag = lex_word( st );
	    if( flag == null ) break parse;

	    // Step 8: parse timestamp
	    seconds = parse_long( st );
	    if( seconds == Long.MIN_VALUE ) break parse;
	    period = lex_char( st );
	    if( period != (char)'.' ) break parse;
	    microseconds = parse_long( st );
	    if( microseconds == Long.MIN_VALUE ) break parse;

	    // Step 9: parse syscall
	    opcode = parse_word( st, opcodes );
	    if( opcode == null ){
		Debug.printtagln( tag+".parse_line", "Unknown Syscall: " +
				  st.sval );
		break parse;
	    }

	    // Step 9.1: remember syscalls encountered
	    opcodes_seen.put( opcode, "dummy" );

	    // Step 10: skip '('
	    paren = lex_char( st );
	    if( paren != (char)'(' ) break parse;

	    // Step 11: parse syscall args
	    if( opcode == OP_OPEN ) {
		// parse filename
		filename = lex_word( st );
		filename = canonical_form( filename, pid );

		// Now fill in the FileOp and return.
		fop = new FileOp();
		fop.op = opcode.intValue();
		fop.timestamp = seconds;	// don't worry about micros.
		fop.filename = filename;
		break parse_loop;
	    
	    } else if( opcode == OP_CHDIR ) {
		filename = lex_word( st );
		filename = canonical_form( filename, pid );

		// Store this pathname (slash-terminated) as the new cwd.
		// NOTE: "/" will end up doubled: "//".  That's alright.
		cwds.put( pid, filename + "/" );

	    } else if( opcode == OP_FORK ) {
		// copy the cwd from the forked process.
		long_pid = parse_long( st );
		filename = (String)cwds.get( new Integer((int)long_pid) );
		if( filename == null ) {
		    Debug.printtagln( tag+".parse_line",
				      "Missing cwd for forked id: " +long_pid);
		} else {
		    cwds.put( pid, filename );
		}
	    }

	    // Step X: skip rest of line for now.
	    break parse;


	} // end parse block
	// This line must have been malformed.  Skip it and continue the loop.
	ttype = next_line( st );
	
	} // end parse_loop

	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_line", "done" );
	return fop;
    }

    /**
     * Advance the stream to the start of the next line.
     *
     * @param st	The stream containing the token.
     * @return <code>StreamTokenizer.TT_EOL</code>, or
     * <code>StreamTokenizer.TT_EOF</code> if the end of the stream was
     * reached. 
     */
    private int next_line( StreamTokenizer st )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".next_line",
				      "Skipping to end of line" );
	int ttype = 0;
	while( (ttype != StreamTokenizer.TT_EOL) &&
	       (ttype != StreamTokenizer.TT_EOF) ) {
	    try {
		ttype = st.nextToken();
	    } catch ( java.io.IOException ioe ) {
		Carp.error( "next_line: Couldn't read Token at line #" +
			    st.lineno() );
		break;
	    }
	    if( DEBUG ) {
		switch( ttype ) {
		case (int)'\"':
		case StreamTokenizer.TT_WORD:
		    Debug.printtagln( tag+".next_line", "WORD: " + st.sval );
		    break;
		case StreamTokenizer.TT_NUMBER:
		    Debug.printtagln( tag+".next_line", "NUMBER: " + st.nval );
		    break;
		case StreamTokenizer.TT_EOL:
		    Debug.printtagln( tag+".next_line", "EOL" );		    
		    break;
		case StreamTokenizer.TT_EOF:
		    Debug.printtagln( tag+".next_line", "EOF" );		    
		    break;
		default:
		    Debug.printtagln( tag+".next_line", "CHAR: " + (char)ttype );
		    break;
		}
	    }
	}
	if( TRACE_DEBUG ) Debug.printtagln( tag+".next_line", "done" );
	return ttype;
    }

    /**
     * Convert a filename into an absolute path.  If the filename does not
     * already begin with a leading slash ('/'), the current working directory
     * is prepended to the filename.  Then, as much redundancy is removed from
     * the path as possible (extra slashes, paths through the parent directory
     * (".."), etc.).
     *
     * @param local_name	The filename provided by a trace record.
     * @param pid		The process operating on the file.
     * @return The full absolute path to the file, including the filename.
     */

    private String canonical_form( String local_name, Integer pid )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".canonical_form",
					    "called: " + local_name );
	String name;
	
	if( local_name.startsWith( "/" ) ) {
	    name = local_name;
	} else {
	    name = (String)cwds.get( pid );
	    if( name == null ) {
		// FIXME!! Combine pid info here, so that we don't get
		// false collisions?
		name = DEFAULT_CWD;
	    }
	    name += local_name;
	}
	
	// Now remove any "//", "/./", or "/../" elements from the path.
	int slash1 = 0;		// iterates across the slashes in a pathname.
	int slash2 = -1;	// subsequent slash after slash1.

	if( !name.endsWith( "/" ) ) {
	    // Append a slash to help detect the redundant elements.
	    name += "/";
	}
	
	// Remove all "//" from the path.
	slash1 = name.indexOf( "//" );
	while( slash1 != -1 ) {
	    // We found a match; drop the first character of the "//".
	    slash2 = slash1 + 1;
	    name = name.substring(0,slash1) +
		name.substring(slash2,name.length());
	    
	    slash1 = name.indexOf( "//", slash1 );
	}

	// Remove all "/./" from the path.
	slash1 = name.indexOf( "/./" );
	while( slash1 != -1 ) {
	    // We found a match; drop the first two characters of the "/./".
	    slash2 = slash1 + 2;
	    name = name.substring(0,slash1) +
		name.substring(slash2,name.length());
	    
	    slash1 = name.indexOf( "/./", slash1 );
	}

	// Remove all '/*/../' from the path.
	slash2 = name.indexOf( "/../" );
	while( slash2 != -1 ) {
	    // We found a match; delete back to the previous '/'.
	    slash1 = name.lastIndexOf( '/', (slash2 - 1) );
	    if( slash1 == -1 ) {
		Debug.printtagln( tag+".canonical_form", 
				  "/../ has occured without a preceding /");
		slash1 = 0;
	    }
	    slash2 += 3;	// end of "/../".
	    name = name.substring(0,slash1) +
		name.substring(slash2,name.length());

	    slash2 = name.indexOf( "/../", slash1 );
	}

	// Finally, remove trailing slash (except for "/").
	if( name.endsWith( "/" ) && ! name.equals("/") ) {
	    name = name.substring(0,(name.length()-1));
	}
	
	// Remember files encountered
	files_seen.put( name, name );

	if( TRACE_DEBUG ) Debug.printtagln( tag+".canonical_form",
					    "done: " + name );
	return name;
    }
    
    /**
     * Convert the next token from a stream using a hashtable.
     *
     * @param st	The stream containing the token.
     * @param lookup	A <code>Hashtable</code> containing the
     * <code>String</code> to <code>Integer</code> mapping.
     * @return 	the corresponding <code>Integer</code> or
     * <code>null</code> if there was no match or any error
     * occurred. 
     */
    private Integer parse_word( StreamTokenizer st, Hashtable lookup )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_word", "called" );
	
	String word = null;
	Integer found = null;

	word = lex_word( st );
	if( word != null ) {
	    // search the Hashtable - (found = null) if no match.
	    if( DEBUG ) Debug.printtagln( tag+".parse_word",
					  "Looking up token: " + word );
	    found = (Integer)lookup.get( word );
	    if( found == null ) {
		if( DEBUG ) Debug.printtagln( tag+".parse_word",
					      "Could not find word " + word );
	    } else {
		if( DEBUG ) Debug.printtagln( tag+".parse_word",
					      "Found " + found );
	    }
	}
	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_word", "done" );
	return found;
    }

    /**
     * Fetch the next <code>long</code> from a stream.
     * 
     * @param st	The <code>StreamTokenizer</code> over the stream.
     * @return	The <code>long</code>, or <code>Long.MIN_VALUE</code> if any
     * errors occurred. Note that, due to the return code, we cannot parse the
     * smallest possible <code>long</code>.  This limitation shouldn't be a
     * problem.
     */
    private long parse_long( StreamTokenizer st )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_long", "called" );
	
	String word = null;
	long found = Long.MIN_VALUE;
	long num;

    parse: {
	    word = lex_word( st );
	    if( word != null ) {
		if( DEBUG ) Debug.printtagln( tag+".parse_long",
					      "Found word " + word );
		try {
		    num = Long.parseLong( word );
		} catch ( NumberFormatException not_long ) {
		    Debug.printtagln( tag+".parse_long",
				      "Found non-long " + word );
		    break parse;
		}
		if( num == Long.MIN_VALUE ) {
		    Debug.printtagln( tag+".parse_long",
				      "Found Long.MIN_VALUE" );
		    break parse;
		}		
	    } else {
		Debug.printtagln( tag+".parse_long", "Found non-word" );
		break parse;
	    }
	    found = num;
	} // end parse block
	if( TRACE_DEBUG ) Debug.printtagln( tag+".parse_long", "done: " + found );
	return found;
    }

    /**
     * Fetch the next <code>String</code> from a stream.
     *
     * @param st	The <code>StreamTokenizer</code> over the stream.
     * @return	The <code>String</code>, or <code>null</code> if the next
     * token was not a word or any errors occurred.
     */

    private String lex_word( StreamTokenizer st )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".lex_word",  "called" );
	
	int ttype = 0;
	String word = null;
	
    lex: {
	    try {
		ttype = st.nextToken();
	    } catch ( java.io.IOException ioe ) {
		Carp.error( "lex_word: Couldn't read Token at line #"
			    + st.lineno() );
		break lex;
	    }
	    if( (ttype != StreamTokenizer.TT_WORD) &&
		(ttype != (int)'\"')) {
		if( DEBUG ) Debug.printtagln( tag+".lex_word",
					      "Found non-word " + st );
		break lex;
	    }
	    word = st.sval;
	} // end lex block

	if( TRACE_DEBUG ) Debug.printtagln( tag+".lex_word", "done: " + word );
	return word;
    }

    /**
     * Fetch the next character from a stream.
     *
     * @param st	The <code>StreamTokenizer</code> over the stream.
     * @return the <code>char</code> read, or <code>0</code> if the next
     * token was not a character or any errors occurred.  Note that, due to
     * the return code, we cannot lex the NULL char.  This limitation
     * shouldn't be a problem.
     */

    private char lex_char( StreamTokenizer st )
    {
	if( TRACE_DEBUG ) Debug.printtagln( tag+".lex_char", "called" );
	
	int ttype = 0;
	char c = 0;
	
    lex: {
	    try {
		ttype = st.nextToken();
	    } catch ( java.io.IOException ioe ) {
		Carp.error( "lex_char: Couldn't read Token at line #" +
			    st.lineno() );
		break lex;
	    }
	    if( (ttype == StreamTokenizer.TT_WORD)
		|| (ttype == StreamTokenizer.TT_NUMBER)
		|| (ttype == StreamTokenizer.TT_EOL)
		|| (ttype == StreamTokenizer.TT_EOF) ) {
		if( DEBUG ) Debug.printtagln( tag+".lex_char",
					      "Found non-char " + st );
		break lex;
	    }
	    // ttype == the character itself.
	    if( ttype == 0 ) {
		Debug.printtagln( tag+".lex_char", "Found (char)0" );
		break lex;
	    }		
	    c = (char)ttype;
	} // end lex block
	
	if( TRACE_DEBUG ) Debug.printtagln( tag+".lex_char", "done: " + c );
	return c;
    }

    
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A simple class to collect the useful portions of a trace line.
     */
    private class FileOp
    {
	////////////////////     CONSTANTS	////////////////////
	////////////////////  STATIC VARIABLES 	////////////////////
	////////////////////     VARIABLES	////////////////////

	/**
	 * The system call number for this operation.
	 */
	public int op;

	/**
	 * The UNIX timestamp for this operation.
	 */
	public long timestamp;
	
	/**
	 * The (hopefully unique) ID of the file.
	 */
	public String filename;
	
	////////////////////    CONSTRUCTORS	////////////////////
	////////////////////   STATIC METHODS 	////////////////////
	//////////////////// 	  METHODS	////////////////////
	////////////////////   INNER CLASSES	////////////////////
    }
}
