/**
 * PatchworkProbeMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkProbeMsg.java,v 1.1 2001/12/04 19:44:26 strib Exp $
 */

package ostore.oil.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.Types;
import ostore.util.NodeId;
import ostore.util.ByteUtils;

/**
 * This is a probing message sent out by the <code>PatchworkStage</code>
 * in order to measure statistics of other nodes.
 *
 * @author Jeremy Stribling
 * @version    $Id: PatchworkProbeMsg.java,v 1.1 2001/12/04 19:44:26 strib Exp $
 */
public class PatchworkProbeMsg extends NetworkMessage {

    private int[] filler;

    public static final int LATENCY = 0;
    public static final int BANDWIDTH_1 = 1;
    public static final int BANDWIDTH_2 = 2;
    public static final int LOSS = 3;

    public static final int MINIMUM_SIZE = 17;

    public int type;
    public int fillerSize;

    public boolean response;

    /**
     * Constucts a probing message of a certain type, with a given
     * approximate size.  The size needs to be at least 
     * <code>MINIMUM_SIZE</code>, and if difference between it and 
     * <code>MINIMUM_SIZE</code> is not divisible by 4 it will be 
     * rounded down to a number that is.
     *
     * @param peer  the node to probe
     * @param totalSize   how big the message should be
     * @param type  the type of the message
     * @param response  whether this is a response
     */
    public PatchworkProbeMsg (NodeId peer, int totalSize, int type, 
			      boolean response) {

	super (peer, false /* outbound */);

	this.response = response;

	// make sure the sender knows it needs to be at least some bytes
	if( totalSize < MINIMUM_SIZE ) {
	  throw new RuntimeException( "Size of PatchworkProbeMsg must " +
				       "be at least " + MINIMUM_SIZE + 
				       "bytes" );
	}

	if( (type != LATENCY) && (type != BANDWIDTH_1) &&
	    (type != BANDWIDTH_2) && (type != LOSS) ) {
	  throw new RuntimeException( "Invalid type for PatchworkProbeMsg" );
	} else {
	  this.type = type;
	}

	fillerSize = totalSize - MINIMUM_SIZE;
	int remainder = fillerSize % 4;	

	// make sure the fillerSize is divisible by 4
	fillerSize -= remainder;

	// make filler to fill the rest of the array size
	filler = new int[ fillerSize/4 ];

    }

    public PatchworkProbeMsg (byte [] data, int [] offset) {
	super (data, offset);
	type = ByteUtils.bytesToInt( data, offset );
	response = ByteUtils.bytesToBool( data, offset );
	fillerSize = ByteUtils.bytesToInt( data, offset );
	filler = new int[0];
	offset[0] += fillerSize;
    }

    public void to_bytes (byte [] data, int [] offset) {
	super.to_bytes (data, offset);
	ByteUtils.intToBytes( type, data, offset );
	ByteUtils.boolToBytes( response, data, offset );
	ByteUtils.intToBytes( fillerSize, data, offset );
	if( data == null ) {
	  data = new byte[fillerSize + offset[0]];	
	}
	ByteUtils.intsToBytes( data, offset[0], filler, 0, fillerSize );
	offset[0] = offset[0] + fillerSize;
    }

    public int type_code () {
	return Types.OIL_PATCHWORK_PROBE_MSG;
    }

    public Object clone () throws CloneNotSupportedException {
	return super.clone ();
    }

    public String toString () {
	return "(PatchworkProbeMsg, type " + type + ", fillerSize " + 
		fillerSize + ", response: " + response + " "
		+ super.toString () + ")";
    }
}

