/*
 * EncryptedQS.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security;
import java.io.FileInputStream;
import java.security.Provider;
import java.security.Security;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.SecureRandom;
import java.security.InvalidKeyException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.BadPaddingException;
import ostore.util.Carp;
import ostore.util.QSIO;
import ostore.util.ByteUtils;
import ostore.util.QSException;
import ostore.util.QSString;
import ostore.util.QuickSerializable;
import ostore.util.TypeTable;
import ostore.util.Types;

public class EncryptedQS implements QuickSerializable {

    public EncryptedQS (QuickSerializable user_data,
	    QSCipher qs_cipher,
	    PublicKey pubkey,
	    SecureRandom random,
	    Cipher engine) throws InvalidKeyException,
	    IllegalBlockSizeException, BadPaddingException {

	if (! qs_cipher.provider  ().equals (
		    engine.getProvider ().getName ()) ||
	    ! qs_cipher.algorithm ().equals (engine.getAlgorithm ())) {
	    throw new InvalidKeyException ();
	}

	_qs_cipher = qs_cipher;

	byte [] plain_text = null;
	while (true) {
	    int [] offset = {0};
	    ByteUtils.intToBytes (user_data.type_code (), plain_text, offset);
	    user_data.to_bytes (plain_text, offset);
	    if (plain_text == null) 
		plain_text = new byte [offset [0]];
	    else
		break;
	}
	if (random != null)
	    engine.init (Cipher.ENCRYPT_MODE, pubkey, random);
	else
	    engine.init (Cipher.ENCRYPT_MODE, pubkey);
	_cipher_text = engine.doFinal (plain_text);
    }

    public EncryptedQS (byte [] data, int [] offset) throws QSException {
	int type_code = ByteUtils.bytesToInt (data, offset);
	_qs_cipher = (QSCipher) 
	    TypeTable.new_object (type_code, data, offset);
	int length = ByteUtils.bytesToInt (data, offset);
	_cipher_text = new byte [length];
	offset [0] += ByteUtils.memcpy (_cipher_text, 0, data, offset [0],
		                        length);
    }

    public QuickSerializable decrypt (PrivateKey privkey, Cipher engine)
    throws QSException, InvalidKeyException, IllegalBlockSizeException, 
    BadPaddingException {
	
	if (! _qs_cipher.provider  ().equals (
		    engine.getProvider ().getName ()) ||
	    ! _qs_cipher.algorithm ().equals (engine.getAlgorithm ())) {
	    throw new InvalidKeyException ();
	}

	engine.init (Cipher.DECRYPT_MODE, privkey);
	byte [] plain_text = engine.doFinal (_cipher_text);

	int [] offset = {0};
	int type_code = ByteUtils.bytesToInt (plain_text, offset);
	return TypeTable.new_object (type_code, plain_text, offset);
    }

    public void to_bytes (byte [] data, int [] offset) {
	ByteUtils.intToBytes (_qs_cipher.type_code (), data, offset);
	_qs_cipher.to_bytes (data, offset);
	ByteUtils.intToBytes (_cipher_text.length, data, offset);
	offset [0] += ByteUtils.memcpy (data, offset [0], _cipher_text, 0,
		                        _cipher_text.length);
    }

    public int type_code () {
	return Types.SECURITY_ENCRYPTED_QS;
    }

    public QSCipher cipher () { return _qs_cipher; }

    private QSCipher _qs_cipher;
    private byte [] _cipher_text;

    public static void main (String [] args) {

	try {
	    if (args.length < 1) {
		System.err.println ("must provide key prefix");
		System.exit (1);
	    } else {
		Carp.out( "running..." );
	    }

	    String key_prefix = args [0];
	    Types.registerAllTypes ();
	    Provider cryptix = new cryptix.jce.provider.CryptixCrypto ();
	    Security.addProvider (cryptix);

	    String clear_text = "Hello, world.";
	    byte [] wire = null;

	    {
		String fname = key_prefix + ".pub";
		FileInputStream file = new FileInputStream (fname);
		QSPublicKey pubkey = (QSPublicKey) QSIO.read (file);
		file.close ();

		Carp.out( "public key read" );
		
		EncryptedQS cipher_text = new EncryptedQS (
			new QSString (clear_text), 
			new QSRsaCipher (),
			pubkey.key (),
			null,
			Cipher.getInstance ("RSA/ECB/PKCS#1", "CryptixCrypto"));

		Carp.out( "cipher text created" );
		
		while (true) {
		    int [] offset = {0};
		    ByteUtils.intToBytes (cipher_text.type_code (), 
			    wire, offset);
		    cipher_text.to_bytes (wire, offset);
		    if (wire != null)
			break;
		    else 
			wire = new byte [offset [0]];
		}

		Carp.out( "cipher text serialized" );
	    }

	    {
		String fname = key_prefix + ".priv";
		FileInputStream file = new FileInputStream (fname);
		QSPrivateKey seckey = (QSPrivateKey) QSIO.read (file);
		file.close ();
		
		Carp.out( "private key read" );
		
		int [] offset = {0};
		int type_code = ByteUtils.bytesToInt (wire, offset);
		EncryptedQS cipher_text = (EncryptedQS)
		    TypeTable.new_object (type_code, wire, offset);

		Carp.out( "cipher text deserialized" );
		
		QSString new_clear_text = (QSString) cipher_text.decrypt (
			seckey.key (), 
			Cipher.getInstance (cipher_text.cipher ().algorithm (), 
			    cipher_text.cipher ().provider ()));

		Carp.out( "clear text recovered" );
		
		if (! clear_text.equals (new_clear_text.toString ())) {
		    Carp.die( "clear text does not match!" );
		} else {
		    Carp.out( "clear text matches" );
		}
	    }

	}
	catch (Exception e) {
	    System.err.println ("Caught exception " + e + 
		    ".  Stack trace follows:");
	    e.printStackTrace ();
	    System.exit (1);
	}

	System.exit (0);
    }
}

