/**
 * ThresholdSignature.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.Signature;
import java.security.SignatureException;
import java.security.InvalidKeyException;
import java.security.InvalidParameterException;
import java.security.MessageDigest;
import ostore.security.QSSignature;
import ostore.security.QSPublicKey;
import ostore.util.QSException;
import ostore.util.ByteUtils;
import ostore.util.Assert;
import ostore.util.Types;
import ostore.util.Debug;
import ostore.util.DebugFlags;

/**
 * A verification signature engine for threshold signatures.  Will not
 * actually produce signatures, only verify them.
 *
 * @author  Sean C. Rhea
 * @version $Id: ThresholdSignature.java,v 1.5 2001/11/16 22:44:45 geels Exp $
 */
public class ThresholdSignature extends Signature implements QSSignature {

    private static final boolean DEBUG = 
	DebugFlags.SECURITY_THRESHOLD_SIGNATURE;
    private static final String class_tag = "ThresholdSignature";

    private static final String ALGORITHM = "Threshold";
    
    public ThresholdSignature () { super (ALGORITHM); } 
    public ThresholdSignature (byte [] data, int [] offset) {
	super (ALGORITHM);
    }

    public String algorithm () { return ALGORITHM; }
    public String provider () { return ThresholdProvider.PROVIDER_STRING; }

    public void to_bytes (byte [] data, int [] offset) {}

    public int type_code () {
	return Types.SECURITY_THRESH_SIGNATURE;
    }

    public boolean equals( Object obj ) {
	return ( obj instanceof ThresholdSignature );
    }

    public int hashCode() {
	return type_code();
    }
    
 
    protected void engineInitSign (PrivateKey key) 
    throws InvalidKeyException {
	final String tag = class_tag + ".engineInitSign";
	Assert.assert (false, tag + ": not yet supported.");
    }

    protected void engineInitVerify (PublicKey key) 
    throws InvalidKeyException {
	final String tag = class_tag + ".engineInitVerify";

	if (! (key instanceof ThresholdPublicKey))
	    throw new InvalidKeyException ();

	if (_md == null) {
	    try {
		_md = MessageDigest.getInstance("SHA-1");
	    }
	    catch (Exception e) {
		Assert.assert (false, tag + ": could not instanciate " +
			"SHA-1 algorithm.");
	    }
	}
	else 
	    _md.reset ();

	_public_key = (ThresholdPublicKey) key;
    }

    protected void engineUpdate (byte b) throws SignatureException {
	final String tag = class_tag + ".engineUpdate";
	_md.update (b);
    }

    protected void engineUpdate (byte[] in, int offset, int length)
    throws SignatureException {
	final String tag = class_tag + ".engineUpdate";
	_md.update (in, offset, length);
    }

    protected byte[] engineSign() throws SignatureException {
	final String tag = class_tag + ".engineSign";
	Assert.assert (false, tag + ": not yet supported.");
	return null;
    }

    protected boolean engineVerify(byte[] signature)
    throws SignatureException {
	final String tag = class_tag + ".engineVerify";

	boolean result = false;
	try {
	    int [] offset = {0};
	    int num_shares = ByteUtils.bytesToInt (signature, offset);
	    SigShare [] sigs = new SigShare [num_shares];
	    for (int i = 0; i < num_shares; ++i) 
		sigs [i] = new SigShare (signature, offset);
	    byte [] data = _md.digest ();
	    result = SigShare.verify (data, sigs, _public_key);
	}
	catch (Exception e) { 
	    if (DEBUG) {
		Debug.printtagln (tag, ": caught " + e + ".");
		e.printStackTrace ();
	    }
	    return false; 
	}
	return result;
    }

    /**
     * Verify the signature without encoding the signature shares into
     * bytes.  This method is more efficient than using {@link
     * #verify(byte[])} if the <code>SigShare</code>s are already
     * deserialized, but it requires knowing the <code>Signature</code>
     * type (i.e., casting down to <code>ThresholdSignature</code>).
     *
     * <p>As with the {@link #verify(byte[])} method, this signature object
     * is reset (via calling {@link #engineInitVerify}) at the end of this
     * method.
     */
    public boolean verify (SigShare [] sigs) {
	final String tag = class_tag + ".verify";

	boolean result = false;
	try {
	    byte [] data = _md.digest ();
	    result = SigShare.verify (data, sigs, _public_key);
	    engineInitVerify (_public_key);
	}
	catch (Exception e) {
	    if (DEBUG) {
		Debug.printtagln (tag, "caught " + e + ".");
		e.printStackTrace ();
	    }
	    return false; 
	}
	return result;
    }

    protected void engineSetParameter(String param, Object value)
	throws InvalidParameterException {
	final String tag = class_tag + ".engineSetParameter";
        throw new InvalidParameterException(tag + ": " + param);
    }

    protected Object engineGetParameter(String param)
    throws InvalidParameterException {
	final String tag = class_tag + ".engineGetParameter";
        throw new InvalidParameterException(tag + ": " + param);
    }

    private MessageDigest _md;
    private ThresholdPublicKey _public_key;
}

