/***
 ** Oceanstore Tapestry Node Implementation
 ** (C) 2000 Ben Zhao
 ** @author Ben Zhao
 ** @version $Id: Link.java,v 1.3 2001/05/14 20:44:46 s_honey Exp $
 ** Implementation of Link Object
 **/

package ostore.tapestry;

import ostore.*;
import ostore.util.*;
import java.util.*;
import java.io.*;

public class Link {

 
    /** important to keep the values in this order -- (i.e. in order
        of severity.).  Used in Link.compare(Link,Link)*/
    public static final int LINK_UP = 0;
    public static final int LINK_DAMAGED = 1;
    public static final int LINK_DOWN = 2;
    public static final int LINK_BLACKLISTED=3;
    public static final int LINK_DISINTIGRATED=4;

    public static final int DEFAULT_LATENCY = Integer.MAX_VALUE;  
  
  /* Where does this link up/down stuff go?  second chances?  I'm
     having a hard time figuring out if the maintenance for old/bad
     nodes should be contained inside the NeighborMap or just inside
     the Tapestry.  If it's in the neighbor map, we'll have to do some
     awkward callback stuff to initialize a refreshing of the
     neighbormap if too many hosts go down...  Hmmm...*/
 
    private SHA1Hash t_guid;
    private NodeId t_nodeId;
 
    private float t_latency;
    private long t_lastRecv;
    private long t_lastSent;
    private int t_status;   
    private boolean t_local;
    
    public Link(SHA1Hash guid,  NodeId nodeId, float latency) {
	t_guid = guid;
	t_latency = latency;
	t_nodeId = nodeId;
	t_status=LINK_DOWN;//Should set up AFTER being created.
	t_lastSent = 0;
	t_lastRecv = 0;
	t_local=false;
    }

    public Link( NodeId nodeId, float latency) {
	this (new SHA1Hash(nodeId),nodeId,latency);
    }
 
    public Link(SHA1Hash guid,  NodeId nodeId) {
	this(guid,nodeId,DEFAULT_LATENCY);
    }
    

    public long lastSent(){
	return t_lastSent;
    }

    public long lastRecv(){
	return t_lastRecv;
    }

    
    /** If this is set, then last sent and recv are always set to max.*/
        public void setLocal(){
	    t_lastSent = Long.MAX_VALUE;
	    t_lastRecv = Long.MAX_VALUE;
	    t_local= true;
	    t_latency=0;
	    setUp();
	    
    }


    /**Run this when something is received from this node.  
       Will do a variety of tasks, including update the last 
       known time this node was alive. (and perhaps keep track 
       of usage stats.)
    */
    public void received(){
	if (!t_local){
	    t_lastRecv = System.currentTimeMillis();
	}
    }
    
    /**Run this when something is sent to this node.  
       Will do a variety of tasks (perhaps keep track of usage stats.)
    */
    public void sent(){
	if (!t_local){
	    t_lastSent = System.currentTimeMillis();
	}
    }
    

    

    /**WIP: debug make a HASHvalue that is essentially the same as
       the SHA1Hash.  That way things like latency won't mess with
       finding the right objects.  */
    
    public int hashCode(){
	return t_guid.hashCode();
    }
    

    
    public void setLatency(float latency) {
	if (!t_local){
	    t_latency = latency;
	}
    }
    
    public boolean isDown(){
	if (t_status==LINK_DOWN) return true;
	return false;
    }  
    
    public boolean isUp(){
	if (t_status==LINK_UP) return true;
	return false;
    }  
    
    public boolean isDamaged(){
	if (t_status==LINK_DAMAGED) return true;
	return false;
    }  
    
    public boolean isBlacklisted(){
	if (t_status==LINK_BLACKLISTED) return true;
	return false;
    }  
    
    
    public void setUp(){
	if (t_status!=LINK_BLACKLISTED){
	    t_status=LINK_UP;
	}
    }    
  
    public void setDown(){
	if ((t_status!=LINK_BLACKLISTED) && (!t_local)){
	    t_status=LINK_DOWN;
	}    
    }
    
    public void setDamaged(){
	if ((t_status!=LINK_BLACKLISTED) && (!t_local)){
	    t_status=LINK_DAMAGED;	    
	}    
    }
    
    public void setBlacklisted(){
	if (!t_local){
	    t_status=LINK_BLACKLISTED;	    
	}    
    }
    
    public void setUnBlacklisted(){
	if ((!t_local)&&isBlacklisted()){
	    t_status=LINK_DOWN;	    
	}
    }    
    
    public void setDisintigrated(){
	if (!t_local){
	    t_status=LINK_DISINTIGRATED;	    
	}    
    }

    public SHA1Hash getGuid() {
	return t_guid;
    }
    
    public NodeId getNodeId() {
	return t_nodeId;
    }
    
    public float getLatency() {
	return t_latency;
    }
    
    public int getStatus() {
	return t_status;
    }
    
    /** Compares which Link is more suited for use in the neighborMap.  Would use compareTo, but it's possible that it would be called on a null object.
	@param a can be null.
	@param b can be null.
	@return negative if A is better, positive if B, zero if tie.
	DONE.
    */

    public static float compare(Link a, Link b){

	if ((a==null) && (b==null)){
	    return 0;  //to prevent infinite loops when sorting two nulls.
	}

	if (a == null) {
	    return 1;
	}

	if (b == null) {
	    return -1;
	}

	//a is more up
	if (a.getStatus()<b.getStatus()){
	    return -1;
	}
	//b is more up.
	if (b.getStatus()<a.getStatus()){
	    return 1;
	}
	//returns neg. if A  is faster, b if not.  0 if equal.
	return   (a.getLatency() - b.getLatency());
	
    }

	 
}

 
    
   
  
 
    
   
  
 
    
   
  /*
 
 OLD, a different world, too.  (quadmaps?)    
   
      public Link(Guid _to, float _latency) {
    to = _to;
    latency = _latency;
    addr = null;  //undefined
  }
 
    
   
  */
