/** ManyTables uses hashtables and vectors to make one-many mappings.
 *
 * @version 1.0
 * @author Scott Honey
 * 
 */

package ostore.tapestry;

import java.util.*;
import ostore.util.*;

public class ManyTable {

    //pseudo constants.    
    private static final int INITIAL_HASH_SIZE = 10000;
    private static final int INITIAL_VECTOR_SIZE = 10;
    
    private int m_hashSize;
    private int m_vectorSize;
    
    private Hashtable m_objectHash;

    /**Default constructor, uses the static INITIAL_*_SIZE values.
     */    
    public ManyTable(){
	this(INITIAL_HASH_SIZE,INITIAL_VECTOR_SIZE);
    }

    /** The more precise constructor
	@param hashSize how big the hashtable should be
	@param vectorSize how big you want vectors to be
    */
    public ManyTable(int hashSize, int vectorSize){
	m_hashSize = hashSize;
	m_vectorSize = vectorSize;
	
	m_objectHash = new Hashtable(m_hashSize);
    }    


    
    /** Adds an object to a particular key's vector
	@param key the key
	@param obj the object...
    */
    public void addObject(Object key, Object obj){
	Vector v;
		
	if (m_objectHash.get(key)==null){
	    //key doesn't exist, make a new vector and add.
	    v = new Vector(m_vectorSize);
	    v.addElement(obj);
	    m_objectHash.put(key,v);
	} 
	else {
	    v = (Vector) m_objectHash.get(key);
	    if (!v.contains(obj)){
		//object doesn't exist, add it.
		v.addElement(obj);
	    }//if
	    
	}//if/else
    }
    
    /** Removes the key (and thus any associated objects) from the table
	@param key the key to remove
    */
    public void removeKey(Object key){
	if (m_objectHash.get(key)==null){
	    //ADDERROR
	    return;
	} else {
	    m_objectHash.remove(key);
	}
    }
    
    

    /** removes all traces of an Object from the table.<br>
	TODO this is probably dog slow.
	@param obj the object to remove.
    */
    public void removeObject(Object obj){
	Enumeration e = m_objectHash.keys();
	
	while (e.hasMoreElements()){
	    removeObject(e.nextElement(),obj);
	}
    }
    

    /** removes an object from a particular key.
	@param key the key
	@param obj the object to remove.
    */
    public void removeObject(Object key, Object obj){
	
	Vector v = (Vector) m_objectHash.get(key);

	if (v==null){
	    //key doesn't exist, do nothing.
	    return;
	}
	
	v.removeElement(obj);
	if (v.isEmpty()){
	    m_objectHash.remove(key);
	}
	
	return;
    }

    
    
    /** Checks to see if a certain key exists
	@param key the key
    */
    public boolean existKey(Object key){
	return m_objectHash.containsKey(key);
    }
    

    /** Checks to see if a particular object exists under any key.<br>
	TODO probably dog slow.
	@param obj the object to check for.
    */
    public boolean existObject(Object obj){
	Enumeration e = m_objectHash.keys();
	
	while (e.hasMoreElements()){
	    if (existObject(e.nextElement(),obj)) {
		return true;
	    }
	    
	}
	return false;
    }
    
    /** Checks to see if an object exists under a certain key
	@param key the key
	@param obj the object
    */
    public boolean existObject(Object key, Object obj){
	Vector v = (Vector) m_objectHash.get(key);
	if (v==null){
	    return false;
	}
	return v.contains(obj);
    }
    
    
    /** Gets all objects associated with a certain key
	@param key the key
    */
    protected Enumeration getObjects(Object key){
	Vector v = (Vector) m_objectHash.get(key);
	if (v==null) {
	    v = new Vector(0);
	}
	return v.elements();
    }
    
    /** @return the number of objects under a certain key.
     */
    protected int getNumObjects(Object key){
	Vector v = (Vector) m_objectHash.get(key);
	if (v==null) {
	    return 0;
	}
	return v.size();
    }
    
    


    /**runs test code, outputs a lot of text
     */
    public static void main(String[] args) {
	//test code
	ManyTable mt = new ManyTable();
	mt.test(100,5);
    }

    
    



    public void test(int numKeys, int numObjects){
	Random r = new Random();
	SHA1Hash key = new SHA1Hash("empty");
	SHA1Hash obj = new SHA1Hash("empty");
	
	System.out.println("Testing ManyTable");
	
	
	for (int i = 0; i < numKeys; i++){
	    key = new SHA1Hash((new Integer(i)).toString());

	    for (int j = 1000; j < 1000 +numObjects; j++){
		obj=new SHA1Hash(Integer.toHexString(j));
		
		addObject(key,obj);
		System.out.print("sent   " + key + "  " + obj);
		if (r.nextFloat()>.75){
		    System.out.print("\tremoved it " );
		    removeObject(key,obj);
		}
		System.out.println("\texists :  " + existObject(obj));
		
	    }
	    if (r.nextFloat()>.50){
		System.out.print("\tRemove Key  " + key );
		removeKey(key);
		System.out.println("\texists :  " + existKey(key));
	    }
	}
	
	for (int i = 0; i < numKeys; i++){
	    key = new SHA1Hash((new Integer(i)).toString());
	    
	    Enumeration e = getObjects(key );
	    System.out.println("Key" + key+" exists?" + existKey(key)+"\tGetting " + getNumObjects(key));
	    
	    while (e.hasMoreElements()){
		System.out.println("\t"+(SHA1Hash) e.nextElement());
	    }
	    System.out.println();
	}
	
	System.out.println("Testing removing objects in groups.");
	for (int j = 1000; j < 1000 +numObjects; j++){
	    obj=new SHA1Hash(Integer.toHexString(j));
	    if (r.nextFloat()>.5) {
		System.out.print("Did exist? "+ existObject(obj) +"\tRemoving " +obj);
		removeObject(obj);
		System.out.println("\t Exists?" + existObject(obj));
	    }
	    
	}
    }
    
}














