/**
 * RouteObjTMessage
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RouteObjTMessage.java,v 1.5 2001/08/16 23:21:18 srhea Exp $
 */

package ostore.tapestry;
import ostore.util.*;
import mdw.sandStorm.api.*;

/*********************
 * Event generated when a stage wants to route a message to the server
 * storing an object with some specified Guid.  This single method can be
 * used in a variety of ways.  The key fields are select, from, and exact
 * match.  select=S and from=F means to keep routing up the routing mesh
 * towards the object root node, stopping when a hop has more than F 
 * locations cached for the object.  The closest F such locations are chosen,
 * out of which S locations are then randomly chosen.  In other words, 
 * locality of the F locations is implied, while the random select S from F
 * is explicitly specified.  The primary use is to find and route a message
 * to the closest copy of a given object, where select=1, from=1.  The 
 * ObjGuid is the object's identifier.
 */
public abstract class RouteObjTMessage extends TMessageEvent 
       implements Cloneable {
    /**
     * The Guid of the object we are sending the message to
     */
    public SecureHash objGuid;

    /**
     * See Class comments, used to choose random object locations when available
     */
    public int select;
    /**
     * Same as select, see Class comments.
     */
    public int from;
    /** 
     * specifies whether the route taken to the object root
     * needs to be exact (as opposed to one that includes possible 
     * surrogate routes), and is false by default.
     */
    public boolean exact_match;

    /**
     * specifies a desired level of parallel queries to use
     * when searching for the object.  If > 0, we issue that many parallel
     * requests using redundant names generated by hashing that many natural
     * numbers with the object ID.
     */
    public int redunLvlRequested = Tapestry.LOCATION_REDUNDANCY;

    /**
     * is a class that stores the necessary fields for 
     * verifying the type of data requested.
     */
    public QuickSerializable verifyFields;

    public boolean validateObj = false;

    public RouteObjTMessage(SecureHash oGuid, int s, int f,
	                    boolean exact, int redLvl, 
	                    QuickSerializable vfields) {
	super(oGuid, 
              false, /* false, not inbound */
              true, /* using MAC by default */
              false); /* verified N/A */
	objGuid = oGuid;
	select = s;
	from = f;
	exact_match = exact;
	redunLvlRequested = redLvl;
	verifyFields = vfields;
    }

    public RouteObjTMessage (byte [] buffer, int [] offset) 
        throws QSException {
	super (buffer, offset);
	int type_code = ByteUtils.bytesToInt (buffer, offset);
	objGuid = (SecureHash) TypeTable.new_object (type_code, buffer, offset);
    }

    public void to_bytes(byte[] buffer, int[] offset) {
	super.to_bytes (buffer, offset);
	ByteUtils.intToBytes (objGuid.type_code (), buffer, offset);
	objGuid.to_bytes (buffer, offset);
    }

    public String toString() {
	return "(RouteObjTMessage "+super.toString()+")";
    }

    public Object clone() throws CloneNotSupportedException {
	RouteObjTMessage result = (RouteObjTMessage) super.clone();
	result.objGuid = objGuid;
	result.select = select;
	result.from = from;
	result.exact_match = exact_match;
	result.redunLvlRequested = redunLvlRequested;
	result.validateObj = validateObj;
	result.verifyFields = verifyFields;
	return result;
    }
}
