/*
 * RouteEntry.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Map;
import ostore.util.Assert;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.NodeId;
import ostore.util.SecureHash;

/**
 * Implements the routing table entry
 *
 * @author     Ben Y. Zhao
 * @version    $Id: RouteEntry.java,v 1.2 2002/01/11 09:40:41 ravenben Exp $
 */
public class RouteEntry {

  private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTER;

  public static final int REDUNDANT_LVL = 3;
    //redundant_lvl of 3 means that there are 2 backups per route entry

  private NodeId [] routes;
  private HashMap latencies;

  public RouteEntry() {
    routes = new NodeId [REDUNDANT_LVL];
    latencies = new HashMap(REDUNDANT_LVL);
  }

  public RouteEntry(NodeId route, Long latency) {
    routes = new NodeId [REDUNDANT_LVL];
    latencies = new HashMap(REDUNDANT_LVL);
    routes[0] = route;
    latencies.put(route, latency);
  }

  public NodeId getFirst() {
    return routes[0];
  }

  public NodeId getAtIndex(int i) {
    if (i < REDUNDANT_LVL) return routes[i];
    return null;
  }

  public void add(NodeId newnode, Long latency) {
    latencies.put(newnode, latency);
    sortbyLatency();
  }

  public void switchRoutes(int from, int to) {
    if (from >= REDUNDANT_LVL || to >= REDUNDANT_LVL ||
  	routes[from] == null || routes[to] == null) {
      if (DEBUG) System.err.println("RouteEntry.switch called with "+
				    "invalid indices.");
      return;
    }
    NodeId t = routes[to];
    routes[to] = routes[from];
    routes[from] = t;
  }

  //resort the nodes by latency
  public void sortbyLatency() {
    
    TreeMap sorted = new TreeMap();
    NodeId node;
    Long lat;
    for (Iterator i=latencies.keySet().iterator(); i.hasNext(); ) {
      node = (NodeId) i.next();
      lat = (Long) latencies.get(node);
      sorted.put(lat, node);
    }
    int j = 0;
    for (Iterator i=sorted.keySet().iterator(); i.hasNext(); j++) {
      lat = (Long) i.next();
      node = (NodeId) sorted.get(lat);
      if (j >= REDUNDANT_LVL) {
	latencies.remove(node);
      } else routes[j] = node;
    }
  }

  /**
   * Search to see if route is actually part of this
   * routeEntry. If so, then update its latency, and
   * then do a sort to reorder the routes by latency
   */
  public void updateLatency(NodeId route, Long latency) {
    if (latencies.containsKey(route)) {
      latencies.put(route, latency);
      sortbyLatency();
    }
  }

  public String toString() {
    return "["+size()+"]"+getFirst();
  }

  //returns number of non-null entries, including the primary
  public int size() {
    int i;
    for (i=0; i<REDUNDANT_LVL; i++)
      if (routes[i]==null) break;
    return i;
  }
}
