/*
 *  Router.java
 *
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Date;
import java.util.Collections;
import java.io.FileInputStream;
import mdw.sandStorm.api.ConfigDataIF;
import mdw.sandStorm.api.EventHandlerException;
import mdw.sandStorm.api.EventHandlerIF;
import mdw.sandStorm.api.QueueElementIF;
import mdw.sandStorm.api.SinkIF;
import mdw.sandStorm.api.SinkException;
import mdw.sandStorm.api.StagesInitializedSignal;
import mdw.sandStorm.api.NoSuchStageException;
import mdw.sandStorm.api.ProfilableIF;
import mdw.sandStorm.internal.SinkProxy;
import ostore.util.QSIO;
import ostore.util.QSByteArray;
import ostore.util.Carp;
import ostore.util.Clock;
import ostore.util.Assert;
import ostore.util.ByteUtils;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;
import ostore.util.NodeId;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.util.QSException;
import ostore.util.QuickSerializable;
import ostore.util.Pair;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.security.QSPublicKey;
import ostore.tapestry.api.TapestryRouteMsg;
import ostore.tapestry.api.TapestryMacRouteMsg;
import ostore.tapestry.api.TapestryPrefixRouteMsg;
import ostore.tapestry.api.TapestryInstance;
import ostore.tapestry.api.TapestryTag;
import ostore.tapestry.api.TapestryLocateMsg;
import ostore.tapestry.api.TapestryLocateFailure;
import ostore.tapestry.api.TapestryPublishMsg;
import ostore.tapestry.api.TapestryQuery;
import ostore.tapestry.api.TapestryQueryState;
import ostore.tapestry.api.TapestryQueryResult;
import ostore.tapestry.api.TapestryQueryResultMatch;
import ostore.tapestry.api.TapestryUnpublishMsg;
import ostore.tapestry.RouteTMessage; // legacy api
import ostore.tapestry.RouteObjTMessage; // legacy api
import ostore.tapestry.PublishTMessage; // legacy api
import ostore.mac.MacHandler;
import ostore.mac.HMACDHHandler;
import ostore.bloom.BloomQuery;

/**
 * Handles the routing functions of Tapestry.
 *
 * TODO:  The back pointers are currently maintained in sorted linked lists
 * which are sorted by distance.  All the elements in a list have the same
 * guid, and the lists are stored in a map indexed by those guids.  I don't
 * have to tell you that this is inefficient, but I can't get Java's
 * TreeSet class to replace the lists.  I think it has something to do with
 * what the Javadocs call compare being "compatible with" equals.  I don't
 * know.  In any case, the backpointers will need to be stored on disk at
 * some point, not all in memory, so this will all need an overhauling
 * then.  --srhea, 12/18/2001
 *
 * @author     Ben Zhao
 * @version    $Id: Router.java,v 1.36 2002/04/05 19:08:33 ravenben Exp $
 */
public class Router implements EventHandlerIF, ProfilableIF {

    private String class_tag = "Router";
    private static final boolean DEBUG = DebugFlags.TAPESTRY_ROUTER;
    private static final boolean DO_MACS = false;

    private void dispatch (QueueElementIF item) {
	final String tag = class_tag + ".dispatch";
	if (DEBUG) Debug.printtagln (tag, "dispatching " +
		item.getClass () + " event: " + item + ".");
	try {
	    Classifier.getClassifier(_self_node_id).dispatch(item);
	} catch (SinkException e) {
	    BUG (class_tag + ".dispatch: could not dispatch " + item);
	}
    }

    private static final void BUG (String msg) {
	Carp.die (msg + "  This is a bug in " +
		"ostore.tapestry.impl.Router.  " +
		"Please let Sean know about it ASAP.");
    }

    public void init (ConfigDataIF config) throws Exception {
	_self_node_id = new NodeId (config.getString ("node_id"));
	
	if (DEBUG) Debug.turnOn (class_tag);
	class_tag = class_tag + ".[" + _self_node_id.port () + "]";
	final String tag = class_tag + ".init";

	Types.registerAllTypes ();

	_using_bloom = config.getBoolean ("using_bloom");

	String stageName = config.getStage().getName();
	try
	{
	    if(config.getBoolean("Profile"))
	    {
		config.getManager().getProfiler().add("ostore.tapestry.impl.Router", this);
		SinkIF sink =
		    config.getManager().getStage(stageName).getSink();
		_sink = new SinkProxy(sink,
			      config.getManager(),
			      config.getStage().getWrapper());
	    }
	    else
		_sink =
		    config.getStage().getSink();
	} 
	catch (NoSuchStageException e) 
	{
	    Carp.die (tag + ": could not get own sink.");
	}

	_waiting_reqs = new LinkedList ();
	_routing_tbl = null;
	
	_network_latencies = new HashMap ();
	_pings_in_flight = new HashMap ();
	_pending_back_pointer_times = new HashMap ();
	_msgs_needing_macs = new HashMap ();

	String dynamic_route = config.getString ("dynamic_route");
	if (dynamic_route == null || dynamic_route.equals("static")) {
	  //starting using static tapestry, so 
	  //router needs to send _backpointers to DynamicTClient
	  _backpointers = new HashMap ();
	} else {
	  //starting using dynamic tapestry, will
	  //wait to get objectMap from DynamicTClient
	  _backpointers = null;
	    Filter filter = new Filter ();
	    if (! filter.requireType(
		  Class.forName("ostore.tapestry.impl.Router$ObjectMap")))
	      BUG (tag + ": could not require type " + 
		   "ostore.tapestry.impl.Router$ObjectMap");
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	if (config.getBoolean ("fake_keys")) {
	    _self_guid = new SHA1Hash (_self_node_id);
	}
	else {
	    String keyfilename = config.getString ("pkey");
	    try {
		FileInputStream keyfile = new FileInputStream (keyfilename);
		_pkey = (QSPublicKey) QSIO.read (keyfile);
		keyfile.close ();
	    }
	    catch (Exception e) {
		Carp.die ("TClient.init:  Caught exception " + e +
			" while trying to read pkey from disk.");
	    }

	    _self_guid = new SHA1Hash (_pkey);
	}

	System.out.println (tag + ": Tapestry Node ID is " + _self_guid + ".");
	TapestryInstance.instance (_self_node_id).set_self_guid (_self_guid);

	String [] event_types = {
	    "mdw.sandStorm.api.StagesInitializedSignal",
	    "ostore.tapestry.impl.RoutingTable$Available",
	    "ostore.tapestry.impl.NetworkLatencyReq",
	    "ostore.tapestry.impl.NetworkLatencyResp",
	    "ostore.tapestry.api.TapestryPublishMsg",
	    "ostore.tapestry.api.TapestryUnpublishMsg",
	    "ostore.tapestry.PublishTMessage", // legacy
	    "ostore.mac.MacHandler$KeyEstablished"
	};

	for (int i = 0; i < event_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	String [] inb_msg_types = {
	    "ostore.tapestry.impl.PongMsg",
	    "ostore.tapestry.impl.PingMsg",
	    "ostore.tapestry.impl.RouteMsg",
	    "ostore.tapestry.impl.RouteObjRootMsg",
	    "ostore.tapestry.impl.RouteObjDestMsg",
	    "ostore.tapestry.impl.PublishMsg",
	    "ostore.tapestry.impl.BloomWrapper"
	};

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}

	String [] outb_msg_types = {
	    "ostore.tapestry.api.TapestryRouteMsg",
	    "ostore.tapestry.api.TapestryPrefixRouteMsg",
	    "ostore.tapestry.api.TapestryLocateMsg",
	    "ostore.tapestry.RouteTMessage", // legacy
	    "ostore.tapestry.RouteObjTMessage" // legacy
	};

	for (int i = 0; i < outb_msg_types.length; ++i) {
	    Filter filter = new Filter ();
	    if (! filter.requireType (Class.forName (outb_msg_types [i])))
		BUG (tag + ": could not require type " + outb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (false)))
		BUG (tag + ": could not require inbound = false for " 
			+ outb_msg_types [i]);
	    Classifier.getClassifier(_self_node_id).subscribe (filter, _sink);
	}
    }

    public void destroy() throws Exception {
	// TODO
    }

    //  methods for ProfilableIF
    public int profileSize()
    {
	return _sink.size();
    }
    //  end methods for ProfilableIF

    /**
     * For use in derived classes.
     */
    protected void handle_stages_init_signal () {
	_self_digits = RoutingTable.digits (_self_guid);
	//if we're using staticTC, then send reference of backpointers
	//to the DynamicTClient stage
	if (dynamic_route == null || dynamic_route.equals("static")) {
	  dispatch(new Router.ObjectMap(_backpointers));
	}
    }

    protected void handle_route_t_message (RouteTMessage msg) {
	final String tag = class_tag + ".handle_route_t_message";
	if (DEBUG) Debug.printtagln (tag, "processing RouteTMessage " + msg);

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};
	NodeId next = route_to_root (dest, true, false, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got RouteTMessage for non-existent " +
		    "destination.  Dropping it.");
	    return;
	}

	if (next.equals (_self_node_id)) {
	    // This is to us.

	    RouteTMessage inb = (RouteTMessage) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;
	    inb.usingMAC = msg.usingMAC;
	    inb.verifiedMAC = true;  // it's from us, after all

	    dispatch (inb);
	}
	else {
	    RouteMsg outb = 
		new RouteMsg (next, _self_guid, msg.peer, false,
			digit [0], msg);

	    if (DO_MACS) {

	    if (msg.usingMAC) {

		if (DEBUG) Debug.printtagln (tag, "Trying to add a MAC " +
			"to the message.");

		QSByteArray mac = null;
		try {
		    mac = HMACDHHandler.getInstance ().doMAC (
			    outb.data_to_mac (), msg.peer);
		}
		catch (MacHandler.GuidNotEstablished e) {

		    if (DEBUG) Debug.printtagln (tag, 
			    "No secret key established yet.");
		    
		    synchronized (_msgs_needing_macs) {
			LinkedList l = 
			    (LinkedList) _msgs_needing_macs.get (msg.peer);
			if (l == null) {
			    if (DEBUG) Debug.printtagln (tag, 
				    "Starting a key agreement.");
		    
			    l = new LinkedList ();
			    _msgs_needing_macs.put (msg.peer, l);
			    HMACDHHandler.getInstance ().establish (msg.peer);
			}
			else if (DEBUG) Debug.printtagln (tag, 
				    "Key agreement already started.");
			    
			l.addLast (outb);
		    }
		    return;
		}
		catch (QSException e) {
		    Carp.die (tag + ": unreachable.");
		}

		if (DEBUG) Debug.printtagln (tag, "MAC added.");

		outb.mac = mac;
	    }

	    }
	    dispatch (outb);
	}
    }
	    
    protected void handle_tapestry_route_msg (TapestryRouteMsg msg) {
	final String tag = class_tag + ".handle_tapestry_route_msg";
	if (DEBUG) Debug.printtagln (tag, "processing TapestryRouteMsg " + msg);

	if (_using_bloom) {
	    dispatch (new BloomWrapper (msg.peer, msg));
	    return;
	}

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};
	NodeId next = route_to_root (dest, true, false, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) {
	      //MARK
	      System.out.println("bad message: "+msg);
	      System.out.println(_routing_tbl);

  	      Carp.out (tag + ": " + 
  			"got TapestryRouteMsg for non-existent " +
  			"destination.  Dropping it.");
	    }
	    return;
	}

	if (next.equals (_self_node_id)) {
	    // This is to us.

	    TapestryRouteMsg inb = (TapestryRouteMsg) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;

	    // The MAC need not be verified; it's from us, after all.

	    dispatch (inb);
	}
	else {
	    RouteMsg outb = 
		new RouteMsg (next, _self_guid, msg.peer, false,
			digit [0], msg);

	    if (DO_MACS) {

	    if (msg instanceof TapestryMacRouteMsg) {
		if (DEBUG) Debug.printtagln (tag, "Trying to add a MAC " +
			"to the message.");

		QSByteArray mac = null;
		try {
		    mac = HMACDHHandler.getInstance ().doMAC (
			    outb.data_to_mac (), msg.peer);
		}
		catch (MacHandler.GuidNotEstablished e) {
		    if (DEBUG) Debug.printtagln (tag, 
			    "No secret key established yet.");

		    synchronized (_msgs_needing_macs) {
			LinkedList l = 
			    (LinkedList) _msgs_needing_macs.get (msg.peer);
			if (l == null) {
			    if (DEBUG) Debug.printtagln (tag, 
				    "Starting a key agreement.");

			    l = new LinkedList ();
			    _msgs_needing_macs.put (msg.peer, l);
			    HMACDHHandler.getInstance ().establish (msg.peer);
			}
			else if (DEBUG) Debug.printtagln (tag, 
				"Key agreement already started.");

			l.addLast (outb);
		    }
		    return;
		}
		catch (QSException e) {
		    Carp.die (tag + ": unreachable.");
		}

		if (DEBUG) Debug.printtagln (tag, "MAC added.");

		outb.mac = mac;
	    }

	    }

	    dispatch (outb);
	}
    }

    protected void handle_tapestry_prefix_route_msg (
	    TapestryPrefixRouteMsg msg) {
	final String tag = class_tag + ".handle_tapestry_prefix_route_msg";
	if (DEBUG) Debug.printtagln (tag, "processing " + 
		"TapestryPrefixRouteMsg, full type " + 
		msg.getClass ().getName () + ". msg==" + msg);

	int [] dest = RoutingTable.digits (msg.peer);
	int [] digit = {0};
	NodeId next = route_to_root (dest, true, true, digit);

	if (next.equals (_self_node_id)) {

	    if (DEBUG) Debug.printtagln (tag, 
		    "It's to us.  Dispatching it locally.");
	
	    // This is to us.

	    TapestryPrefixRouteMsg inb = 
		(TapestryPrefixRouteMsg) ByteUtils.qs_clone (msg);
	    inb.peer = _self_guid;
	    inb.inbound = true;

	    dispatch (inb);
	}
	else {
	    if (DEBUG) Debug.printtagln (tag, 
		    "It's not to us.  Forwarding it to " + next + ".");
	
	    RouteMsg outb = 
		new RouteMsg (next, _self_guid, msg.peer, true,
			digit [0], msg);

	    dispatch (outb);
	}
    }

    protected void handle_route_msg (RouteMsg msg) {
	final String tag = class_tag + ".handle_route_msg";
	if (DEBUG) Debug.printtagln (tag, "processing RouteMsg, " +
		"msg==" + msg);

	int [] dest = RoutingTable.digits (msg.dest);
	int [] digit = {msg.hop + 1};
	NodeId next = route_to_root (dest, true, msg.prefix_only, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got RouteMsg for non-existent " +
		    "destination.  Dropping it.");
	    return;
	}

	if (next.equals (_self_node_id)) {
	    // This is to us.

	    if (DEBUG) Debug.printtagln (tag, "We're the end of the line.");

	    QuickSerializable user_data = null;
	    QuickSerializable data_to_mac = null;
	    try {
		user_data = msg.user_data ();
		data_to_mac = msg.data_to_mac ();
	    }
	    catch (QSException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+"caught " + e + 
			" while decoding " + msg + ".  Dropping it...");
		return;
	    }

	    if (user_data instanceof TapestryRouteMsg) {

		TapestryRouteMsg inb = null;
		try {
		    inb = (TapestryRouteMsg) user_data;
		}
		catch (ClassCastException e) {
		    BUG (tag + ": should have caught this above.");
		    return;
		}
		inb.peer = msg.src;
		inb.inbound = true;

		if (inb instanceof TapestryMacRouteMsg) {
		    if (DO_MACS) {

		    try {
			if ((msg.mac == null) || 
				(! HMACDHHandler.getInstance ().authenticate (
				    data_to_mac, msg.mac, msg.src))) {
			    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out
			    (tag + ": " +
				    "Got RouteMsg from " + msg.src + 
				    "via " + msg.peer + " with bad MAC." +
				    "  Dropping it.");
			    // TODO: maybe retry it instead?
			    return;
			}
		    }
		    catch (MacHandler.GuidNotEstablished e) {
			if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out
			(tag + ": " +
				"Got RouteMsg from " + msg.src + 
				"via " + msg.peer + " with a MAC, but " +
				"no secret key is currently established" +
				" with that node.  Dropping it.");
			// TODO: maybe retry it instead?
			return;
		    }

		    }
		}

		dispatch (inb);
	    }
	    else if (user_data instanceof TapestryPrefixRouteMsg) {
		TapestryPrefixRouteMsg inb = null;
		try {
		    inb = (TapestryPrefixRouteMsg) user_data;
		}
		catch (ClassCastException e) {
		    BUG (tag + ": should have caught this above.");
		    return;
		}
		inb.peer = msg.src;
		inb.inbound = true;

		dispatch (inb);
	    }
	    else if (user_data instanceof RouteTMessage) {
		RouteTMessage inb = null;
		try {
		    inb = (RouteTMessage) user_data;
		}
		catch (ClassCastException e) {
		    BUG (tag + ": should have caught this above.");
		    return;
		}
		inb.peer = msg.src;
		inb.inbound = true;
		inb.usingMAC = (msg.mac != null);

		if (DO_MACS) {

		if (msg.mac != null) {
		    inb.verifiedMAC = true;
		    try {
			if ((msg.mac == null) || 
				(! HMACDHHandler.getInstance ().authenticate (
				data_to_mac, msg.mac, msg.src))) {
			    if (DEBUG) Debug.printtagln (tag,
				    "Got RouteMsg from " + msg.src + 
				    "via " + msg.peer + " with bad MAC.");
			    inb.verifiedMAC = false;
			}
		    }
		    catch (MacHandler.GuidNotEstablished e) {
			if (DEBUG) Debug.printtagln (tag,
				"Got RouteMsg from " + msg.src + 
				"via " + msg.peer + " with a MAC, but " +
				"no secret key is currently established" +
				" with that node.");
			inb.verifiedMAC = false;
		    }
		}
		
		}
		else { // ! DO_MACS
		    inb.verifiedMAC = (msg.mac != null);
		}

		dispatch (inb);
	    }
	    else {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+
			"got RouteMsg with bad payload.  " +
			"Dropping it.");
		return;
	    }
	}
	else {

	    if (DEBUG) Debug.printtagln (tag, "Forwarding to " + next + ".");

	    RouteMsg outb = null;
	    try {
		outb = (RouteMsg) msg.clone ();
	    }
	    catch (CloneNotSupportedException e) {
		BUG (tag + ": wtf?");
	    }
	    outb.hop = digit [0];
	    outb.peer = next;
	    outb.inbound = false;

	    dispatch (outb);
	}
    }

    protected void handle_publish_t_message (PublishTMessage msg) {
	final String tag = class_tag + ".handle_publish_t_message";
	if (DEBUG) Debug.printtagln (tag, "processing PublishTMessage " + msg);

	publish (_self_guid, msg.objGuid, null /* no tag */,
		 true /*publishing*/, _self_node_id);
    }

    protected void handle_tapestry_publish_msg (TapestryPublishMsg msg) {
	final String tag = class_tag + ".handle_tapestry_publish_msg";
	if (DEBUG) Debug.printtagln (tag, "processing TapestryPublishMsg " + msg);

	publish (_self_guid, msg.guid, msg.tag, true /*publishing*/,
		 _self_node_id);
    }

    protected void handle_tapestry_unpublish_msg(TapestryUnpublishMsg msg) {
      final String tag = class_tag + ".handle_tapestry_unpublish_msg";
      if (DEBUG) Debug.printtagln(tag, "processing TapestryUnpublishMsg "+
				  msg);
      publish(_self_guid, msg.guid, msg.tag, false /*unpublishing*/,
	      _self_node_id);
    }

    protected void handle_publish_msg (PublishMsg msg) {
	final String tag = class_tag + ".handle_publish_msg";
	if (DEBUG) Debug.printtagln (tag, "processing PublishMsg " + msg);
	publish (msg.src, msg.guid, msg.tag, msg.publish, msg.peer);
    }

    protected void handle_route_obj_t_message (RouteObjTMessage msg) {
	final String tag = class_tag + ".handle_route_obj_t_message";
	if (DEBUG) Debug.printtagln (tag, "processing RouteObjTMessage " + msg);

	try {
	if (! route_to_dest (msg.objGuid, _self_guid, null, null, msg,
		    0, 0, 0, null)) {

	    int [] dest = RoutingTable.digits (msg.objGuid);
	    int [] digit = {0};
	    NodeId next = route_to_root (dest, false, false, digit);

	    if (! next.equals (_self_node_id)) {
		RouteObjRootMsg outb = new RouteObjRootMsg (
			next, _self_guid, msg.objGuid, null, null, null,
			digit [0], msg, _self_node_id);
		dispatch (outb);
	    }
	}
	}
	catch (QSException e) {
	    Carp.die (tag + ": unreachable.");
	}
    }
    
    protected void handle_bloom_wrapper (BloomWrapper wrapper) {
	final String tag = class_tag + ".handle_bloom_wrapper";
	if (wrapper.success) {
	    if (wrapper.user_data instanceof TapestryLocateMsg) {
		TapestryLocateMsg inb = 
		    (TapestryLocateMsg) ByteUtils.qs_clone (wrapper.user_data);

		inb.peer = wrapper.src;
		inb.inbound = true;
		inb.guid = wrapper.dest;
		inb.query = null;
		inb.query_state = null;

		dispatch (inb);
	    }
	    else if (wrapper.user_data instanceof TapestryRouteMsg) {
		TapestryRouteMsg inb = 
		    (TapestryRouteMsg) ByteUtils.qs_clone (wrapper.user_data);

		inb.peer = wrapper.src;
		inb.inbound = true;

		dispatch (inb);
	    }
	    else {
		BUG (tag + ": don't know what to do with " + 
			wrapper.user_data + ".");
	    }
	}
	else {
	    if (wrapper.user_data instanceof TapestryLocateMsg) {

		RouteObjRootMsg r = new RouteObjRootMsg (null, wrapper.src,
			wrapper.dest, null, null, null, -1, wrapper.user_data,
							 _self_node_id);

		handle_route_obj_root_msg (r);
	    }
	    else if (wrapper.user_data instanceof TapestryRouteMsg) {

		RouteMsg r = new RouteMsg (null, wrapper.src, wrapper.dest,
			false, -1, wrapper.user_data);

		handle_route_msg (r);
	    }
	    else {
		BUG (tag + ": don't know what to do with " + 
			wrapper.user_data + ".");
	    }
	}
    }

    protected void handle_tapestry_locate_msg (TapestryLocateMsg msg) {
	final String tag = class_tag + ".handle_tapestry_locate_msg";
	if (DEBUG) Debug.printtagln (tag, "processing TapestryLocateMessage "+
				     msg);

	if (_using_bloom && (msg.query == null) && (msg.id == null)) {
	    dispatch (new BloomWrapper (msg.guid, msg));
	    return;
	}

	try {
	if (! route_to_dest (msg.guid, _self_guid, msg.query,
		    msg.query_state, msg, 0, 0, 0, null)) {

	    int [] dest = RoutingTable.digits (msg.guid);
	    int [] digit = {0};
	    NodeId next = route_to_root (dest, false, false, digit);

	    if (! next.equals (_self_node_id)) {
		RouteObjRootMsg outb = new RouteObjRootMsg (
			next, _self_guid, msg.guid, msg.query,
			msg.query_state, msg.id, digit [0], msg,
			_self_node_id);
		if(DEBUG) Debug.printtagln(tag, 
					   "we are NOT the root." +
					   " we are forwarding the" +
					   " TapestryLocatemsg==" + msg +
					   " towards the root." +
					   "  outbound==" + outb);
		dispatch (outb);
	    }
	    else {
		// If we are the root, and we have no matching
		// backpointers, the location request has failed.
		if (msg.id != null)
		{
		    if(DEBUG) 
		      Debug.printtagln(tag, "If we are the root, and we" +
		      " have no matching backpointers, the location" +
				       " request has failed.");
		    dispatch (new TapestryLocateFailure 
		      (_self_guid, msg.id, msg.guid, _self_node_id));
		}
		else
		  if(DEBUG) Debug.printtagln(tag, 
					     "we ARE the root." +
					     " TapestryLocateMsg==" + msg +
					     " we have followed some" +
					     " backpointers so we are" +
					     " dropping msg!");
	    }
	}
	else
	  if(DEBUG) Debug.printtagln(tag, 
				     "We have followed ENOUGH backpointers" +
				     " so we are stopping!" +
				     " TapestryLocateMsg==" + msg);
	}
	catch (QSException e) {
	    Carp.die (tag + ": unreachable.");
	}
    }

    protected void handle_route_obj_root_msg (RouteObjRootMsg msg) {
	final String tag = class_tag + ".handle_route_obj_root_msg";
	if (DEBUG) Debug.printtagln (tag, "processing RouteObjRootMsg " + msg);

	try {
	    boolean result = false;
	    
	    if (msg._user_data_buffer != null) {
		result = route_to_dest (msg.guid, msg.src, msg.query, 
			msg.query_state,
			null, msg._user_data_length, msg._user_data_offset, 
			msg._user_data_type, msg._user_data_buffer);
	    }
	    else {
		// if we're running under the simulator, the message has
		// never been serialized, so we have to use user_data
		// instead of the _user_data_buffer
		result = route_to_dest (msg.guid, msg.src, msg.query, 
			msg.query_state, msg.user_data (), 0, 0, 0, null);
	    }

        if (! result) {

	    if (DEBUG) Debug.printtagln (tag, "no route to dest" );
	    int [] dest = RoutingTable.digits (msg.guid);
	    int [] digit = {msg.hop + 1};
	    NodeId next = route_to_root (dest, false, false, digit);
	    if (DEBUG) Debug.printtagln (tag, "next: " + next);

	    if (! next.equals (_self_node_id)) {
		RouteObjRootMsg outb = null;
		try {
		    outb = (RouteObjRootMsg) msg.clone ();
		}
		catch (CloneNotSupportedException e) {
		    BUG (tag + ": unreachable.");
		}
		outb.peer = next;
		outb.inbound = false;
		outb.hop = digit [0];
		if (DEBUG) Debug.printtagln (tag, "outb: " + outb );
		dispatch (outb);
	    } 
	    else {
		QuickSerializable user_data = null;
		try {
		    user_data = msg.user_data ();
		}
		catch (QSException e) {
		    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + "caught " + e + 
			    " while decoding " + msg + ".  Dropping it...");
		    return;
		}

		// If we are the root, and we have no matching
		// backpointers, the location request has failed.
		if (user_data instanceof TapestryLocateMsg) {
		    // We don't support failures for RouteObjTMessages.
		    if (msg.id != null)
                    {
		      dispatch (new TapestryLocateFailure (
			        _self_guid, msg.id, msg.guid, msg.requestor));
                    }
		}
	    }
	}
	}
	catch (QSException e) {
	    Carp.die (tag + ": unreachable.");
	}
    }

    public void handle_route_obj_dest_msg (RouteObjDestMsg msg) {
	final String tag = class_tag + ".handle_route_obj_dest_msg";
	if (DEBUG) Debug.printtagln (tag, "processing RouteObjDestMsg " + msg);

	int [] dest = RoutingTable.digits (msg.dest);
	int [] digit = {msg.hop + 1};
	NodeId next = route_to_root (dest, true, false, digit);

	if (next == null) {
	    if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": " + 
		    "got RouteObjDestMsg for non-existent " +
		    "destination.  Dropping it.");
	    return;
	}

	if (next.equals (_self_node_id)) {
	    QuickSerializable user_data = null;
	    try {
		user_data = msg.user_data ();
	    }
	    catch (QSException e) {
		if (DebugFlags.LOG_BAD_NETWORK_EVENTS) Carp.out (tag + ": "
		+"caught " + e + 
			" while decoding " + msg + ".  Dropping it...");
		return;
	    }

	    if (user_data instanceof TapestryLocateMsg) {
		TapestryLocateMsg inb = (TapestryLocateMsg) user_data;
		inb.peer = msg.src;
		inb.inbound = true;
		inb.guid = msg.guid;
		inb.query = null;
		inb.query_state = null;
		dispatch (inb);
	    }
	    else {
		RouteObjTMessage inb = (RouteObjTMessage) user_data;
		inb.inbound = true;
		inb.usingMAC = false;
		inb.verifiedMAC = false;
		inb.peer = msg.src;
		Assert.assert (inb.objGuid.equals (msg.guid), 
			tag + ": crap!");
		dispatch (inb);
	    }
	}
	else {
	    RouteObjDestMsg outb = null;
	    try {
		outb = (RouteObjDestMsg) msg.clone ();
	    }
	    catch (CloneNotSupportedException e) {
		BUG (tag + ": wtf?");
	    }
	    outb.peer = next;
	    outb.inbound = false;
	    outb.hop = digit [0];

	    dispatch (outb);
	}
    }

    /**
     * Take into account the result of a NetworkLatencyReq query.  Unlike
     * handle_network_latency_req, below, this will remain part of Router
     * even after the network introspection code that services a
     * NetworkLatencyReq is moved out.
     */
    private void handle_network_latency_resp (NetworkLatencyResp resp) {
	final String tag = class_tag + ".handle_network_latency_resp";
	if (DEBUG) Debug.printtagln (tag, "processing NetworkLatencyResp " + resp);

	LinkedList s = null;
	synchronized (_pending_back_pointer_times) {
	    s = (LinkedList) 
		_pending_back_pointer_times.remove (resp.node_guid);
	}

	if (s == null) 
	    return;

	while (! s.isEmpty ()) {
	    Pair p = (Pair) s.removeFirst ();
	    PublishInfo pi = (PublishInfo) p.first;
	    LinkedList l = (LinkedList) p.second;
	    if (DEBUG) Debug.printtagln (tag, "Changing distance of " +
		    pi + " to " + resp.latency_millis + " ms.  " +
		    "Current elts are " + l);

	    synchronized (l) {
		PublishInfo pi2 = null;
		ListIterator i = l.listIterator (); 
		while (i.hasNext ()) {
		    pi2 = (PublishInfo) i.next ();
		    if (pi.equals (pi2))
			break;
		}
		i.remove ();

		pi2.distance = resp.latency_millis;

		i = l.listIterator ();
		while (i.hasNext ()) {
		    PublishInfo pi3 = (PublishInfo) i.next ();
		    if (pi2.distance < pi3.distance) {
			i.previous ();
			break;
		    }
		}
		i.add (pi2);
	    }
	}
    }

    /** 
     * Handle the event requesting the latency to some other network node
     * through Tapestry.  This should be moved to some introspective
     * network stage at some point in the future.
     */
    private void handle_network_latency_req (NetworkLatencyReq req) {
	final String tag = class_tag + ".handle_network_latency_req";
	if (DEBUG) Debug.printtagln (tag, "processing NetworkLatencyReq " + req);

	Long l = (Long) _network_latencies.get (req.node_guid);
	if (l != null) {
	    if (DEBUG) Debug.printtagln (tag, "Distance to " +
		    req.node_guid + " is already known to be " +
		    l + ".");
	    dispatch (new NetworkLatencyResp (req.node_guid, l.longValue ()));
	}
	else {
	    if (DEBUG) Debug.printtagln (tag, "Distance to " +
		    req.node_guid + " is not yet known.");
	    Long k = (Long) _pings_in_flight.get (req.node_guid);
	    if (k == null) {
		if (DEBUG) Debug.printtagln (tag, "Sending a ping.");
		_pings_in_flight.put (req.node_guid, 
			new Long ((Clock.current_date (_self_node_id)).getTime ()));
		handle_tapestry_route_msg (new PingMsg (req.node_guid));
	    }
	}
    }

    /**
     * Implements the functionality of NetworkLatencyReq.  This should be
     * moved to some introspective network stage at some point in the
     * future.
     */
    private void handle_ping_msg (PingMsg msg) {
	final String tag = class_tag + ".handle_ping_msg";
	if (DEBUG) Debug.printtagln (tag, "processing PingMsg " + msg);

	// TODO: think about flood attacks, timeouts, sequence numbers...
	handle_tapestry_route_msg (new PongMsg (msg.peer));
    }

    /**
     * Implements the functionality of NetworkLatencyReq.  This should be
     * moved to some introspective network stage at some point in the
     * future.
     */
    private void handle_pong_msg (PongMsg msg) {
	final String tag = class_tag + ".handle_pong_msg";
	if (DEBUG) Debug.printtagln (tag, "processing PongMsg " + msg);

	Long l = (Long) _pings_in_flight.get (msg.peer);
	if (l != null) {
	    long now = (Clock.current_date (_self_node_id)).getTime ();
	    long lat = now - l.longValue ();

	    if (DEBUG) Debug.printtagln (tag, "Updating distance to " +
		    msg.peer + " to be " + lat + " ms.");
	    
	    _network_latencies.put (msg.peer, new Long (lat));
	    _pings_in_flight.remove (msg.peer);
	    dispatch (new NetworkLatencyResp (msg.peer, lat));
	}
	else if (DEBUG) Debug.printtagln (tag, "got pong without " +
		"corresponding ping from node " + msg.peer + ".");
    }

  private void handle_object_map (Router.ObjectMap msg) {
    final String tag = class_tag +".handle_object_map";
    if (DEBUG) Debug.printtagln(tag, "processing ObjectMap "+msg);
    if (DEBUG)
      Assert.assert(_backpointers == null, 
		    tag + ": got two versions of object map");
    
    _backpointers = msg.objmap;

    //check if routing table is also complete
    //if so, then start normal operation and execute queued events
    if (_routing_tbl != null) {
      try {
	while (! _waiting_reqs.isEmpty ()) {
	  QueueElementIF e = 
	    (QueueElementIF) _waiting_reqs.removeFirst ();
	  handleEvent (e);
	}
      }
      catch (EventHandlerException e) {
	BUG (tag + ": what do I do now?"); // TODO
      }
    }
  }
    
    private void handle_routing_table_available (RoutingTable.Available msg) {

	final String tag = class_tag + ".handle_routing_table_available";
	if (DEBUG) Debug.printtagln (tag, "processing RoutingTable.Available " + msg);

	if (DEBUG)
	  Assert.assert (_routing_tbl == null, tag + ": got two versions " +
			 "of the routing table.");

	_routing_tbl = msg.table;
	
	if (dynamic_route != null && 
	    dynamic_route.equals("static")) {
	  dispatch (new RoutingTable.Available (_routing_tbl));
	} else {
	  System.out.println("Tapestry: ready.\n");
	  System.out.flush();
	}	  

	//check, and only execute queued events if _backpointers is there
	if (_backpointers != null) {
	  try {
	    while (! _waiting_reqs.isEmpty ()) {
	      QueueElementIF e = 
		(QueueElementIF) _waiting_reqs.removeFirst ();
	      handleEvent (e);
	    }
	  }
	  catch (EventHandlerException e) {
	    BUG (tag + ": what do I do now?"); // TODO
	  }
	}
    }

    private void handle_key_established (MacHandler.KeyEstablished event) {
	final String tag = class_tag + ".handle_key_established";

	if (DO_MACS) {

	if (DEBUG) Debug.printtagln (tag, 
		"processing MacHandler.KeyEstablished" );

	if (DEBUG) Debug.printtagln (tag, "Key established for node " +
		"with guid " + event.guid + ".");

	LinkedList l = null;
	synchronized (_msgs_needing_macs) {
	    l = (LinkedList) _msgs_needing_macs.remove (event.guid);
	}
	while (! l.isEmpty ()) {
	    RouteMsg outb = (RouteMsg) l.removeFirst ();
	    QSByteArray mac = null;
	    try {
		mac = HMACDHHandler.getInstance ().doMAC (
			outb.data_to_mac (), event.guid);
	    }
	    catch (MacHandler.GuidNotEstablished e) {
		synchronized (_msgs_needing_macs) {
		    l.addFirst (outb);
		    LinkedList l2 = 
			(LinkedList) _msgs_needing_macs.get (event.guid);
		    if (l2 != null) {
			while (! l2.isEmpty ()) 
			    l.addLast (l2.removeFirst ());
		    }
		    _msgs_needing_macs.put (event.guid, l);
		    HMACDHHandler.getInstance ().establish (event.guid);
		}
		return;
	    }
	    catch (QSException e) {
		Carp.die (tag + ": unreachable.");
	    }
	    outb.mac = mac;

	    int [] dest = RoutingTable.digits (outb.dest);
	    int [] digit = {0};
	    NodeId next = route_to_root (dest, true, false, digit);

	    Assert.assert (next != _self_node_id, tag + ": bug.");

	    outb.peer = next;
	    outb.hop = digit [0];

	    dispatch (outb); 
	}

	}
    }

    public void handleEvent(QueueElementIF item) 
	throws EventHandlerException {
        final String tag = class_tag + ".handleEvent";

	if (DEBUG) Debug.printtagln (tag, "got " + item.getClass () + 
		    " event: " + item + ".");

	if (item instanceof StagesInitializedSignal) {
	    handle_stages_init_signal ();
	}
	else if (item instanceof RoutingTable.Available) {
	    handle_routing_table_available ((RoutingTable.Available) item);
	}
	else if (item instanceof Router.ObjectMap) {
	  handle_object_map((Router.ObjectMap) item);
	}
	else if (item instanceof BloomWrapper) {
	    handle_bloom_wrapper ((BloomWrapper) item);
	}
	else if (_routing_tbl == null || 
		 _backpointers == null) {
	    if (DEBUG) Debug.printtagln (tag, "Not up yet, queuing it."+
					 "tbl="+_routing_tbl+"; backs="+
					 _backpointers);
	    _waiting_reqs.addLast (item);
	    return;
	}
	else {
	    if (item instanceof RouteTMessage) {
		handle_route_t_message ((RouteTMessage) item);
	    }
	    else if (item instanceof TapestryRouteMsg) {
		if (((TapestryRouteMsg) item).inbound) {
		    if (item instanceof PingMsg) {
			handle_ping_msg ((PingMsg) item);
		    }
		    else if (item instanceof PongMsg) {
			handle_pong_msg ((PongMsg) item);
		    }
		    else {
			BUG (tag + ": should be a ping or pong");
		    }
		}
		else
		    handle_tapestry_route_msg ((TapestryRouteMsg) item);
	    }
	    else if (item instanceof TapestryPrefixRouteMsg) {
		handle_tapestry_prefix_route_msg ((TapestryPrefixRouteMsg) item);
	    }
	    else if (item instanceof RouteMsg) {
		handle_route_msg ((RouteMsg) item);
	    }
	    else if (item instanceof PublishTMessage) {
		handle_publish_t_message ((PublishTMessage) item);
	    }
	    else if (item instanceof TapestryPublishMsg) {
		handle_tapestry_publish_msg ((TapestryPublishMsg) item);
	    }
	    else if (item instanceof TapestryUnpublishMsg) {
	        handle_tapestry_unpublish_msg ((TapestryUnpublishMsg) item);
	    }
	    else if (item instanceof PublishMsg) {
		handle_publish_msg ((PublishMsg) item);
	    }
	    else if (item instanceof RouteObjTMessage) {
		handle_route_obj_t_message ((RouteObjTMessage) item);
	    }
	    else if (item instanceof TapestryLocateMsg) {
		handle_tapestry_locate_msg ((TapestryLocateMsg) item);
	    }
	    else if (item instanceof RouteObjRootMsg) {
		handle_route_obj_root_msg ((RouteObjRootMsg) item);
	    }
	    else if (item instanceof RouteObjDestMsg) {
		handle_route_obj_dest_msg ((RouteObjDestMsg) item);
	    }
	    else if (item instanceof NetworkLatencyReq) {
		handle_network_latency_req ((NetworkLatencyReq) item);
	    }
	    else if (item instanceof NetworkLatencyResp) {
		handle_network_latency_resp ((NetworkLatencyResp) item);
	    }
	    else if (item instanceof MacHandler.KeyEstablished) {
		handle_key_established ((MacHandler.KeyEstablished) item);
	    }
	}
    }

    // if user_data is null, we will use the remaining four arguments
    // instead.  this is done so that we don't deserialize payload until we
    // reach the destination
    protected boolean route_to_dest (SecureHash guid, SecureHash src, 
	    TapestryQuery query, TapestryQueryState state, 
	    QuickSerializable user_data, int user_data_length,
	    int user_data_offset, long user_data_type, 
	    byte [] user_data_buffer) throws QSException {

	final String tag = class_tag + ".route_to_dest";
	if (DEBUG) Debug.printtagln (tag, "called" );
	
	LinkedList l = (LinkedList) _backpointers.get (guid);
	boolean stop = false;

	if (l != null) {
	    if (DEBUG) Debug.printtagln (tag, "scanning backpointers " + l +
					 " l.size==" + l.size());
	    LinkedList dispatches = new LinkedList ();

	    Iterator i = l.iterator ();
	    while (i.hasNext ()) {
		PublishInfo ii = (PublishInfo) i.next ();
		if (DEBUG) Debug.printtagln (tag, "checking: " + ii );

		SecureHash [] matches = null;
		if (query != null) {
		    if (DEBUG) Debug.printtagln (tag, "running query: " + query );
		    TapestryQueryResult qr = 
			query.query (ii.src, ii.tag, state);
		    if (DEBUG) Debug.printtagln (tag, "result: " + qr );
		    if (qr instanceof TapestryQueryResultMatch) {
			TapestryQueryResultMatch m = 
			    (TapestryQueryResultMatch) qr;
			matches = m.matches ();
			stop = m.stop ();
		    }
		    else 
			stop = false;
		}
		else {
		    if (DEBUG) Debug.printtagln (tag, "no query" );
		    matches = new SecureHash [1];
		    matches [0] = ii.src;
		    stop = true;
		}

		if (DEBUG) Debug.printtagln (tag, "matches: " + matches );
		if( matches != null ) {
		    for (int j = 0; j < matches.length; ++j) {
			
			if (DEBUG) Debug.printtagln (tag, "matches[" + j +
						     "]: " + matches[j] );
			int [] dest = RoutingTable.digits (matches [j]);
			int [] digit = {0};
			//if (matches[j].equals(_self_guid) && DEBUG) {
			//  Debug.printtagln(tag, "Warning: match is US!");
			  //TODO: would like to print out routing table here
			//}
			NodeId next = route_to_root (dest, true, false, digit);
			if (DEBUG) Debug.printtagln (tag, "next: " + next );
			    
			if (next == _self_node_id) {
			    if (DEBUG) Debug.printtagln (tag, "next is self" );

			    QuickSerializable event = null;
			    if (user_data != null) {
				event = ByteUtils.qs_clone (user_data);
			    }
			    else {
				int [] offset = {user_data_offset};
				event = TypeTable.new_object (
					user_data_type, 
					user_data_buffer, 
					offset);
			    }

			    if (event instanceof RouteObjTMessage) {
				RouteObjTMessage r = (RouteObjTMessage) event;
				r.peer = src;
				r.inbound = true;
				r.usingMAC = false;
				r.verifiedMAC = false;
			    }
			    else {
				TapestryLocateMsg r = (TapestryLocateMsg) event;
				r.peer = src;
				r.inbound = true;
				r.guid = guid;
				r.query = null;
				r.query_state = null;
				if (DEBUG) Debug.printtagln (tag, 
					"event==" + event +
					" r==" + r);
			    }
			    dispatches.addLast (event);
			}
			else {
			    RouteObjDestMsg outb = null;
			    
			    if (user_data != null) {
				// if we're running under the simulator,
				// the message has never been serialized,
				// so we have to use user_data instead of
				// the _user_data_buffer

				outb = new RouteObjDestMsg (next, 
					src, ii.src, guid, digit [0], user_data);
			    }
			    else {
				outb = new RouteObjDestMsg (next, 
					src, ii.src, guid, digit [0], 
					user_data_length, user_data_offset,
					user_data_type, user_data_buffer);
			    }
			    dispatches.addLast (outb);
			}
		    }
		}

		if (stop) 
		    break;
	    }

	    while (! dispatches.isEmpty ()) {
		QueueElementIF d = (QueueElementIF) dispatches.removeFirst ();
		if (DEBUG) Debug.printtagln (tag, "dispatching: "+ d );
		dispatch (d);
	    }
	} else {
	    if (DEBUG) Debug.printtagln (tag, "no backpointers exist" );
	}
	if (DEBUG) Debug.printtagln (tag, "done (" + (stop?"stop":"go") + ")");
	return stop;
    }

    protected void publish (SecureHash src, SecureHash guid,
	    TapestryTag ttag, boolean publish, NodeId lasthop) {
	final String tag = class_tag + ".publish";

	// Store a backpointer here, initially at infinite distance.
	PublishInfo pi = new PublishInfo (src, ttag, lasthop);
	if (src.equals (_self_guid))
	  pi.distance = 0;

	LinkedList l = (LinkedList) _backpointers.get (guid);
	if (l == null) {
	    l = new LinkedList ();
	    if (publish)
	      _backpointers.put (guid, l);
	    else if (DEBUG)
	      Debug.printtagln(tag, "Unpublishmsg received for " +
		      "non-existent object pointer. Will forward on.");
	}
	if(DEBUG)
	  Debug.printtagln(tag, (publish ? "adding publishInfo==" :
				 "removing publishInfo==") + pi + 
			   " for guid==" + guid + 
			   " to backpointers set==" + l + " BEFORE."); 
	synchronized (l) {
	  if (publish) {
	      if (pi.distance == 0)
		  l.addFirst (pi);
	      else
		  l.addLast (pi);
	  } else {
	    if (l.contains(pi)) {
	      l.remove(pi);
	      if (DEBUG)
		Debug.printtagln(tag, "Removing publishinfo...");
	    }
	  }
	}
 
	if (publish && pi.distance == Long.MAX_VALUE) {
	    // Start looking for the real distance to the publisher.
	    synchronized (_pending_back_pointer_times) {
		LinkedList s = (LinkedList) 
		    _pending_back_pointer_times.get (src);
		if (s == null) {
		    s = new LinkedList ();
		    _pending_back_pointer_times.put (src, s);
		    dispatch (new NetworkLatencyReq (src));
		}
		s.addLast (new Pair (pi, l));
	    }
	} else if (!publish && pi!=null && !lasthop.equals(pi.lasthop)) {
	  //if we're unpublishing, check to see if we need to backtrack
	  PublishMsg back = new PublishMsg(pi.lasthop, src, guid,
					   ttag, 0, false);
	  dispatch(back);
	}

	// Route the message on towards the root.
	int [] dest = RoutingTable.digits (guid);
	int [] digit = {0};
	NodeId next = route_to_root (dest, false, false, digit);

	if (! next.equals (_self_node_id)) {
	    PublishMsg outb = new PublishMsg (next, src, guid, ttag, 
		    digit [0], publish);
	    dispatch (outb);
	}
    }

    public void handleEvents(QueueElementIF items[]) 
    throws EventHandlerException {
	for (int i=0; i < items.length; i++)
	    handleEvent (items[i]);
    }

    /**
     * Route to the root of the tree for <code>dest_digits</code>, assuming
     * that the first <code>digit [0] - 1</code> digits have already been
     * accounted for, and return the next node in the routing path.  
     * 
     * @param dest_digits the digits of the destination GUID
     * @param exact       whether this should be an exact match; if this
     *                    parameter is true, the function will return
     *                    <code>null</code> if there is no such node.
     * @param digit       the digit to start routing from
     *
     * @return            the next node in the routing path, if any; return
     *                    <code>null</code> if there is no such node and
     *                    return <code>_self_node_id</code> if the current
     *                    node is the deterministic root
     */
    public NodeId route_to_root (int [] dest_digits, boolean exact, 
	    boolean prefix, int [] digit) {
	final String tag = class_tag + ".route_to_root";

	if (DEBUG) {
	    String msg = "called.  Our guid: "; 
	    for (int i = 0; i < dest_digits.length; ++i) {
		msg += Integer.toHexString (_self_digits [i]);
	    }
	    msg += ".  Looking for: ";
	    for (int i = 0; i < dest_digits.length; ++i) {
		msg += Integer.toHexString (dest_digits [i]);
	    }
	    msg += (prefix ? "prefix." : (exact ? ", exact." : ", fuzzy."));
	    Debug.printtagln (tag, msg);
	}
	
	// Start with the requested digit.

	for ( ; digit [0] < _routing_tbl.digits_per_guid (); ++digit [0]) {
	    
	    int val = dest_digits [digit [0]];

	    //	    if (DEBUG) Debug.printtagln (tag, "digit " + digit [0] + 
	    //		    " looking for " + Integer.toHexString (val) + ".");

	    NodeId entry = _routing_tbl.read (digit [0], val);

	    // If there is no routing entry at the given digit (not even
	    // ourselves) find the next best one by adding circularly to
	    // the desired value.  This is then the ``surrogate'' root.

	    if (entry == null) {
		if (prefix) {
		    // Go back to the previous digit if there is one.
		    if (digit [0] > 0) 
			digit [0] -= 1;

		    // Look up the routing entry.
		    val = dest_digits [digit [0]];
		    entry = _routing_tbl.read (digit [0], val); 

		    // If there is no such entry, pick the next one
		    // circularly.
		    if (entry == null) {
			int i = (val == _routing_tbl.digit_values () - 1) ? 0 : val;

			while (_routing_tbl.read (digit [0], i) == null)
			    i = (i == _routing_tbl.digit_values () - 1) ? 0 : i + 1;

			entry = _routing_tbl.read (digit [0], i);
		    }
		    return entry; // Which is now non-null.
		}
		else if (exact) {
		    if (DEBUG) Debug.printtagln (tag, 
			    "no one matches; failing.");
		    return null;
		}

		if (DEBUG) Debug.printtagln (tag, 
			"no one matches; approximating");
			
		entry = _routing_tbl.surrogate (dest_digits, val, digit [0]);
		if (entry == null) 
		    return _self_node_id;  // finish early
	    }

	    // Now entry != null.

	    if (! entry.equals (_self_node_id)) {

		// If the surrogate root is not us, we need to forward on
		// to whoever it is.

		if (DEBUG) Debug.printtagln (tag, "Node " + entry +
			", guid " + _routing_tbl.node_id_to_guid (entry) + 
			" matches.");

		return entry;
	    }

	    // If the (surrogate) root is us, continue to the next digit.
	}

	// If we are the surrogate root for the last digit, we are
	// deterministically the root for all nodes.

	if (DEBUG) Debug.printtagln (tag, "we match.");
	return _self_node_id;
    }

  public static class ObjectMap implements QueueElementIF {
    public ObjectMap (Map backpointers) {
      this.objmap = backpointers;
    }

    public Map objmap;
  }

    private String dynamic_route;
    private NodeId _self_node_id;
    private SecureHash _self_guid;
    private int [] _self_digits;
    private SinkIF _sink;
    private LinkedList _waiting_reqs;
    private Map _backpointers;
    private QSPublicKey _pkey;

    private Map _network_latencies;
    private Map _pings_in_flight;
    private Map _pending_back_pointer_times;

    private RoutingTable _routing_tbl;
    private Map _msgs_needing_macs;
    private boolean _using_bloom;
}

