/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * PriorityQueue is a (blocking) priority queue. One of these
 * will be used to store updates (and it will be read by the update thread)
 * and the other stores query/success/fail messages for the location
 * algorithm thread. Objects with small priorities come out first.
 *
 * @author     Westley Weimer
 * @version    $Id: PriorityQueue.java,v 1.6 2001/10/18 00:20:03 eaton Exp $
 */
public class PriorityQueue 
{
    public static final boolean DEBUG = ostore.util.DebugFlags.PRIORITY_QUEUE;

    private Object[] _heap;	// the vector in which we embed the heap
    private long[] _prio;	// priorities
    private int _capacity;	// maximum size
    private int _heap_size;	// current heap size


    /**
     * The constructor creates a new priority queue.
     * @param	initial_capacity	the initial capacity of the heap.
     */
    public PriorityQueue(int initial_capacity) {
        _capacity = initial_capacity + 1; // compensate for skipping index 0.
	_heap = new Object[_capacity];
	_prio = new long[_capacity];
	_heap_size = 0;
    }

    /**
     * returns the number of objects currently in the queue (0 means empty).
     */
    public synchronized int size() {
	return _heap_size;
    }

    /**
     * returns the number of objects currently in the queue (0 means empty).
     */
    public synchronized boolean empty () {
	return _heap_size == 0;
    }

    /**
     * returns the maximum size of the heap. 
     */
    public synchronized int maxSize() {
	return _capacity;
    }
    /**
     * returns the current capacity of the heap.
     */
    public synchronized int capacity() {
      return _capacity-1;	// correct for skipping index 0.
    }

    /**
     * inserts an object with the given priority. If the queue is full,
     * the internal storage arrays are doubled in a very naive way.
     * @param blocking	ignored.
     */
    public synchronized boolean insert(Object o, long prio, boolean blocking) {
// 	if (blocking) {
// 	    while (_heap_size >= _capacity - 1) {
// 	      myWait();
// 	    }
// 	} else if (_heap_size >= _capacity - 1) {
// 	    return false;
// 	}
      if( _heap_size >= _capacity - 1 ) {
	if (DEBUG) Debug.printtagln("PriorityQueue", 
			 "doubling capacity...");
	int new_capacity = (2*_capacity)-1;	// remember array is 1-based.
	Object[] new_heap = new Object[new_capacity];
	long[] new_prio   = new long[new_capacity];
	for( int i=0; i<_capacity; ++i ) {
	  new_heap[i] = _heap[i];
	  new_prio[i] = _prio[i];
	}
	_heap = new_heap;
	_prio = new_prio;
	_capacity = new_capacity;
      }
	
        _heap_size ++;
	int i = _heap_size;
	while (i > 1 && _prio[i/2] > prio) {
	    _heap[i] = _heap[i/2];
	    _prio[i] = _prio[i/2];
	    i = i/2;
	}
	_heap[i] = o;
	_prio[i] = prio;
	if (DEBUG) Debug.println("PriorityQueue", 
	    "PriorityQueue: inserted obj with key " + prio);
	if (DEBUG) Debug.println("PriorityQueue", 
	    "PriorityQueue: size is now " + _heap_size);
	notifyAll();
	return true;
    }

    /**
     * removes the given object from the priority queue. returns true if
     * the object was found and removed, false otherwise. 
     */
    public synchronized boolean remove(Object o) {
	int i;

	if (DEBUG) Debug.printtagln("PriorityQueue", 
		      "remove, size = " + _heap_size);

	for (i=1; i<=_heap_size; i++) {
	    if (DEBUG) Debug.printtagln("PriorityQueue", 
			     "remove, looking for " + o);
	    if (DEBUG) Debug.printtagln("PriorityQueue", "comparing to " + _heap[i]);
	    
	    if ((_heap[i]).equals(o)) {
		// remove it!
		_heap[i] = _heap[_heap_size];
		_heap[_heap_size] = null;
		_prio[i] = _prio[_heap_size];
		_heap_size --;

		// Should be heapify(i)?
                // Was heapify(1)
		heapify(i);
		if (DEBUG) Debug.println("PriorityQueue", 
			"PriorityQueue: removed " + o);
		notifyAll();
		return true;
	    }
	}
	return false;
    }


    /**
     * removes every instance the given object from the priority queue.
     * returns true if the object was found and removed, false otherwise. 
     */
    public synchronized boolean removeAll(Object o) {
	int i;
	boolean ret = false;

	if (DEBUG) Debug.printtagln("PriorityQueue", 
		      "remove, size = " + _heap_size);

	for (i=1; i<=_heap_size; i++) {
	    if (DEBUG) Debug.printtagln("PriorityQueue", 
			     "remove, looking for " + o);
	    if (DEBUG) Debug.printtagln("PriorityQueue", "comparing to " + _heap[i]);
	    
	    if ((_heap[i]).equals(o)) {
		// remove it!
		_heap[i] = _heap[_heap_size];
		_heap[_heap_size] = null;
		_prio[i] = _prio[_heap_size];
		_heap_size --;

		// Should be heapify(i)?
                // Was heapify(1)
		heapify(i);
		if (DEBUG) Debug.println("PriorityQueue", 
			"PriorityQueue: removed " + o);
		ret = true;

		// Be sure to check the object we replaced
		// the removed object with
		i--;
	    }
	}
	notifyAll();
	return ret;
    }


    /**
     * topPriority() returns the smallest priority in the queue, or 0 if
     * the queue is empty and blocking is not set.
     */
    public synchronized long topPriority(boolean blocking) {
	if (blocking)
	    while (_heap_size < 1) {
	      myWait();
	    }
	else if (_heap_size < 1)
	    return 0;
	return _prio[1];
    }

    /**
     * topObject() returns the smallest priority object in the queue, or 
     * null if the queue is empty and blocking is not set.
     */
    public synchronized Object topObject(boolean blocking) {
	if (blocking)
	    while (_heap_size < 1) {
	      myWait();
	    }
	else if (_heap_size < 1)
	    return null;
	return _heap[1];
    }

    /**
     * extract() returns the object with the smallest priority in the
     * queue and removes it from the queue. If the queue is empty and
     * blocking is set, this operation blocks. If the queue is empty and
     * blocking is not set, null is returned.
     */
    public synchronized Object extract(boolean blocking) {
	if (blocking) 
	    while (_heap_size < 1) {
	      myWait();
	    }
	else if (_heap_size < 1) 
	    return null;
	Object max = _heap[1];
	long prio = _prio[1];
	_heap[1] = _heap[_heap_size];
	_heap[_heap_size] = null;
	_prio[1] = _prio[_heap_size];
	_heap_size --;

	heapify(1);
	if (DEBUG) Debug.println("PriorityQueue", 
	    "PriorityQueue: extracted obj with key " + prio);
	notifyAll();
	return max;
    }

    /**
     * heapify (aka "fix-down") maintains the heap invariant after an
     * extraction.
     */
    private synchronized void heapify(int i) {
	int l, r, largest;

	while (true) {
	    l = i * 2;
	    r = i * 2 + 1;
	    if (l <= _heap_size && _prio[l] < _prio[i]) 
		largest = l;
	    else
		largest = i;
	    if (r <= _heap_size && _prio[r] < _prio[largest])
		largest = r;
	    if (largest != i) {
		Object temp_o;
		long temp_p;
		// swap i and largest
		temp_o = _heap[i];
		temp_p = _prio[i];
		_heap[i] = _heap[largest];
		_prio[i] = _prio[largest];
		_heap[largest] = temp_o;
		_prio[largest] = temp_p;

		i = largest;
	    } else return;
	}
    }
    /**
     * simple wrapper for Object.wait().
     */
    private final void myWait () {
      if (DEBUG) Debug.printtagln( "Wait", "SLEEP on PriorityQueue - "
			+ Thread.currentThread() );
      try { wait (); } catch (InterruptedException e) {}
      if (DEBUG) Debug.printtagln( "Wait", "WAKING UP! on PriorityQueue - "
			+ Thread.currentThread() );
    }
}
/* eof: PriorityQueue.java */

