/**
 *  QSWindow.java
 *
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * <p>The <code>QSWindow</code> object is a data structure similiar to
 * a queue.  It can store up to a fixed amount of 
 * <code>QuickSerializable</code> objects, and once it contains that
 * many objects, it begins overwriting the least recent object
 * added to it.  In this way, it represents a kind of window of
 * objects in time that it can hold.</p><p>
 * There is probably a lot of functionality that can be added to this 
 * object to make it more useful. Please feel free to add to it
 * as necessary.</p>
 *
 * @author      Jeremy Stribling
 * @version     $Id: QSWindow.java,v 1.1 2001/12/04 19:45:13 strib Exp $
 */
public class QSWindow implements QuickSerializable
{
    ////////////////////       CONSTANTS	////////////////////

    public static final int DEFAULT_SIZE = 1000;

    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    private QSVector data;
    private int index;
    private int windowSize;

    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Construct a <code>QSWindow</code> object with the given maximum 
     * capacity.
     *
     * @param windowSize   the maximum capacity of this window
     */
    public QSWindow( int windowSize ) {
	this.windowSize = windowSize;
	index = 0;
	data = new QSVector();
    }

    /**
     * Construct a <code>QSWindow</code> with a default maximum capacity.
     * Currently this default is 1000.
     */
    public QSWindow( ) {
	this( DEFAULT_SIZE );
    }

    /**
     * Constructs a <code>QSWindow</code> from its
     * <code>QuickSerializable</code> form.
     */
    public QSWindow( byte[] data, int[] offset ) throws QSException
    {
 	windowSize = ByteUtils.bytesToInt( data, offset );
 	index = ByteUtils.bytesToInt( data, offset );
	this.data = new QSVector( data, offset );
    }
    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    //  methods for ostore.util.QuickSerializable interface

    /**
     * Specified by the <code>QuickSerializable</code> interface.
     */
    public void to_bytes(byte [] data, int[] offset)
    {
	ByteUtils.intToBytes( windowSize, data, offset );
	ByteUtils.intToBytes( index, data, offset );
	(this.data).to_bytes( data, offset );
    }

    /**
     * Specified by the <code>QuickSerializable</code> interface.
     */
    public int type_code() 
    {
        return ostore.util.Types.QSWINDOW;
    }

    /**
     * Add this object to the window.  If the window is at maximum
     * capacity, overwrite the least recently added object.
     *
     * @param datum  the object to add
     */
    public void add( QuickSerializable datum ) {

        data.ensureCapacity( index + 1 );
        if( data.size() <= index ) {
          data.add( index, datum );
        } else {
          data.set( index, datum );
        }
        index++;
                                      
        // if we've filled up our data, start again at
        // the beginning
        if( index >= windowSize ) {
            index = 0;
        }
    }

    /**
     * Forget about a certain number of the least-recently added
     * data.
     *
     * @param num   the number of elements to forget
     */
    public synchronized void clear( int num )
    {
        
        int recentSize = data.size();
        int startClear;
        int endClear;
        
        // if we're clearing more than we have, just
        // clear 'em all
        if( num >= recentSize ) {
            clear();
            return;
        }
     
        // start clearing right after the current index, or
        // start clearing from the beginning if the index is
        // at the end.
        if( index < recentSize ) {
            startClear = index;
        } else {
            startClear = 0;
            index = index - num;
        }
     
        // end clearing at startClear + num, or if that's too big,
        // clear to the end and then clear as much from the beginning
        // as needed  
        endClear = startClear + num;
        if( endClear > recentSize ) {
            endClear = recentSize;
        
            data.removeRange( startClear, endClear );  
            int numCleared = endClear - startClear;
            startClear = 0;
            endClear = num - numCleared;
            index = index - (num - numCleared);
        }
            
        data.removeRange( startClear, endClear );

    }

    /**
     * Forget about all elements currently in the window.
     */
    public synchronized void clear()
    {
        data.removeAllElements();
        index = 0;
    } 

    /**
     * Convert the elements in this window to an array.  No order of the
     * elements is guaranteed within this array.
     */
    public QuickSerializable[] toArray() {
	return data.toArray();
    }

}
