/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * SHA1Verify is the class which stores a list of SHA1Hashes
 * which sould be serially concatenated and hashed to produce
 * a GUID or other verifier.
 *
 * @author	Chris WElls
 * @version	$Id: SHA1Verify.java,v 1.5 2002/04/04 05:02:48 hweather Exp $
 */
public class SHA1Verify implements SecureVerify {

	public static final int TYPE_CODE = Types.SHA1_VERIFY;

	/**
	 * This is the array of SHA1Hashes
	 */
	protected SHA1Hash[] _hashes;

	/**
	 * Default constructor
	 */
	public SHA1Verify() {
		_hashes = new SHA1Hash[0];
	}

	/**
	 * Read the bytes of the value of this verifier in from the input
	 * byte array
	 * @param array out of which to read the bytes of this verifier
	 * @param offset in data at which to begin reading
	 * @return number of bytes read (always NUM_BYTES)
	 */
        public SHA1Verify (byte [] data, int [] offset) {
		int length = ByteUtils.bytesToInt(data, offset);
		_hashes = new SHA1Hash [length];
		for (int i = 0; i < _hashes.length; ++i)
			_hashes[i] = new SHA1Hash (data, offset);
	}

	/**
	 * Initializing constructor
	 * @param size is the number of hashes to store in this verifier
	 */
	public SHA1Verify(int size) {
		_hashes = new SHA1Hash[size];
		/**
		 * Add by Hakim Weatherspoon 07/02/2001
		for(i = 0; i < size; i++)
		    _verify[i] = new SHA1Hash();
		*/
	}

	/**
	 * Get the i^th hash in this verifier
	 * @param i the index of the hash to get
	 * @return the i^th hash
	 */
	public SHA1Hash getHash(int i) {
		return _hashes[i];
	}

	/**
	 * Set the i^th hash in this verifier
	 * @param i the index of the hash to set
	 * @param hash the value of the hash
	 */
	public void setHash(int i, SHA1Hash hash) {
		_hashes[i] = hash;
	}

	/**
	 * This routine returns the array of hashes
	 * @return the array of hashes
	 */
	public SHA1Hash [] getArray() {
		return _hashes;
	}

	/**
	 * Return the length of the verification array
	 * @return hash array length
	 */
	public int length() {
		return _hashes.length;
	}

	/**
	 * Create a new SHA1Verify with the values of this one
	 * @param verify the preexisting value of a SHA1Verify
	 */
	public Object clone() throws CloneNotSupportedException {
		return super.clone();
	}

	/**
	 * Write the bytes of this verifier to the input byte array
	 * @param array into which to write the bytes of this verifier
	 * @param offset in data at which to begin writing
	 * @return number of bytes written
	 */
        public void to_bytes (byte [] data, int [] offset) {
		ByteUtils.intToBytes (_hashes.length, data, offset);
		for (int i = 0; i < _hashes.length; ++i) {
			_hashes[i].to_bytes (data, offset);
		}
        }

        /** Specified by ostore.util.Digestible */
        public void add_to_digest(SecureHashBuffer buffer){
		buffer.add (_hashes.length);
		for (int i = 0; i < _hashes.length; ++i) {
			buffer.add(_hashes[i]);
		}
	}

        /**
	 * Specified by ostore.util.QuickSerializable 
	 * @deprecated */
        public int type_code() {
	        throw new NoSuchMethodError();
		//return TYPE_CODE;
	}
}
