/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import ostore.security.*;
import ostore.util.*;

/**
 * A class to store the compile-time constants for the types of each
 * subtype of QuickSerializable in the system, and the mapping between
 * these two.
 * 
 * <p> The use of this class by programmers other than the author should be
 * only to add new type constants, and to add elements to the {@link
 * #types} array.  For instance, let's say you've added a new class,
 * <code>ostore.foo.Bar</code> to the system.  First, you would add a type
 * code for that class to this one, adding a member like this:
 *
 * <pre>static public final int FOO_BAR = 0xdeadbeef;</pre>
 *
 * Next, you would make sure that the member function
 * <code>type_code</code> of <code>ostore.foo.Bar</code> was written like
 * this:
 *
 * <pre>public int type_code () { return ostore.util.Types.FOO_BAR; }</pre>
 *
 * Finally, you would add a new element to the <code>types</code> array
 * containing the mapping from your new Type to the type code, like this:
 *
 * <pre>new Type ("ostore.foo.Bar", FOO_BAR),</pre>
 *
 * And then you would be done.  There is no need to call {@link
 * TypeTable#register_type} yourself, it will be done by {@link
 * #registerAllTypes} (which these days is called by the classifier), using
 * the information you provided in the <code>types</code> array.
 * 
 * @author   Sean C. Rhea
 * @version  "$Id: Types.java,v 1.150 2002/04/06 18:30:36 hweather Exp $"
 * 
 * @see ostore.util.TypeTable
 * @see ostore.util.QuickSerializable
 * @see ostore.util.QSException
 */
public class Types
{
    // Stage IDs
    // TODO: move these to ostore.util.StageId
    static public final int TEST_GENERATE_FRAGS_BLK        = 0xf005;
    static public final int GENERATE_LOG_STAGE             = 0xf006;
    static public final int RECONSTRUCT_OBJ_STAGE          = 0xf00A;

    // Put type codes here.

    static public final int QSINT                          = 0x0006;
    static public final int QSLONG                         = 0x0007;
    static public final int QSSTRING                       = 0x0008;
    static public final int QSVECTOR                       = 0x0009;
    static public final int QSDATE                         = 0x000A;
    static public final int QSBYTEARRAY			   = 0x000B;
    static public final int QSARRAY			   = 0x000C;
    static public final int QSBOOL			   = 0x000D;
    static public final int QSWINDOW			   = 0x000E;

    static public final int SHA1_HASH                      = 0x1002;
    static public final int SHA1_VERIFY                    = 0x1003;
    static public final int TIMESTAMP                      = 0x1004;
    static public final int NODEID                         = 0x1005;
    static public final int VERIFIABLE_BLOCK_TAG           = 0x1007;

    static public final int VID				= 0x1008;

    static public final int NETWORK_IDENTITY_MESSAGE       = 0x2000;
    static public final int DUMP_STATS_MSG       = 0x2001;
    static public final int RESET_STATS_MSG       = 0x2002;

    static public final int STARTREK_ROUTE_MESSAGE         = 0x3000;
    static public final int STARTREK_PUBLISH_MESSAGE       = 0x3001;
    static public final int STARTREK_UNPUBLISH_MESSAGE     = 0x3002;
    static public final int STARTREK_TEST_TAPESTRY_MESSAGE = 0x3003;
    static public final int STARTREK_GENERICTAPESTRYROUTEMESSAGE = 0x3103;
    static public final int STARTREK_GENERICTAPESTRYROUTEOBJMESSAGE = 0x3104;

    /*static public final int STARTREK_DISINTEGRATETMESSAGE        = 0x3100;
    static public final int STARTREK_GENERICOPERATIONFAILTERROR  = 0x3101;
    static public final int STARTREK_GENERICROUTETERROR          = 0x3102;
    static public final int STARTREK_GENERICTAPESTRYROUTEMESSAGE = 0x3103;
    static public final int STARTREK_INTEGRATETMESSAGE           = 0x3104;
    static public final int STARTREK_MOVEOBJTMESSAGE             = 0x3105;
    static public final int STARTREK_NODENOTFOUNDTERROR          = 0x3106;
    static public final int STARTREK_OBJECTNOTFOUNDTERROR        = 0x3107;
    static public final int STARTREK_PUBLISHTMESSAGE             = 0x3108;
    static public final int STARTREK_ROUTEOBJTMESSAGE            = 0x3109;
    static public final int STARTREK_ROUTEPARENTTMESSAGE         = 0x3110;
    static public final int STARTREK_ROUTETMESSAGE               = 0x3111;*/

    static public final int TAPESTRY_DISINTEGRATETMESSAGE  = 0x4000;
    static public final int TAPESTRY_INTEGRATETMESSAGE     = 0x4001;
    static public final int TAPESTRY_MOVEOBJTMESSAGE       = 0x4002;
    static public final int TAPESTRY_PUBLISHTMESSAGE       = 0x4003;
    static public final int TAPESTRY_ROUTEOBJTMESSAGE      = 0x4004;
    static public final int TAPESTRY_ROUTEPARENTTMESSAGE   = 0x4005;
    static public final int TAPESTRY_ROUTETMESSAGE         = 0x4006;
    static public final int TAPESTRY_PING_MSG              = 0x4007;
    static public final int TAPESTRY_PONG_MSG              = 0x4008;
    static public final int TAPESTRY_LOCATE_FAILURE        = 0x4009;

    static public final int TAPESTRY_STATIC_ALIVE_MSG      = 0x4100;
    static public final int TAPESTRY_STATIC_ALL_NODES_MSG  = 0x4101;
    static public final int TAPESTRY_STATIC_READY_MSG      = 0x4102;
    static public final int TAPESTRY_STATIC_BEGIN_MSG      = 0x4103;
    static public final int TAPESTRY_STATIC_PING_MSG       = 0x4104;
    static public final int TAPESTRY_STATIC_PONG_MSG       = 0x4105;
    // to test new type code stuff --srhea
    // static public final int TAPESTRY_ROUTE_MSG             = 0x4106;
    static public final int TAPESTRY_ROUTE_OBJ_ROOT_MSG    = 0x4107;
    static public final int TAPESTRY_ROUTE_OBJ_DEST_MSG    = 0x4108;
    static public final int TAPESTRY_PUBLISH_MSG           = 0x4109;
    static public final int TAPESTRY_BLOOM_WRAPPER         = 0x410a;

  static public final int TAPESTRY_DYNAMIC_REQ_MSG       = 0x4200;
  static public final int TAPESTRY_DYNAMIC_MAP_MSG       = 0x4201;
  static public final int TAPESTRY_DYNAMIC_NOT_MSG       = 0x4202;
  static public final int TAPESTRY_DYNAMIC_MCAST_MSG     = 0x4203;
  static public final int TAPESTRY_DYNAMIC_OBJMAP_MSG    = 0x4204;
  static public final int TAPESTRY_DYNAMIC_OBJREQ_MSG    = 0x4205;
  static public final int TAPESTRY_DYNAMIC_SURRREQ_MSG   = 0x4206;
  static public final int TAPESTRY_DYNAMIC_SURR_MSG      = 0x4207;
  static public final int TAPESTRY_DYNAMIC_PING_MSG      = 0x4208;
  static public final int TAPESTRY_DYNAMIC_PONG_MSG      = 0x4209;
  static public final int TAPESTRY_DYNAMIC_GATEWAY_MSG   = 0x4210;

  static public final int TAPESTRY_PUBLISH_INFO          = 0x4300;
  static public final int TAPESTRY_TEST_READY_MSG        = 0x4301;
  static public final int TAPESTRY_TEST_FOUND_MSG        = 0x4302;
  static public final int TAPESTRY_TEST_LOCATE_MSG       = 0x4303;

    static public final int DATAOBJECT                                = 0x2506;
    static public final int METADATA                                  = 0x2508;

    static public final int VERSION_INFO                              = 0x2507;
    static public final int DATAOBJECT_DATABLOCK                      = 0x2515;
    static public final int BTREE                                     = 0X2505;
    static public final int BTREE_INTERIOR_NODE                       = 0x2501;
    static public final int LONG_KEY                                  = 0x2585;
    static public final int SHA1HASH_KEY                              = 0x2586;

    static public final int SELECTION_ALL			= 0x2601;
    static public final int SELECTION_RANGE			= 0x2602;
    static public final int SELECTION_KEYED			= 0x2603;
    static public final int SELECTION_COMPOSITE			= 0x2604;
    static public final int SELECTION_NULL                      = 0x2605;

    static public final int RPCERT                          = 0x2777;
    static public final int PRCERT                          = 0x2778;
    static public final int ACL                             = 0x2780;

    static public final int UPDATE                                    = 0x7000;
    static public final int UPDATE_TUPLE                              = 0x7001;

    // TODO : remove these two types:
    static public final int READ_REQUEST                              = 0x7004;
    static public final int READ_RESULT                               = 0x7005;

    static public final int BYTES_READ_REQ			= 0x7010;
    static public final int BYTES_READ_RESP			= 0x7011;
    static public final int SLICE_READ_REQ			= 0x7012;
    static public final int SLICE_READ_RESP			= 0x7013;
    static public final int REPLICA_BYTES_READ_REQ		= 0x7020;
    static public final int REPLICA_BYTES_READ_RESP		= 0x7021;
    static public final int REPLICA_SLICE_READ_REQ		= 0x7022;
    static public final int REPLICA_SLICE_READ_RESP		= 0x7023;

    static public final int REMOTE_CACHE_REQUEST_MESSAGE        = 0x7050;
    static public final int REMOTE_CACHE_REPLY_MESSAGE          = 0x7051;

    static public final int DATE_PRED				= 0x7100;
    static public final int FRESH_PRED				= 0x7101;
    static public final int HEAD_PRED				= 0x7102;
    static public final int NUMBER_PRED				= 0x7103;
    static public final int VID_PRED				= 0x7104;
    
    static public final int ACTION                         = 0x7500;
    static public final int ACTION_NULL                    = 0x7501;
    static public final int ACTION_REPLACE_BLOCKS          = 0x7502;
    static public final int ACTION_APPEND_BLOCKS           = 0x7503;
    static public final int ACTION_SET_METADATA            = 0x7504;
    static public final int ACTION_TRUNCATE                = 0x7505;
    static public final int ACTION_EXPAND                  = 0x7506;

    static public final int PREDICATE                      = 0x7600;
    static public final int PREDICATE_TRUE                 = 0x7601;

    static public final int UPDATE_INTERRUPTED_EXCEPTION   = 0x7802;

    static public final int READ_INTERRUPTED_EXCEPTION     = 0x9800;

    static public final int ARRAY			   = 0xA000;
    static public final int MATRIX			   = 0xA001;
    static public final int SEGMENT_REQUEST		   = 0xA011;
    static public final int SEGMENT_REV_REQUEST		   = 0xA012;
    static public final int HMM			   	   = 0xA100;
    static public final int HMM_DELTA			   = 0xA101;
    static public final int HMM_PREDICTION		   = 0xA102;
    static public final int HMM_SEGMENT			   = 0xA103;
    static public final int SCREEN			   = 0xA110;

    // patchwork types:
    static public final int OIL_PATCHWORK_MODEL		   = 0xA200;
    static public final int OIL_PATCHWORK_DELTA		   = 0xA201;
    static public final int OIL_PATCHWORK_OBSERVATION	   = 0xA202;
    static public final int OIL_PATCHWORK_PROBE_MSG	   = 0xA203;


    static public final int TESTBLOCK                      = 0xC000;

    static public final int SECURITY_RSA_PUBLIC_KEY        = 0x11000;
    static public final int SECURITY_RSA_PRIVATE_KEY       = 0x11001;
    static public final int SECURITY_SHA1_RSA_SIGNED_QS    = 0x11002;
    static public final int SECURITY_QS_SHA1_RSA_SIGNATURE = 0x11003;
    static public final int SECURITY_QS_RSA_PUBLIC_KEY     = 0x11004;
    static public final int SECURITY_QS_RSA_PRIVATE_KEY    = 0x11005;
    static public final int SECURITY_SIGNED_QS             = 0x11006;
    static public final int SECURITY_QS_DH_PUBLIC_KEY      = 0x11007;

    static public final int SECURITY_THRESH_PUBLIC_KEY     = 0x11008;
    static public final int SECURITY_THRESH_SIGNATURE      = 0x11009;
    static public final int SECURITY_THRESH_KEY_SHARE      = 0x1100A;
    static public final int SECURITY_SIGNED_QS_ARRAY_IND   = 0x1100B;

    static public final int SECURITY_ENCRYPTED_QS          = 0x1100C;
    static public final int SECURITY_QS_RSA_CIPHER         = 0x1100D;

    static public final int RP_CREATE_OBJECT_REQ           = 0x12000;
    static public final int RP_CREATE_OBJECT_REQ_MSG       = 0x12001;
    static public final int RP_INNER_RING_CREATE_REQ_MSG   = 0x12003;

    static public final int INNER_CREATE_OBJECT_RESP       = 0x13000;
    //static public final int INNER_CREATE_OBJECT_RESP_MSG   = 0x13001;
    static public final int INNER_HEARTBEAT                = 0x13002;
    static public final int INNER_UPDATE_REQ_MSG           = 0x13003;
    static public final int INNER_UPDATE_RESP              = 0x13004;
    static public final int INNER_UPDATE_RESP_MSG          = 0x13005;
    static public final int INNER_LATEST_HEARTBEAT_REQ     = 0x13006;
    static public final int INNER_LATEST_HEARTBEAT_REQ_MSG = 0x13007;
    static public final int INNER_LATEST_HEARTBEAT_RESP    = 0x13008;
    static public final int INNER_LATEST_HEARTBEAT_RESP_MSG= 0x13009;
    //static public final int INNER_TIMED_HEARTBEAT_REQ_MSG  = 0x13010;
    static public final int INNER_TIMED_HEARTBEAT_RESP_MSG = 0x13011;
    static public final int INNER_BLOCK_READ_REQ_MSG       = 0x13012;
    static public final int INNER_BLOCK_READ_RESP_MSG      = 0x13013;
    static public final int INNER_CREATE_OBJECT_REJECT     = 0x13014;
    static public final int INNER_UPDATE_REJECT            = 0x13015;
    static public final int INNER_LATEST_HEARTBEAT_REJECT  = 0x13016;
    static public final int INNER_INNER_RING_QUERY         = 0x13017;
    static public final int INNER_INNER_RING_TAG           = 0x13018;
    static public final int INNER_VGUID_GENERATION_REQ     = 0x13019;
    static public final int INNER_VGUID_GENERATION_REJECT  = 0x1301A;

    static public final int AGREE_FAKE_BYZ_PREPARE_MSG     = 0x14000;
    static public final int AGREE_FAKE_BYZ_RESULT_MSG      = 0x14001;
    static public final int AGREE_FAKE_BYZ_COMMIT_MSG      = 0x14002;
    static public final int AGREE_FAKE_BYZ_FORWARD_MSG     = 0x14003;

    static public final int BLOOM_UPDATE_ID                = 0x15000;
    static public final int BLOOM_UPDATE_MESSAGE           = 0x15001;
    static public final int BLOOM_TOPOLOGY_DISCOVER_QUERY  = 0x15002;
    static public final int BLOOM_TOPOLOGY_DISCOVER_RESPONSE = 0x15003;
    static public final int BLOOM_SIMULATED_READ           = 0x15004;
    static public final int BLOOM_SIMULATED_READ_RESP      = 0x15005;

    static public final int MAC_DHMESSAGE                  = 0xC1FF0;
    static public final int MAC_STSVERIFY                  = 0xC1FF1;
    static public final int MAC_MACEDQS                    = 0xC1FF2;
    static public final int MAC_MACMESSAGE                 = 0xC1FF3;
    static public final int MACTYPE_HMACMD5                = 0xC2FF0;

    static public final int CACHEABLE_IDENTITY             = 0xD0000;

    static public final int GENERATE_FRAGS_BLK_REQ_CONTINUATION    = 0x5015;
    static public final int GENERATE_FRAGS_BLK_REQ                 = 0x5000;
    static public final int GENERATE_FRAGS_BLK_RESP                = 0x5001;
    static public final int GENERATE_FRAGS_CHKPT_REQ               = 0x5002;
    static public final int GENERATE_FRAGS_CHKPT_RESP              = 0x5003;
    static public final int GENERATE_FRAGS_LOG_REQ                 = 0x5004;
    static public final int GENERATE_FRAGS_LOG_RESP                = 0x5005;
    static public final int DISSEMINATE_FRAGS_REQ                  = 0x5006;
    static public final int DISSEMINATE_FRAGS_RESP                 = 0x5007;
    static public final int DISSEMINATE_TOMBSTONE_REQ              = 0x5008;
    static public final int DISSEMINATE_TOMBSTONE_RESP             = 0x5009;
    static public final int RECONSTRUCT_BLK_REQ                    = 0x500C;
    static public final int RECONSTRUCT_BLK_RESP                   = 0x500D;
    static public final int RECOALESCE_BLK_REQ                     = 0x500E;
    static public final int RESURRECT_TOMB_REQ                     = 0x500F;
    static public final int RESURRECT_TOMB_RESP                    = 0x5010;
    static public final int QUERY_REQ                              = 0x5011;
    static public final int QUERY_RESP                             = 0x5012;
    static public final int FRAGMENT                               = 0x5013;
    static public final int FRAGMENT_BUCKET                        = 0x5014;
    static public final int FRAGMENT_TAG                           = 0x5016;
    static public final int REQUEST_FRAG_MSG                       = 0x5017;
    static public final int REQUEST_FRAG_QUERY                     = 0x5018;
    static public final int REQUEST_FRAG_QUERY_STATE               = 0x5019;
    static public final int REQUEST_FRAG_RESPONSE_MSG              = 0x501A;
    static public final int DISSEMINATE_FRAG_MSG                   = 0x501B;
    static public final int DISSEMINATE_FRAG_ACK_MSG               = 0x501C;
    static public final int TYPE_REQUEST_MSG                       = 0x501D;
    static public final int TYPE_RESPONSE_MSG                      = 0x501E;
    static public final int DISSEMINATION_SET_CREATOR_TAG          = 0x501F;
    static public final int REQUEST_DISSEMINATION_SET_MSG          = 0x5020;
    static public final int REQUEST_DISSEMINATION_SET_QUERY        = 0x5021;
    static public final int REQUEST_DISSEMINATION_SET_RESPONSE_MSG = 0x5022;
    static public final int BUCKETS_TO_DISSEMINATE_EVENT           = 0x5023;
    static public final int GUID_RESULT                            = 0x5024;

    static public final int DTREE_MSG				= 0x8000;
    static public final int DTREE_LOCATE_MSG			= 0x8001;
    static public final int DTREE_CREATE_REQ			= 0x8011;
    static public final int DTREE_CREATE_RESP			= 0x8012;
    static public final int DTREE_JOIN_REQ			= 0x8013;
    static public final int DTREE_JOIN_RESP			= 0x8014;
    static public final int DTREE_LEAVE_REQ			= 0x8015;
    static public final int DTREE_LEAVE_RESP			= 0x8016;
    static public final int DTREE_DISCONNECT_REQ		= 0x8017;
    static public final int DTREE_DISCONNECT_RESP		= 0x8018;
    
    static public final int REPLICA_TAG				= 0x6000;
    static public final int REPLICA_QUERY			= 0x6001;
    static public final int REPLICA_OPEN_REQ			= 0x6010;
    static public final int REPLICA_OPEN_RESP			= 0x6011;
    static public final int REPLICA_CLOSE_REQ			= 0x6012;
    static public final int REPLICA_CLOSE_RESP			= 0x6013;

     // Associate type codes with class names here.

    static public final Type [] types = {

 	//  ostore.util types
	new Type ("ostore.util.QSInt", QSINT),
	new Type ("ostore.util.QSLong", QSLONG),
	new Type ("ostore.util.QSString", QSSTRING),
	new Type ("ostore.util.QSDate", QSDATE),
	new Type ("ostore.util.QSByteArray", QSBYTEARRAY),
	new Type ("ostore.util.QSArray", QSARRAY),
        new Type("ostore.util.QSBool", QSBOOL),
        new Type("ostore.util.QSWindow", QSWINDOW),

	new Type ("ostore.util.SHA1Hash", SHA1_HASH),
	new Type ("ostore.util.SHA1Verify", SHA1_VERIFY),
	new Type ("ostore.util.Timestamp", TIMESTAMP),
	new Type ("ostore.util.NodeId", NODEID),
//	new Type ("ostore.util.GuidPair", GUIDPAIR),
//	new Type ("ostore.util.VerifiableBlockTag", VERIFIABLE_BLOCK_TAG),

//	new Type ("ostore.util.VID", VID),

	// ostore.network types
	new Type ("ostore.network.IdentityMessage",
		  NETWORK_IDENTITY_MESSAGE),
        /*
	new Type ("ostore.network.DumpStatsMsg",
		  DUMP_STATS_MSG),
	new Type ("ostore.network.ResetStatsMsg",
		  RESET_STATS_MSG),
        */

	// ostore.startrek types
	/*
	new Type ("ostore.startrek.PublishMessage",
		  STARTREK_PUBLISH_MESSAGE),
	new Type ("ostore.startrek.UnPublishMessage",
		  STARTREK_UNPUBLISH_MESSAGE),
	new Type ("ostore.startrek.RouteMessage",
		  STARTREK_ROUTE_MESSAGE),
	new Type ("ostore.startrek.GenericTapestryRouteMessage",
		  STARTREK_GENERICTAPESTRYROUTEMESSAGE),
	new Type ("ostore.startrek.GenericTapestryRouteObjMessage",
		  STARTREK_GENERICTAPESTRYROUTEOBJMESSAGE),
	*/

	// ostore.cache types (TODO - moved up temporarily)
/*        new Type("ostore.cache.TestBlock", 
		 TESTBLOCK),
	new Type("ostore.cache.CacheableIdentity",
		 CACHEABLE_IDENTITY),
        new Type("ostore.cache.RemoteCacheRequestMessage",
                 REMOTE_CACHE_REQUEST_MESSAGE),
        new Type("ostore.cache.RemoteCacheReplyMessage",
                 REMOTE_CACHE_REPLY_MESSAGE),

	//  ostore.dataobj types
        new Type("ostore.dataobj.Acl",
                 ACL),
	new Type("ostore.dataobj.Btree",
                 BTREE),
	new Type("ostore.dataobj.Btree$InteriorNode",
	         BTREE_INTERIOR_NODE),
        new Type("ostore.dataobj.DataObject",
                 DATAOBJECT),
	new Type("ostore.dataobj.DataObject$DataBlock", 
	         DATAOBJECT_DATABLOCK),
        new Type("ostore.dataobj.LongKey",
                 LONG_KEY),
	new Type("ostore.dataobj.Metadata",
                 METADATA),
	new Type("ostore.dataobj.PRCert",
                 PRCERT),
	new Type("ostore.dataobj.RPCert",
                 RPCERT),
        new Type("ostore.dataobj.SHA1HashKey",
                 SHA1HASH_KEY),
        new Type("ostore.dataobj.All",
                 SELECTION_ALL),
        new Type("ostore.dataobj.Range",
                 SELECTION_RANGE),
        new Type("ostore.dataobj.KeyedSelection",
                 SELECTION_KEYED),
        new Type("ostore.dataobj.CompositeSelection",
                 SELECTION_COMPOSITE),
        new Type("ostore.dataobj.NullSelection",
                 SELECTION_NULL),
*/

	// ostore.tapestry.impl types

	new Type("ostore.tapestry.impl.StaticAliveMsg", 
		TAPESTRY_STATIC_ALIVE_MSG),
	new Type("ostore.tapestry.impl.StaticAllNodesMsg", 
		TAPESTRY_STATIC_ALL_NODES_MSG),
	new Type("ostore.tapestry.impl.StaticReadyMsg", 
		TAPESTRY_STATIC_READY_MSG),
	new Type("ostore.tapestry.impl.StaticBeginMsg", 
		TAPESTRY_STATIC_BEGIN_MSG),
	new Type("ostore.tapestry.impl.StaticPingMsg", 
		TAPESTRY_STATIC_PING_MSG),
	new Type("ostore.tapestry.impl.StaticPongMsg", 
		TAPESTRY_STATIC_PONG_MSG),
	// to test new type code stuff --srhea
	// new Type("ostore.tapestry.impl.RouteMsg", 
	// 	TAPESTRY_ROUTE_MSG),
	new Type("ostore.tapestry.impl.RouteMsg"),
	new Type("ostore.tapestry.impl.RouteObjRootMsg", 
		TAPESTRY_ROUTE_OBJ_ROOT_MSG),
	new Type("ostore.tapestry.impl.RouteObjDestMsg", 
		TAPESTRY_ROUTE_OBJ_DEST_MSG),
	new Type("ostore.tapestry.impl.PublishMsg", 
		TAPESTRY_PUBLISH_MSG),
	new Type("ostore.tapestry.impl.BloomWrapper", 
		TAPESTRY_BLOOM_WRAPPER),
	new Type("ostore.tapestry.impl.PingMsg", 
		TAPESTRY_PING_MSG),
	new Type("ostore.tapestry.impl.PongMsg", 
		TAPESTRY_PONG_MSG),
	new Type("ostore.tapestry.api.TapestryLocateFailure", 
		TAPESTRY_LOCATE_FAILURE),
	// ostore.tapestry.impl dynamic msgs
	new Type("ostore.tapestry.impl.DynamicRequestMsg",
		 TAPESTRY_DYNAMIC_REQ_MSG),
	new Type("ostore.tapestry.impl.DynamicMapMsg",
		 TAPESTRY_DYNAMIC_MAP_MSG),
	new Type("ostore.tapestry.impl.DynamicNotifyMsg",
		 TAPESTRY_DYNAMIC_NOT_MSG),
	new Type("ostore.tapestry.impl.DynamicDMCastMsg",
		 TAPESTRY_DYNAMIC_MCAST_MSG),
	new Type("ostore.tapestry.impl.DynamicObjMapMsg",
		 TAPESTRY_DYNAMIC_OBJMAP_MSG),
	new Type("ostore.tapestry.impl.DynamicObjReqMsg",
		 TAPESTRY_DYNAMIC_OBJREQ_MSG),
	new Type("ostore.tapestry.impl.DynamicSurrReqMsg",
		 TAPESTRY_DYNAMIC_SURRREQ_MSG),
	new Type("ostore.tapestry.impl.DynamicSurrMsg",
		 TAPESTRY_DYNAMIC_SURR_MSG),
	new Type("ostore.tapestry.impl.DynamicPingMsg", 
		 TAPESTRY_DYNAMIC_PING_MSG),
	new Type("ostore.tapestry.impl.DynamicPongMsg", 
		 TAPESTRY_DYNAMIC_PONG_MSG),
	new Type("ostore.tapestry.impl.DynamicGatewayMsg",
		 TAPESTRY_DYNAMIC_GATEWAY_MSG),
	// ostore.tapestry.impl test msg
	new Type("ostore.tapestry.impl.TestFoundMsg",
		 TAPESTRY_TEST_FOUND_MSG),
	new Type("ostore.tapestry.impl.TestLocateMsg",
		 TAPESTRY_TEST_LOCATE_MSG),
	new Type("ostore.tapestry.impl.TestReadyMsg",
		 TAPESTRY_TEST_READY_MSG),

	// ostore.oil types
  	new Type ("ostore.oil.Array", ARRAY),
  	new Type ("ostore.oil.Matrix", MATRIX),
  	new Type ("ostore.oil.SegmentPager$Request", SEGMENT_REQUEST),
  	new Type ("ostore.oil.SegmentPager$ReverseRequest",
		  SEGMENT_REV_REQUEST),
  	new Type ("ostore.oil.HMM", HMM),
  	new Type ("ostore.oil.HMM$Delta", HMM_DELTA),
  	new Type ("ostore.oil.HMM$Prediction", HMM_PREDICTION),
  	new Type ("ostore.oil.HMM$Segment", HMM_SEGMENT),
  	new Type ("ostore.oil.Screen", SCREEN),


	// patchwork types
  	new Type ("ostore.oil.patchwork.PatchworkModel$Delta", 
		  OIL_PATCHWORK_DELTA),
  	new Type ("ostore.oil.patchwork.PatchworkModel", 
		  OIL_PATCHWORK_MODEL),
  	new Type ("ostore.oil.patchwork.PatchworkModel$Observation", 
		  OIL_PATCHWORK_OBSERVATION),
  	new Type ("ostore.oil.patchwork.PatchworkProbeMsg", 
		  OIL_PATCHWORK_PROBE_MSG),


	//TODO - moved above for ordering purposes
	// ostore.cache types (for testing)
	//new Type ("ostore.cache.TestBlock", TESTBLOCK),

        // ostore.security types
	new Type("ostore.security.QSDHPublicKey", 
		SECURITY_QS_DH_PUBLIC_KEY),
	new Type("ostore.security.SignedQS", SECURITY_SIGNED_QS),
	new Type("ostore.security.QSSha1RsaSignature", 
		SECURITY_QS_SHA1_RSA_SIGNATURE),
	new Type("ostore.security.QSRsaPublicKey", 
		SECURITY_QS_RSA_PUBLIC_KEY),
	new Type("ostore.security.QSRsaPrivateKey", 
		SECURITY_QS_RSA_PRIVATE_KEY),
	new Type("ostore.security.thresh.ThresholdPublicKey", 
		SECURITY_THRESH_PUBLIC_KEY),
	new Type("ostore.security.thresh.ThresholdSignature", 
		SECURITY_THRESH_SIGNATURE),
	new Type("ostore.security.thresh.KeyShare", 
		SECURITY_THRESH_KEY_SHARE),
	new Type("ostore.security.thresh.ThresholdSignedQS"),
	new Type("ostore.security.SignedQSArray$Individual", 
		SECURITY_SIGNED_QS_ARRAY_IND),
	new Type("ostore.security.EncryptedQS", 
		SECURITY_ENCRYPTED_QS),
	new Type("ostore.security.QSRsaCipher", 
		SECURITY_QS_RSA_CIPHER),

	// ostore.bloom types

	new Type ("ostore.bloom.UpdateId",
		BLOOM_UPDATE_ID),
	new Type ("ostore.bloom.UpdateMessage",
		BLOOM_UPDATE_MESSAGE),
	new Type ("ostore.bloom.TopologyDiscoverQuery",
		BLOOM_TOPOLOGY_DISCOVER_QUERY),
	new Type ("ostore.bloom.TopologyDiscoverResponse",
		BLOOM_TOPOLOGY_DISCOVER_RESPONSE),
	new Type ("ostore.bloom.SimulatedRead",
		BLOOM_SIMULATED_READ),
	new Type ("ostore.bloom.SimulatedReadResp",
		BLOOM_SIMULATED_READ_RESP),

  	// ostore.mac types
	new Type ("ostore.mac.DHMessage",
		 MAC_DHMESSAGE),
	new Type ("ostore.mac.STSVerify",
	     MAC_STSVERIFY),
	/*new Type ("ostore.mac.MACedQS",
		 MAC_MACEDQS),
	new Type ("ostore.mac.MACMessage",
		 MAC_MACMESSAGE),*/
	new Type ("ostore.mac.QSHMACMD5",
		 MACTYPE_HMACMD5),

    };

    //**********************************************************************
    //***************** Don't touch anything below here. *******************
    //**********************************************************************

    static private class Type {
	public Type (String name, int code) {
	    this.name = name; this.code = code;
	}
	public Type (String name) {
	    this.name = name; this.code = -1;
	}
	public String name;
	public int code;
    }

    static public void registerAllTypes ()
    {
	if (registered) return;

	boolean errors = false;
	for (int i = 0; i < types.length; ++i) {
	    try
	    {
		if (types [i].code == -1)
		    TypeTable.register_type (types [i].name);
		else
		    TypeTable.register_type (types [i].name, types [i].code);
	    }
	    catch (Exception e)
	    {
		Carp.carp ("Caught a fatal exception initializing " +
			   "type (" + types [i].name + ", " + 
			   Integer.toHexString (types [i].code) + "): " + e);
		e.printStackTrace ();
		errors = true;
	    }
	}
	if (errors)
	    Carp.die ("Dying...");

	registered = true;
    }

    static private boolean registered = false;
}

