/*
 * Graph.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package soss.network;
import java.util.HashMap;
import java.util.Map;
import java.util.HashSet;
import java.util.Set;
import java.util.Collection;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.Vector;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.StreamTokenizer;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.Reader;
import ostore.util.Assert;
import ostore.util.Carp;
import ostore.util.Pair;
import ostore.util.PriorityQueue;

/**
 * Directed graph class for Transit-Stub style graphs.  
 *
 * @author Sean C. Rhea
 * @version $Id: TransitStub.java,v 1.5 2002/03/22 00:15:54 srhea Exp $
 */
public class TransitStub {

    private static final String class_tag = "TransitStub";
    private static final double MULT = 1000000.0;

    public static final class Edge {
	public Node tail;
	public double alpha, beta;
	public Edge next;
	public Map byte_cnts;
	public boolean inter_domain;

	public Edge (Node n, double a, double b, boolean interdom) {
	    tail = n; alpha = a; beta = b;
	    inter_domain = interdom;
	    next = null;
	    clear_byte_cnts ();
	}

	public void add_bytes (long type, long bytes) {
	    Long Type = new Long (type);
	    Long prev = (Long) byte_cnts.get (Type);
	    if (prev == null) {
		byte_cnts.put (Type, new Long (bytes));
	    }
	    else {
		byte_cnts.put (Type, 
			new Long (bytes + prev.longValue ()));
	    }
	}

	public void clear_byte_cnts () {
	    byte_cnts = new HashMap ();
	}
    }

    public static final class DomainId {
	public int dom, dnn, snum;
	public DomainId (int dom, int dnn, int snum) {
	    this.dom = dom;
	    this.dnn = dnn;
	    this.snum = snum;
	}
	public boolean equals (Object other) {
	    DomainId rhs = (DomainId) other;
	    if (dom != rhs.dom) return false;
	    if (snum != rhs.snum) return false;
	    if (snum == -1) return true;
	    return dnn == rhs.dnn;
	}
	public int hashCode () {
	    if (snum == -1) 
		return dom;
	    else 
		return (dom << 16) ^ (dnn << 8) ^ snum;
	}
    }

    public static final class Node {
	public int number;
	public Domain domain;
	public Edge edges;

	public Node (int n, Domain d) { 
	    number = n; 
	    this.domain = d;
	    edges = null;
	}

	public final void add_edge (Edge e) { 
	    e.next = edges;
	    edges = e;
	}

	public final Edge get_edge (Node sink) {
	    Edge result = edges;
	    while ((result != null) && (! result.tail.equals (sink)))
		result = result.next;
	    return result;
	}

	public boolean equals (Object other) {
	    Node rhs = (Node) other;
	    return rhs.number == number;
	}

	public int hashCode () { 
	    return number; 
	}
    }

    public static final class Domain {
	public DomainId id;
	public Set /* of Node */ nodes; 
	public Set /* of Node */ interdomain_edges; 

	public Domain (DomainId id) {
	    this.id = id;
	    nodes = new HashSet ();
	    interdomain_edges = new HashSet ();
	}
    }

    public static final void single_source (
	    Node [] G,
	    Node [] pred,
	    double [] dist,
	    Node source,
	    Set domain) {

	PriorityQueue Q = new PriorityQueue (G.length);
	for (int i = 0; i < G.length; ++i) {
	    Node v = G [i];
	    if (v.equals (source))
		dist [v.number] = 0.0;
	    else
		dist [v.number] = Double.MAX_VALUE;
	    pred [v.number] = null;
	}

	Q.insert (source, 0, false);

	while (! Q.empty ()) {
	    Node u = (Node) Q.extract (false);
	    for (Edge e = u.edges; e != null; e = e.next) {
		Node v = e.tail;

		assert (v.number != u.number);

		if ((domain == null) || (domain.contains (v))) {
		    if (dist [v.number] > dist [u.number] + e.alpha) {
			if (pred [v.number] == null)
			    Q.remove (v);
			dist [v.number] = dist [u.number] + e.alpha;
			pred [v.number] = u;
			Q.insert (v, Math.round (dist [v.number] * MULT), 
				false);
		    }
		}
	    }
	}
    }

    public static final void store_single_source (
	    Node [] G,
	    Node [] pred,
	    double [] dist,
	    Node source,
	    int [] pred_tbl, int pred_tbl_offset, 
	    double [] alpha, int alpha_offset) {

	int [] path = new int [G.length];
	int path_length = 0;

	for (int dest = 0; dest < G.length; ++dest) {

	    double total_alpha = 0.0;
	    Node walker = G [dest];

	    path [0] = walker.number;
	    path_length = 1;

	    if (pred [walker.number] == null)
		pred_tbl [pred_tbl_offset + dest] = -1;
	    else
		pred_tbl [pred_tbl_offset + dest] = 
		    pred [walker.number].number;

	    while (pred [walker.number] != null) {
		path [path_length++] = pred [walker.number].number;
		for (Edge e = walker.edges; true; e = e.next) {
		    if (e.tail.equals (pred [walker.number])) {
			total_alpha += e.alpha;
			break;
		    }
		}
		walker = pred [walker.number];
	    }

	    alpha [alpha_offset + dest] = total_alpha;
	}
    }

    public static final void store_single_source (
	    Node [] G,
	    Node [] pred,
	    double [] dist,
	    Node source,
	    Set domain,
	    Map pred_tbl, Map alpha) {
	
	int [] path = new int [G.length];
	int path_length = 0;

	for (Iterator d = domain.iterator (); d.hasNext (); ) {

	    double total_alpha = 0.0;
	    Node dest = (Node) d.next ();
	    Node walker = dest;

	    path [0] = walker.number;
	    path_length = 1;

	    pred_tbl.put (walker, pred [walker.number]);

	    while (pred [walker.number] != null) {
		path [path_length++] = pred [walker.number].number;
		for (Edge e = walker.edges; true; e = e.next) {
		    if (e.tail.equals (pred [walker.number])) {
			total_alpha += e.alpha;
			break;
		    }
		}
		walker = pred [walker.number];
	    }

	    alpha.put (dest, new Double (total_alpha));
	}
    }

    public TransitStub (String filename) throws IOException {

	FileInputStream is = new FileInputStream (filename);
	Reader reader = new BufferedReader (new InputStreamReader (is));
	StreamTokenizer tok = new StreamTokenizer (reader);

	tok.resetSyntax ();
	tok.whitespaceChars ((int) ' ', (int) ' ');
	tok.wordChars ((int) '-', (int) '-');
	tok.wordChars ((int) '0', (int) '9');

	int lineno = 1;
	tok.nextToken ();
	while (tok.ttype == StreamTokenizer.TT_EOL) {
	    tok.nextToken ();
	    lineno++;
	}

	// Read in the node count.
	if (tok.ttype != StreamTokenizer.TT_WORD) {
	    System.err.println ("Expected node count on line " + 
		    lineno + ": " + tok);
	    System.exit (1); 
	}

	_N = Integer.parseInt (tok.sval);
	_G = new Node [_N];
	_domains = new HashMap ();

	Carp.out ("Graph size = " + _N);

	tok.nextToken ();
	while (tok.ttype == StreamTokenizer.TT_EOL) {
	    tok.nextToken ();
	    lineno++;
	}

	// Read in all the vertices.

	for (int vertex = 0; vertex < _N; ++vertex) {
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected vertex number on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    if (Integer.parseInt (tok.sval) != vertex)
		Carp.die ("Expected vertex = " + vertex);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected dom on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    int dom = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected dnn on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    int dnn = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected snum on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    int snum = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected snn on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    int snn = Integer.parseInt (tok.sval);

	    DomainId did = new DomainId (dom, dnn, snum);
	    Domain d = (Domain) _domains.get (did);
	    if (d == null) {
		d = new Domain (did);
		_domains.put (did, d);
	    }

	    _G [vertex] = new Node (vertex, d);
	    d.nodes.add (_G [vertex]);

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }
	}

	// Read in all the edges.

	int edge_count = 0;
	LinkedList interdomain_edges = new LinkedList ();

	while (tok.ttype != StreamTokenizer.TT_EOF) {

	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected from on line " + 
			lineno + ": " + tok);
		System.exit (1); 
	    }
	    int from = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected to on line " + 
			lineno + ".");
		System.exit (1); 
	    }
	    int to = Integer.parseInt (tok.sval);

	    tok.nextToken ();
	    if (tok.ttype != StreamTokenizer.TT_WORD) {
		System.err.println ("Expected length on line " + 
			lineno + ".");
		System.exit (1); 
	    }
	    double alpha = ((double) Integer.parseInt (tok.sval)) * 0.001;

	    edge_count += 1;

	    // TODO: for now, beta's are assiged as below.
	    double beta = 0.0;

	    Node src = _G [from];
	    Node dst = _G [to];
	    boolean interdomain_edge = 
		! src.domain.id.equals (dst.domain.id);

	    if (src.domain.id.snum == -1) {
		// The source is in a transit domain.

		if (dst.domain.id.snum != -1) {
		    // The destination is in a stub domain.  This is a
		    // transit to stub edge.
		    beta = 8.0 / 1000000.0 / 1.5;
		}
		else {
		    // The destination is in a transit domain.  This is a
		    // transit to transit edge.

		    beta = 8.0 / 1000000.0 / 45;
		}
	    }
	    else {
		// The source is in a stub domain.

		if (dst.domain.id.snum == -1) {
		    // The destination is in a transit domain.  This is a
		    // stub to transit edge.

		    beta = 8.0 / 1000000.0 / 1.5;
		}
		else {
		    // The destination is in a stub domain.  This is a stub
		    // to stub edge.

		    if (interdomain_edge) {
			// They're in different stub domains.
			beta = 8.0 / 1000000.0 / 1.5;
		    }
		    else {
			// They're in the same stub domain.
			beta = 8.0 / 1000000.0 / 100;
		    }
		}
	    }

	    if (interdomain_edge) {
		
		Node n = new Node (_N + interdomain_edges.size (), 
			null);
		n.add_edge   (new Edge (src, alpha / 2.0, beta, true));
		n.add_edge   (new Edge (dst, alpha / 2.0, beta, true));
		src.add_edge (new Edge (n, alpha / 2.0, beta, true));
		dst.add_edge (new Edge (n, alpha / 2.0, beta, true));

		src.domain.interdomain_edges.add (n);
		dst.domain.interdomain_edges.add (n);

		interdomain_edges.add (n);
	    }
	    else {
		src.add_edge (new Edge (dst, alpha, beta, false));
		dst.add_edge (new Edge (src, alpha, beta, false));
	    }

	    tok.nextToken ();
	    while (tok.ttype == StreamTokenizer.TT_EOL) {
		tok.nextToken ();
		lineno++;
	    }
	}

	Carp.out ("Edges = " + edge_count);
	Carp.out ("Interdomain Edges = " + interdomain_edges.size ());

	is.close ();

	// Create a new graph, converting each interdomain edge to a node
	// with two edges only.

	Node [] G = new Node [_N + interdomain_edges.size ()];
	{
	    int i = 0;
	    while (i < _N) { 
		G [i] = _G [i];
		++i;
	    }
	    while (i < G.length) {
		G [i] = (Node) interdomain_edges.removeFirst ();
		++i;
	    }
	}
	_G = G;

	Node [] predecessors = new Node [_G.length];
	double [] distances  = new double [_G.length];

	Carp.out ("Modified graph size = " + _G.length);

	_ide_to_all_pred  = new int    [(_G.length - _N) * _G.length];
	_ide_to_all_alpha = new double [(_G.length - _N) * _G.length];

	// Compute a single source distance from each interdomain edge node
	// to all other nodes.

	for (int i = _N; i < _G.length; ++i) {
	    single_source (_G, predecessors, distances, _G [i], null);
	    store_single_source (_G, predecessors, distances, _G [i], 
		    _ide_to_all_pred, (i - _N) * _G.length, 
		    _ide_to_all_alpha, (i - _N) * _G.length);
	}

	// Compute an all pairs shortest path from every node to all of the
	// other nodes in its domain.

	_intra_dom_pred  = new HashMap (_N);
	_intra_dom_alpha = new HashMap (_N);

	for (Iterator i = _domains.keySet ().iterator (); i.hasNext (); ) {
	    Domain d = (Domain) _domains.get (i.next ());
	    for (Iterator j = d.nodes.iterator (); j.hasNext (); ) {
		Node src = (Node) j.next ();
		single_source (_G, predecessors, distances, src, d.nodes);
		Map p = new HashMap (d.nodes.size ());
		Map a = new HashMap (d.nodes.size ());
		store_single_source (_G, predecessors, distances, src, 
			d.nodes, p, a);
		_intra_dom_pred.put (src, p);
		_intra_dom_alpha.put (src, a);
	    }
	}
    }

    public int vertices () { return _N; }

    public class PathInfo {
	public double route_time;
	public int total_hops;
	public Set interdomain_edges;

	public PathInfo (double r, int t, Set i) {
	    route_time = r;
	    total_hops = t;
	    interdomain_edges = i;
	}
    }
	
    /**
     * Returns (route time, hops), keeps record of bytes sent if record is
     * true.
     */
    public PathInfo shortest_path (int s, int d, long bytes, boolean record,
	    long type) {

	assert (s < _N);
	assert (d < _N);

	Node src = _G[s];
	Node dst = _G[d];

	double total_alpha = Double.MAX_VALUE;
	double max_beta = 0.0;
	int total_hops = 0;
	Set interdomain_edges = null;

	if (src.domain.id.equals (dst.domain.id)) {
	    Map alphas = (Map) _intra_dom_alpha.get (src);
	    Double alpha = (Double) alphas.get (dst);
	    total_alpha = alpha.doubleValue ();

	    Node walker = dst;
	    //System.out.print ("Walking " + src.number + "->" + dst.number + 
		    //".  Path = " + walker.number);
	    while (! walker.equals (src)) {
		Map preds = (Map) _intra_dom_pred.get (src);
		Node pred = (Node) preds.get (walker);
		//System.out.print (" " + pred.number);
		Edge e = walker.get_edge (pred);
		if (e.beta > max_beta)
		    max_beta = e.beta;
		if (record) e.add_bytes (type, bytes);
		walker = pred;
		total_hops += 1;
	    }
	    //System.out.println ("");
	}
	else {

	    // The interdomain edge that the source will go to.
	    Node best_ides = null;
	    
	    // The interdomain edge that the destination will go to.
	    Node best_ided = null;

	    // Minimize the sum:
	    //
	    //  src -> ide -> ide -> dest
	    //
	    for (Iterator i = src.domain.interdomain_edges.iterator (); 
		    i.hasNext (); ) {

		Node ides = (Node) i.next ();
		double src_to_ide = _ide_to_all_alpha [
		    (ides.number - _N) * _G.length + src.number];

		for (Iterator j = dst.domain.interdomain_edges.iterator (); 
			j.hasNext (); ) {

		    Node ided = (Node) j.next ();
		    double dst_to_ide = _ide_to_all_alpha [
			(ided.number - _N) * _G.length + dst.number];

		    double ide_to_ide = _ide_to_all_alpha [
			(ided.number - _N) * _G.length + ides.number];

		    double total = src_to_ide + dst_to_ide + ide_to_ide;

		    if (total < total_alpha) {
			total_alpha = total;
			best_ides = ides;
			best_ided = ided;
		    }
		}
	    }

	    interdomain_edges = new HashSet ();
	    interdomain_edges.add (new Integer (best_ides.number));
	    interdomain_edges.add (new Integer (best_ided.number));

	    //System.out.print ("Walking interdomain path " + src.number + "->" + 
		    //dst.number + ".  Path =");

	    // Walk from src to best_ides.
	    Node walker = src;
	    while (! walker.equals (best_ides)) {
		//System.out.print (" " + walker.number);
		Node pred = predecessor (walker, best_ides);
		Edge e = walker.get_edge (pred);
		if (e.beta > max_beta)
		    max_beta = e.beta;
		if (record) e.add_bytes (type, bytes);
		walker = pred;
		if (walker.number >= _N) 
		    interdomain_edges.add (new Integer (walker.number));
		// Count interdomain edges once, others twice.
		if ((walker.number >= _N) ^ (pred.number >= _N))
		    total_hops += 1;
		else 
		    total_hops += 2;
	    }

	    // Walk from best_ides to best_ided.
	    walker = best_ides;
	    while (! walker.equals (best_ided)) {
		//System.out.print (" " + walker.number);
		Node pred = predecessor (walker, best_ided);
		Edge e = walker.get_edge (pred);
		if (e.beta > max_beta)
		    max_beta = e.beta;
		if (record) e.add_bytes (type, bytes);
		walker = pred;
		if (walker.number >= _N) 
		    interdomain_edges.add (new Integer (walker.number));
		// Count interdomain edges once, others twice.
		if ((walker.number >= _N) ^ (pred.number >= _N))
		    total_hops += 1;
		else 
		    total_hops += 2;
	    }

	    // Walk from dst to best_ided.
	    walker = dst;
	    while (! walker.equals (best_ided)) {
		//System.out.print (" " + walker.number);
		Node pred = predecessor (walker, best_ided);
		Edge e = walker.get_edge (pred);
		if (e.beta > max_beta)
		    max_beta = e.beta;
		if (record) e.add_bytes (type, bytes);
		walker = pred;
		if (walker.number >= _N) 
		    interdomain_edges.add (new Integer (walker.number));
		// Count interdomain edges once, others twice.
		if ((walker.number >= _N) ^ (pred.number >= _N))
		    total_hops += 1;
		else 
		    total_hops += 2;
	    }
	    //System.out.println ("");

	    assert (total_hops % 2 == 0);
	    total_hops /= 2;
	}

	double route_time = total_alpha + bytes * max_beta;

	return new PathInfo (route_time, total_hops, interdomain_edges);
    }

    public Node predecessor (Node walker, Node dst) {
	assert (dst.number >= _N);
	int p = _ide_to_all_pred [(dst.number - _N) * _G.length +
	    walker.number];
    
	if (p == -1) return dst; else return _G [p];
    }

    public void test () {
	for (int i = 0; i < _N; ++i)
	    for (int j = 0; j < _N; ++j)
		shortest_path (i, j, 1, true, 0);
    }

    public static void print_mem_usage () {
	System.gc ();
	double result = (Runtime.getRuntime ().totalMemory () - 
	                 Runtime.getRuntime ().freeMemory ()) / 1024.0;
	System.err.println ("Memory in use " + result + "k");
    }

    public static void main (String [] args) throws Exception {
	print_mem_usage ();
	System.err.print ("Computing tables...");
	long start = System.currentTimeMillis ();
	TransitStub ts = new TransitStub (args [0]);
	long end = System.currentTimeMillis ();
	System.err.println ("done in " + (end - start) + " ms.");

	print_mem_usage ();
	System.err.print ("Printing results...");
	start = System.currentTimeMillis ();
	ts.test ();
	end = System.currentTimeMillis ();
	System.err.println ("done in " + (end - start) + " ms.");
	print_mem_usage ();
    }

    public void print_byte_counts () {
	for (int src = 0; src < _N; ++src) {
	    Node nsrc = _G [src];
	    for (Edge e = nsrc.edges; e != null; e = e.next) {
		for (Iterator i = e.byte_cnts.keySet ().iterator ();
			i.hasNext (); ) {
		    Long type = (Long) i.next ();
		    Long bytes = (Long) e.byte_cnts.get (type);
		    System.out.println (src + " " + e.tail.number +
			    " " + bytes + 
			    (e.inter_domain ? " TRUE " : " FALSE ") + type);
		}
	    }
	}
    }

    public void clear_byte_counts () {
	for (int src = 0; src < _G.length; ++src) {
	    Node nsrc = _G [src];
	    for (Edge e = nsrc.edges; e != null; e = e.next) {
		e.clear_byte_cnts ();
	    }
	}
    }

    private static void assert (boolean expr) {
	Assert.assert (expr, "BUG");
    }

    int _N;
    Node [] _G;
    Map _domains;
    int [] _ide_to_all_pred;
    double [] _ide_to_all_alpha;
    Map _intra_dom_pred;
    Map _intra_dom_alpha;
}


